#!/usr/bin/perl
#
# syntax:
# 	patchlist.pl $difffile 
#
# whatfor?
#	for taking a modified kernel, and an original src kernel
#	tree and creating a set of patch files, with one patch file
#	per modified changed file.
#
# assumptions: 
#       The diff was done like this:
#	$ diff -r -c  older.dir  newer.dir > difffile
#	we assume a context diff
#
# output of patchlist.pl is:
#	1. a set of patch files named according to the diff pathname 
#	2. a file with all of the Only files listed.  onlylist
#
# a kernel may be patched by putting the difffiles in a directory
# say diff
#
# 1. to make the diff file
#	assume old files in /somewhere/older.dir
#       assume new files in /somewhere/newer.dir
#  $ diff -r -c older.dir newer.dir > diff.list
#  $ mkdir diffs
#  $ mv diff.list diffs/diff.list
#  $ cd diffs;  patchlist.pl diff.list
#  $ cd ..
#
# 2. to patch older.diff to be the same as newer.dir 
#
# assume the kernel to be patched is in /somewhere/older.diff
#
# cd to /somewhere 
# cat diffs/*patch | patch -s
#
# algorithm (roughly):
# loop through the file.diff - read a line
#	if line matches ONLY - toss
#	if line matches diff
#		if in DIFFMODE
#			close DIFFMODE file
#			if file has no count
#				unlink it
#		split and find last 2 arguments
#		open filename  last-1.last.patch
#		enter DIFFMODE
#		set count to 0
#	else
#		put line in file
#		increment count

#
# assume diff -r -c
$count = 1;
$mode  =  0;   # 
$onlyfile = "onlylist";
open (ONLYFILE, "> $onlyfile") || die "open failed: $!\n";
while (<>) {
	if (/^Only/) {
		if ($mode) {
			close (PATCHFILE);
			# zap if nothing in it
			if ($count == 0) {
				unlink($patchfile);
			}
			$mode = 0;
			$count = 0;
		}
		print ONLYFILE $_;
	}
	elsif (/^diff/) {
		if ($mode) {
			close (PATCHFILE);
			# zap if nothing in it
			if ($count == 0) {
				unlink($patchfile);
			}
			$mode = 0;
		}
		# split the diff line up into words
		@words=split;
		# extract the old dir and new dir
		@olddir = split(/\/+/, $words[3]);
		#@newdir = split(/\/+/, $words[4]);
		
		$oldlast = $olddir[$#olddir];  
		#$newlast = $newdir[$#newdir];  
		$oldprev = $olddir[$#olddir-1];  
		#$newprev = $newdir[$#newdir-1];  
		# name is next to last glued to last glued to ".patch"
		$patchfile = "$oldprev+" . $oldlast . "+patch";
		open (PATCHFILE, "> $patchfile") || die "open failed: $!\n";
		$mode = 1;
		$count = 0;
	}
	else {
		if ($mode == 1 ) {
			print PATCHFILE $_;
			$count++;
		}
	}
}
close (PATCHFILE);
close (ONLYFILE);
