
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * alist.h
 *
 *	structure of alist member.  mn list of agents.
 */

#define TRUE	1
#define FALSE	0
#include "route.h"


/* identity values */
#define OURHA		0	/* its our HA */
#define OURFA		1	/* its our FA */
#define THEIRFA   	2 	/* its a FA, but not one of ours */
#define MOBILENODE	3	/* its not an agent, but an mn */

#define MAXLOCATION     255	/* max location string len for beacon */
#ifndef MAXURL
#define MAXURL		254
#endif


/* signal strength measurement and acceptance constants */
#define MIN_GOODAVE	15	/* minimum sample required for a "good"
				 * signal strength average */
#define MIN_GOODSIG	10	
#define MAX_SIGSTR	100
#define MAX_NOBEACON	60	/* (default) max diff in seconds since last
				 *   beacon arrived */
				/* TBD: vary this according to speed
				 * of movement?  */
#define AGENTCYCLE	1	/* check for better signal at this second
				 * interval */
#define SIGFUDGE	10	/* keep the current agent unless the new
				 * agent's signal is better by this
				 * percentage */
#define RETRYWAIT	(5 * 60)	/* period of time (seconds) to wait
					 * before attempting registration
					 * with a given agent if it fails
					 * to respond after retryBackoff
					 * cycle */


/* 
 * list of agents OR mns who sent beacons (used by mn)
 */

struct alist {
	int identity;		/* ha or fa or whatever */	 

	/* secpolicies are in udp_adhoc.h */
	int secpolicy;		/* secpolicy to use with agent */

	/* public signal strength values */
	int	sigstrength;	/* signal strength of last beacon */
	int	sigaverage;	/* signal average over <= MIN_GOODAVE */

	/* private signal strength values */
	int	sigarray[MIN_GOODAVE]; /* last <= MIN_GOODAVE strength values */
	int	sigindex;	/* index into sigarray: next element to write */
	int	sigsum;		/* sum of all current members of sigarray[] */
	u_long  sigcount;	/* # of sequential checks to signal cache */
	time_t  sigtime;	/* time of last check to signal cache */

	int seqno;		/* seqno in beacon */
	int lifetime;		/* lifetime in beacon */
	int ticks;		/* lifetime in list */
	int retryWait;		/* seconds left to wait before attempting
				 * a new registration if agent failed due
				 * to backoff */
	char sawAuth;		/* beacon was authenticated */
	char auth_ok;		/* authentication matched */
	u_long auth_tssecs;	/* ts for authentication in ntp secs */
	u_long auth_tsfract;    /* ts for authentication in ntp fracts */
	struct in_addr coa;	/* for use at ha. mn's coa */
	struct in_addr bip;	/* ip addr (from) for beacon */
	short sawMac;		/* true if mac field valid */
	short mnroute;		/* true if mn route installed (ad hoc) */
	char mac[6];		/* mac address */
	char location[MAXLOCATION];
	int lsize;
	int notinlist;		/* TRUE when entry has been deleted */
	struct alist *next;	/* next list pointer */
};




/* alist prototypes
*/
struct alist * alist_search(struct alist *, struct in_addr );

struct alist * alist_findFirst(struct alist *head, int daemonType, int match);

struct alist * alist_create(struct alist **, struct in_addr );

void alist_sort( struct alist **head, struct alist *a );

int alist_delete(struct alist **, struct in_addr );

void alist_initwalk(struct alist *);

struct alist * alist_nextwalk();

void alist_dump(struct alist *);

void alist_header(int longform);

void alist_clrRetry( struct alist *head );

void checkAlist(	
	 	int useha,	
	   	struct in_addr ip_src,
 		int mse_sequence,
		int mse_lifetime,
		struct in_addr mse_coa,
		int auth_type,
		int auth_ok,
		u_long auth_tssecs, 
		u_long auth_tsfract, 
		char *location,
		int lsize,
		int sawMac,
		char * mac_addr,
		int sawmn,
		int doArp,
		struct in_addr 
		);

