/*
 * Copyright (c) 2003 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * divertInterfaces.c
 * Common functions for divert used by MN and HA.
 *
 * functions:
*/
#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_fw.h>
#include <netinet/in_systm.h>
#include <math.h>

#include "mnstate.h"
#include "mnconfig.h"
#include "udp_reg.h"
#include "mip_log.h"
#include "divertInterfaces.h" 

/*
 *  This Function returns a Divert Socket.
 *  Note : The option IPDIVERT has to be placed in the 
 *         kernel configuration file.
 */

int 
getDivertSock()
{
	int fd = -1;        
	
	/* Open A Divert Socket */
	fd=socket(PF_INET, SOCK_RAW, IPPROTO_DIVERT);
	
	if (fd < 0) {
		syslog(LOG_ERR, "getDivertSocket(): socket (%m)");
		syslog(LOG_ERR, "IPDIVERT in kernel config?");	    
		exit(1);
	}             
	return fd;
}

/*
 *   This function generates a MIP message header for data messages
 *   in the shared NAT/UDP tunnel. This header 
 *   is used to differentiate between the data and registration  
 *   messages during  the NAT traversal.
 */
 
void mipMsgHdr_gen(char *packet){
  
	struct mip_MsgHdr *msghdr;        
	
	msghdr = (struct mip_MsgHdr *) packet;
	
	msghdr->type       = MIP_MSGHDR_TYPE;
	msghdr->nextHeader = IPVERSION;
	return;
}

/*
 *   Binds the Divert Socket to a port.
 *
 */

int 
divertBind(int sock,int port)
{

	struct sockaddr_in bindPort;  
	int ret;
	
	bindPort.sin_family=AF_INET;
	bindPort.sin_addr.s_addr=INADDR_ANY;
	bindPort.sin_port=htons(port);
	
	ret=bind(sock, (struct sockaddr *)&bindPort,
		 sizeof bindPort);
	
	if (ret < 0) {
		close(sock);
		syslog(LOG_ERR, "getDivertSocket():  bind (%m)");
		exit(1);
	}
	
	return ret;
}

void 
encapsulate(unsigned char *packet,int n, struct in_addr sAddr,
		 struct in_addr dAddr, u_short port)
{
	int ii;    
	struct in_addr srcAddr;
	struct in_addr dstAddr;
	struct udphdr *udphdr;
	struct mip_MsgHdr *msgHdr;
	unsigned char encapPacket[BUFSIZE];
	unsigned char temp[BUFSIZE];
	extern int debug;
	
	srcAddr.s_addr = sAddr.s_addr;
	dstAddr.s_addr = dAddr.s_addr;
	
	/* Generate an IP-Header: note calculates ip hdr checksum
	 * which is only over ip header.
	 */ 
	ip_gen(packet,IPPROTO_UDP,srcAddr,dstAddr,
		sizeof(struct ip) + sizeof (struct udphdr) 
				 + sizeof(struct mip_MsgHdr) + n);

	/* form a pointer to udp part */
	udphdr = (struct udphdr *)(packet + sizeof(struct ip));

	/* form a pointer to mip-message part */
	msgHdr  = (struct mip_MsgHdr *) (packet + sizeof(struct ip)
					 + sizeof(struct udphdr));
	
	/* Generate the MIP message Header */ 
	mipMsgHdr_gen((char *)msgHdr);

	/* Generate the UDP Header */
	udp_gen((unsigned char *)udphdr,htons(UDP_PORT),port,
		(sizeof(struct udphdr) + n));
	
	/* udp header checksum
	 */
	udphdr->uh_sum = trans_check(IPPROTO_UDP,(unsigned char *)udphdr,
				     (n + sizeof(struct udphdr)),
				     srcAddr,
				     dstAddr);
	
	return;
}


