
/*
 * Copyright (c) 1999 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * getmac.c
 *
 * idiot code for extracting mac address for i/f at boottime.
 */
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <net/if_var.h>
#include <net/if_dl.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <arpa/inet.h>

#define	NSIP
#include <netdb.h>

#include <sys/protosw.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mip_log.h"

static char macbuf[6];

/*
 * input: char * ptr to interface name
 * output: mac address associated with interface name or NULL
 */
char *
getmac(char *ifname)
{
	int             s;
	struct ifconf   ifc;
#define MAX_INTERFACES 50	/* Yeah right. */
	char            buffer[MAX_INTERFACES * sizeof(struct ifreq)];
	struct ifreq   *ifptr, *end;
	char 		name[1024]; 
	int		rc;
	struct 		sockaddr_dl *sdl;
	char 		*c;
	struct ifreq    ifr;
	int		i;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		syslog(LOG_ERR, "getIfs(): UDP ifconfig socket (%m).");
		exit(1);
	}
	ifc.ifc_len = sizeof(buffer);
	ifc.ifc_buf = buffer;

	/*
	 * use SIOCGIFCONF socket ioctl to suck in all the interface
	 * structures in one swell foop
	 */
	if (ioctl(s, SIOCGIFCONF, (char *) &ifc) < 0) {
		syslog(LOG_ERR, "getIfs(): ifconfig (SIOCGIFCONF) (%m)");
		exit(1);
	}
	close(s);

	/*
	 * setup end pointer, ifc.ifc_len was returned with true length
	 */
	end = (struct ifreq *) (ifc.ifc_buf + ifc.ifc_len);
	ifptr = ifc.ifc_req;
	while (ifptr < end) {
#ifdef TESTMAIN
		sprintf(ifr.ifr_name, "%s", ifptr->ifr_name);
		sprintf(name, "%s", ifptr->ifr_name);
		printf("name %s, family %d\n", name,
				ifptr->ifr_addr.sa_family);
#endif
 		if (ifptr->ifr_addr.sa_family == AF_LINK) { 
                        sdl = (struct sockaddr_dl *) &ifptr->ifr_addr;
			if (strncmp(ifname, ifptr->ifr_name, 20) == 0) {
				c = (char *) LLADDR(sdl);
				for ( i = 0; i < 6; i++) {
					macbuf[i] = *c++;
				}
				return(macbuf);
			}
                }
advance:
		/* advance ifptr
		*/
		if (ifptr->ifr_addr.sa_len)	/* Dohw! */
			ifptr = (struct ifreq *) ((caddr_t) ifptr +
						  ifptr->ifr_addr.sa_len -
						  sizeof(struct sockaddr));
		ifptr++;
	}
	return(NULL);
}

#ifdef TESTMAIN
main(argc, argv)
	int argc;
	char *argv[];
{
	int i;
	u_char *mac_addr;
	if (argc<2) {
		fprintf(stderr, "usage: ether if_name\n");
		exit(1);
	}
	if (mac_addr=getmac(argv[1])) {
		printf("MAC addr for %s is ", argv[1]);
		for (i=0;i<6;i++) 
			printf("%02x%c", *mac_addr++ & 0xff, (i<5)?':':' ');
		printf("\n");
	}
	else printf("No mac addr for %s\n", argv[1]);
}
#endif
