/*	harp.c */
#ifdef HARP
#define MIPSEC

#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>

#include "harp.h"
#include "mip_log.h"
#include "mipconfig.h"
#include "route.h"
#include "vlist.h"
#include "udp_reg.h"
#include "faha_pinfo.h"
#include "mipstats.h"

struct sockaddr_in harpRegRem;  /* remote IP ( co-HA ) */
struct sockaddr_in harpRegLoc;  /* local IP ( this HA ) */

extern struct ha_stats ha_stats;

extern struct mip_config mip_config;
extern struct vlist *hahead;
extern int debug;
extern harpSocket;
int harp_ackTimer = HARP_ACK_TIMELIMIT;
static int harp_sendSock;
int harp2 = FALSE;

/* authentication variables, not exported, except for the key itself */
#ifdef MIPSEC
char *haha_key;	/* haha Md5 authentication key */
static u_long haha_spi;
static int haha_keysize;
static int haha_transform;
static u_long dummy;
#endif

#ifdef MIPSEC
/* external authentication functions */
extern char* storeMd5( char*, int, unsigned char*, char* );  
extern char* computeMd5( char*, int, char*, char* );
extern char* get_key( struct keylist*, u_long, int*, u_long*, int*, u_long*, int* );
#endif

int initHarpSocket( int );	/* open TCP connection to co-HA */
u_long GetHaIp();		/* get HA ip address */
void recvHarpDump( int );	/* request registration dump from co-HA */
void harpRegDump( int );	/* dump registration information to co-HA */
void vlist_to_nat_ext(struct vlist v, struct mse_nat_ext_req *nat_ext);
void vlist_to_reg( struct vlist, struct udp_register* );
void harpReadSock( int, char*, int );	/* read from harp TCP socket */
void harpForward( int, struct packet_info *, int );
void harp2ArpOff();
void harp2ArpOn();
void harp2Arp(int);
extern void haUdp(struct mip_ifs *, struct sockaddr_in *, struct packet_info *, int);
extern void sendOneUdp(int , char *, int , struct in_addr, u_short);
extern void set_arp( char*, char*, short );
extern int dumpme( char*, int );

/* open TCP connection with co-HA. & create UDP send port */
int initHarpSocket( int port_num )
{
	int sd;
	int rc;
	struct timeval time_out;  /* socket timeout limit */
	int dummy2;

#ifdef MIPSEC
	/* get haha authentication key */
	haha_key = get_key( mip_config.sa[HAHASECURITY].kl, 
				mip_config.coha.sin_addr.s_addr, 
				&haha_keysize, &haha_spi, &haha_transform,
				&dummy, &dummy2 );
#endif

	/* create udp send socket */
	harp_sendSock = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
	if ( harp_sendSock < 0 ){
		mip_log( MLOG_ERR, "initHarpSocket(): failure in opening HARP send socket\n" );
		return ERROR;
	}

	/* set timeout limits */
	time_out.tv_sec = 5;  
	time_out.tv_usec = 0;

	/* configure TCP connection at remote end  */
	harpRegRem.sin_family = AF_INET;
	harpRegRem.sin_port = htons( port_num );
	harpRegRem.sin_addr.s_addr = mip_config.coha.sin_addr.s_addr;

	/* configure TCP connection at local end */
	harpRegLoc.sin_family = AF_INET;
	harpRegLoc.sin_port = htons( port_num );
	harpRegLoc.sin_addr.s_addr = htonl( INADDR_ANY );

	/*  create HARP registration socket */
	if (( sd = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP ))< 0 ){
		mip_log( MLOG_ERR, "initHarpSocket(): harp TCP socket failed\n" );
		return ERROR;
	}
		
	/* set socket to non-blocking */
	rc = setsockopt( sd, SOL_SOCKET, SO_RCVTIMEO, &time_out, sizeof( struct timeval ));
	if ( rc < 0 ){
		perror( "setsockopt failed in initHarpSocket()" );
		exit(1);
	} 

	/*  we are booting, so we will attempt to get a *
	*	registration dump from our co-HA 	*/
	if ( connect( sd, (struct sockaddr*)&harpRegRem, 
		sizeof(struct sockaddr_in)) > ERROR ){
			if ( debug ){
				printf( "HARP:initHarpSocket(): connecting to co-HA\n" );
			}
			ha_stats.harpconnects++;
			recvHarpDump( sd );
	} 
	else if ( debug ){
		printf( "HARP:initHarpSocket(): failed to get reg dump from co-HA\n" );
	}
	close( sd );
	if (( sd = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP ))< 0 ){  
		mip_log( MLOG_ERR, "HARP:initHarpSocket(): harp TCP socket failed\n" );
		return ERROR;
	}

	/* bind to socket and wait to do Reg Dump for our co-HA */
	if ( bind( sd, (struct sockaddr*)&harpRegLoc, sizeof(harpRegLoc))< 0 ){
		mip_log( MLOG_ERR, "initHarpSocket(): harp bind failed\n" );
		perror("bind failed");
		exit(1);
	}
	listen(sd, 5);
	if ( debug ){
		printf("HARP:initHarpSocket(): bound and listening for co-HA requests\n" );
	}

	return sd;
}

/* 	Get ip address of HA interface, Used by Harp Ping & Ack */
u_long getHaIp()
{
	int i;
	for ( i=0; i< MAX_INTERFACES; i++){
		if ( mip_config.mip_ifs[i].amfa == 0 ){/* found HA interface */
			return (mip_config.mip_ifs[i].if_sin.sin_addr.s_addr );
		}
	}
	return 0;
}

/*	Dump our registration table down HARP tcp connection 
    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     type=HARP_REG_DUMP          |    size=8                   |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                   Number of Registrations                     |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                   Size of Registrations                       |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                  Mobile-IP registration packet ...            |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                  Mobile-IP registration packet ...            |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |              etc.                             | 

note: registration packet MAY have extensions.  

*/


void harpRegDump( int s )
{
	int newsd;	/* socket descriptor */
	int sz;		/* size of sockaddr */
	unsigned long count=0, reg_size; /* number of, size of registrations */ 
	unsigned short type, size;
	struct vlist *vptr; 
	char packet[1024];
	char *pkt_ptr = packet;
	char hash[16];
	int nwritten;
	struct packet_info pinfo;
	struct udp_register reg;

	pinfo.reg = &reg;
	pinfo.hasNat = 0;
	pinfo.sendNatRep = 0;
	
	/*  accept socket connection */ 
	sz = sizeof(struct sockaddr);
	newsd = accept( s, (struct sockaddr*)&harpRegRem, &sz );
	if ( newsd < 0 ){
		mip_log( MLOG_ERR, "HARP: harpRegDump(): accept failed\n" );
		if ( debug ){
			printf( "HARP: harpRegDump(): accept failed\n" );
		}
		return;
	}
	ha_stats.harpaccepts++;

	/* recv dump request */
	harpReadSock(newsd, pkt_ptr, sizeof(short));

	type = ntohs(*((u_short*)pkt_ptr));
	pkt_ptr += sizeof(short);
	if ( type != HARP_DUMP_REQ ){
		mip_log( MLOG_ERR, "HARP: harpRegDump(): type failure\n" );
		return;
	}
	/*	read size */
	harpReadSock( newsd, pkt_ptr, sizeof(short));
	size = ntohs(*((u_short*)pkt_ptr));
	pkt_ptr += sizeof(short);
	if ( size != HARP_DUMP_REQ_SIZE ){
		mip_log( MLOG_ERR, "HARP: harpRegDUMP(): size failure\n" );
		return;
	}
#ifdef MIPSEC
	/*	read authentication */
	harpReadSock( newsd, pkt_ptr, 16 );
	if ( debug ){
		printf( "HARP: harpRegDump(): read auth (%x)\n", pkt_ptr );
	}
	computeMd5( packet, (pkt_ptr-packet), haha_key, hash );
	if ( bcmp( hash, pkt_ptr, 16 ) != 0){
		mip_log(MLOG_ERR, "HARP: harpRegDump(): auth failed h=(%x) p=(%x)\n", hash, pkt_ptr );
		return;
	}
#endif

	if ( debug ){
		printf("HARP: harpRegDump(): recvd reg dump request from co-HA\n" );
	}

	pkt_ptr = packet;	/* reset pointer */

	/* send dump message Type */
	type = HARP_REG_DUMP;
	type = htons( type );
	memcpy( pkt_ptr, (char*)&type, sizeof(short));
	nwritten = write( newsd, pkt_ptr, sizeof( short ));
	pkt_ptr += sizeof(short);
	if ( nwritten < sizeof( short )){
		mip_log( MLOG_ERR, "HARP: harpRegDump(): socket Type write failure\n" );
		return;
	}
	/* send size Length */
	size = HARP_REG_DUMP_SIZE;
	size = htons( size );
	memcpy( pkt_ptr, (char*)&size, sizeof(short));
	nwritten = write( newsd, pkt_ptr, sizeof( short ));
	pkt_ptr += sizeof(short);
	if ( nwritten < sizeof( short )){
		mip_log( MLOG_ERR, "HARP: harpRegDump(): socket Length write failure\n" );
	}
	if ( debug ){
		printf( "HARP:harpRegDump(): sent dump message\n" );
	}

	/* count entries in vlist */
	vlist_initwalk(hahead);
	while ( vlist_nextwalk()){
		count++;
	}
	/*  send number of vlist entries ( L ) */
	count = htonl( count );
	memcpy( pkt_ptr, (char*)&count, sizeof(long));
	nwritten = write( newsd, pkt_ptr, sizeof( long ));
	pkt_ptr += sizeof(long);
	if ( nwritten < sizeof( long )){
		mip_log( MLOG_ERR, "HARP: harpRegDump(): socket vlist count write failure\n" );
		return;
	}
	if ( debug ){
		printf( "HARP:harpRegDump(): send vlist count: (%ld)\n", ntohl(count)); 
	}
	reg_size = sizeof( struct udp_register );
	reg_size = htonl( reg_size );
	memcpy( pkt_ptr, (char*)&reg_size, sizeof(long));
	nwritten = write( newsd, pkt_ptr, sizeof( long ));
	pkt_ptr += sizeof(long);
	if ( nwritten < sizeof( long )){
		mip_log( MLOG_ERR, "HARP: harpRegDump(): reg size message failed\n" );
		return;
	}
#ifdef MIPSEC
	/*	send hash */
	storeMd5( packet, (pkt_ptr-packet), hash, haha_key );
	nwritten = write( newsd, hash, 16 );
	if ( nwritten < 16 ){
		mip_log( MLOG_ERR, "HARP: harpRegDump(): write failed on hash\n");  
		return;
	}
	if ( debug ){
		printf("HARP: harpRegDump(): sent hash (%x)\n", hash );
	}
#endif
	
	/*  dump vlist ( V ) */
	vlist_initwalk(hahead);
	while( vptr = vlist_nextwalk()){

		struct udp_register *r;
		struct mse_nat_ext_req *nat;

		pkt_ptr = packet;
		r = (struct udp_register*)pkt_ptr;

		if ( debug ){
			printf( "HARP: harpRegDump(): dumping entry\n" );
		}
		pinfo.psize = sizeof( struct udp_register );

		vlist_to_reg( *vptr, r );
		pinfo.packet = packet;

		pkt_ptr += sizeof( struct udp_register );
		
#ifdef DIVERT
		/* Setup a pointer to the NAT Registration Extension, if any */
		nat = (struct mse_nat_ext_req*)pkt_ptr;
		
		/* test for nat in vlist
		*/
		pinfo.hasNat = vptr->hasNat; 
		if(vptr->hasNat) {
			/* Load NAT Ext, if any */
			vlist_to_nat_ext(*vptr, nat);
			pinfo.nat_ext_req = nat;
			pinfo.udp_src.sin_addr.s_addr = vptr->udp_src.sin_addr.s_addr;
			pinfo.udp_port = vptr->udp_port;
			pinfo.reg = r;
		}
#endif
		harpForward( newsd, &pinfo, HARP_REG_DUMP ); 
	}
}

#ifdef DIVERT
void 
vlist_to_nat_ext(struct vlist v, struct mse_nat_ext_req *nat_ext)
{
	nat_ext->type          = MN_HA_NAT_TYPE_REQ;  
	nat_ext->length        = v.length;
	nat_ext->sub_type      = MN_HA_NAT_SUBTYPE;
	nat_ext->reserved_1    = v.reserved_1;
	nat_ext->reserved_2    = v.reserved_2;
	nat_ext->Encapsulation = v.Encapsulation;
	nat_ext->reserved_3    = ntohs(v.reserved_3);
}
#endif
	
void 
vlist_to_reg( struct vlist v, struct udp_register *reg )
{
	reg->type = UDP_REGISTER_TYPE;
	reg->flags = 0;		/* none supported */
	reg->lifetime = v.orig_lifetime;
	reg->home_address = v.mn;
	reg->home_agent = v.ha;
	reg->coa = v.coa;
	reg->i.l[0] = v.i.l[0];
	reg->i.l[1] = v.i.l[1];
}

/*   send dump request byte then recieve registration table, *
 *	create pinfo structure & forward to haUdp()          
    0               1               2               3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     type=HARP_DUMP_REQ         |        size=0                |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*/
void recvHarpDump( int s )
{
	char buf[1024];
	char *buf_ptr = buf;
	char hash[1024];
	int i, nwritten;
	int offset = 0;
	int leng = 0;
	int auth = TRUE;
	u_char ext_type;
	unsigned short type, size;
	unsigned long count=0, reg_size;
	struct packet_info pinfo;
	struct mse_nat_ext_req *nat;
	struct harpNat_ext     *harpNat;



	/*	write dump request */
	type = HARP_DUMP_REQ;
	type = htons( type );
	memcpy( buf_ptr, (char*)&type, sizeof(short));
	nwritten = write( s, buf_ptr, sizeof(short));
	buf_ptr += sizeof(short);
	if ( nwritten < sizeof(short) ){
		mip_log( MLOG_ERR, "HARP: recvHarpDump(): socket write error\n" );
		return;
	}
	/*  write message size */
	size = HARP_DUMP_REQ_SIZE;
	size = htons( size );
	memcpy( buf_ptr, (char*)&size, sizeof(short));
	nwritten = write( s, buf_ptr, sizeof( short ));
	buf_ptr += sizeof(short);
	if ( nwritten < sizeof( short )){
		mip_log( MLOG_ERR, "HARP: recvHarpDump(): socket write error\n" );
		return;
	}

#ifdef MIPSEC
	storeMd5( buf, (buf_ptr-buf), hash, haha_key );
	nwritten = write( s, hash, 16 );
	if ( nwritten < 16 ){
		mip_log( MLOG_ERR, "HARP: recvHarpDump(): socket write error\n" );
	}

	if ( debug ){
		printf( "HARP: recvHarpDump(): hash= (%x)\n", hash );
		printf( "HARP: recvHarpDump(): wrote dump request\n" );
	}
#endif

	buf_ptr = buf;	/* reset buf pointer */

	/*	read dump message (T) */
	harpReadSock( s, buf_ptr, sizeof(short));
	type = ntohs(*((u_short*)buf_ptr));
	buf_ptr += sizeof(short);
	if ( type != HARP_REG_DUMP ){
		mip_log( MLOG_ERR, "HARP: recvHarpDump(): dump message error\n" );
		return;
	}

	/*	read message size (L) */
	harpReadSock( s, buf_ptr, sizeof(short));
	size = ntohs(*((u_short*)buf_ptr));
	buf_ptr += sizeof(short);
	if ( size != HARP_REG_DUMP_SIZE ){
		mip_log( MLOG_ERR, "HARP: recvHarpDump(): dump messages error\n" );
		return;
	}

	/* 	read in registration count */
	harpReadSock( s, buf_ptr, sizeof( long ));
	count = ntohl(*((u_long*)buf_ptr));
	buf_ptr += sizeof(long);
	if ( debug ){
		printf( "HARP: recvHarpDump(): read in vlist count (%ld)\n", count );
	}

	/*	read in registration size */
	harpReadSock( s, buf_ptr, sizeof( long ));
	reg_size = ntohl(*((u_long*)buf_ptr));
	buf_ptr += sizeof(long);
	if ( reg_size != ( sizeof( struct udp_register )
			)){
		mip_log( MLOG_ERR, "recvHarpDump(): reg size wrong\n" );
		return;
	}

#ifdef MIPSEC
	/*	read authentication */
	harpReadSock( s, buf_ptr, 16 );
	computeMd5( buf, (buf_ptr-buf), haha_key, hash );
	if ( bcmp( hash, buf_ptr, 16 ) != 0){
		mip_log( MLOG_ERR, "recvHarpDump(): auth failed, ignoring registrations\n" );
		auth = FALSE;
	}
#endif

	/* read in registrations */
	for (i=0; i< count; i++ ){
		u_short t, sz;
		char *packet; 

		bzero(&pinfo, sizeof(struct packet_info));
		buf_ptr = buf;
		if ( debug ){
			printf("HARP: recvHarpDump(): reading registration \n");
		}

		/* read type */
		harpReadSock( s, buf_ptr, sizeof( short ));
		t = ntohs( *((u_short*)buf_ptr ));
		buf_ptr += sizeof(short);
		if ( debug ){
			printf("HARP: recvHarpDump(): read type %d\n", t );
		}

		/* read size */
		harpReadSock( s, buf_ptr, sizeof( short ));
		sz = ntohs( *((u_short*)buf_ptr ));
		buf_ptr += sizeof(short);
		if ( debug ){
			printf("HARP: recvHarpDump(): read size %d\n", sz );
		}
			
		/* read entire registration data */
		harpReadSock( s, buf_ptr, sz );
		if ( debug ){
			printf("HARP: dumping read registration data\n");
			dumpme( buf_ptr, sz );
		}
		packet = buf_ptr;
		buf_ptr += sz;
		if ( debug ){
			printf("HARP: recvHarpDump(): read data\n" );
		}
		
		ha_stats.harpdump++;

		/* skip udp registration */
		offset += sizeof(struct udp_register); 

		/* 
		 * Examine the size value of the HARP_FORWARD message 
		 * We have already dealt with the basic mip registration.
		 * However there may be extensions following it.
		 */
		if ( sz > sizeof(struct udp_register)) {

			/* Process Extensions */

			/* Present Order of Extensions is ..
			 *   
			 *  First NAT Registration Request Extension.   
			 *  Second HARP_NAT Extension.
			 *    
			 */

			while( offset < sz ) {
				type = ( *((u_char*)(packet  + offset)));
				leng = ( *((u_char*)(packet  + offset + 1)));
				switch(type) {
					
				case MN_HA_NAT_TYPE_REQ :
					if (debug) {
						printf("MN_HA_NAT_TYPE_REQ found --> ");
						printf("Length of Ext: %x \n",leng); 
					}
					pinfo.nat_ext_req = (struct mse_nat_ext_req *) (packet+offset);
					pinfo.hasNat      = 1;
					ha_stats.harpdumpnat++;
					break;

				case HARP_NAT_EXT:
					if (debug) {
						printf("HARP_NAT_EXT found --> ");
						printf("Length of Ext: %x \n",leng);
					}
					pinfo.harp_nat_ext = (struct harpNat_ext *) (packet+offset);
					ha_stats.harpdumpharpnat++;
					break;
						
				default:
					if (debug) {
						printf("Ignoring Unrecognized Extension \n");
					}
					break;
				}
				offset = offset + leng + 2;
			}

		} 

		/* Set the offset to Zero for the next iteration, if any */
		offset = 0;
#ifdef MIPSEC
		/* read hash */

		harpReadSock( s, buf_ptr, 16 );
		if (debug) { 	
			dumpme(buf_ptr,16);
		}

		computeMd5( buf, (buf_ptr-buf), haha_key, hash );
		if ( bcmp( hash, buf_ptr, 16 ) != 0){
			mip_log(MLOG_ERR, "HARP: recvHarpDump() auth failed, ignoring all further registrations\n" );
			auth = FALSE;
		}
		if ( debug ){
			printf( "recvHarpDump: compared hashes (%x) (%x)\n" ,hash, buf_ptr );
		}
#endif

		if ( t != HARP_FORWARD ){
			mip_log( MLOG_ERR, "recvHarpDump(): wrong type\n" );
			return; 
		}
		
		pinfo.packet = packet;
		pinfo.psize = sz;

		pinfo.reg = (struct udp_register*)packet;
		pinfo.fa_x_auth = 0;
		pinfo.fa_x_auth_size = 0;
		pinfo.fa_x_auth_count = 0;

		pinfo.postmnha_size = 0;
		pinfo.forward = TRUE;

		/* call haUdp() with this packet */
		if ( debug ){
			printf( "HARP: recvHarpDump(): calling haUdp()\n" );
		}
		if ( auth == TRUE ){
			haUdp( NULL, &harpRegRem, &pinfo, NULL );
		}
		else if ( debug ){
			printf(" recvHarpDump(): ignoring reg due to bad auth\n" );
		}
	}
}

/*	read size bytes from tcp socket */
void harpReadSock( int sd, char *buf, int size )
{
	int nread;

	while( size ){
		nread = read( sd, buf, size );
		if ( nread < 0 ){
			if ( errno == EWOULDBLOCK && debug ){
				printf( "HARP: harpReadSock(): would have blocked on read\n" );
			}
			mip_log( MLOG_ERR, "HARP: harpReadSock(): socket read error\n ");
			return;
		}
		buf += nread;
		size -=nread;
	}
	return;
}

/* forward UDP registration packet to co-HA
    0               1               2               3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     type=HARP_FORWARD          |     size                     |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                1 Mobile-IP registration packet ...            |
   |                                                               |

note: the above includes any extensions of interest; i.e., udp registration
part plus supplemental extensions.
*/

/*
 * harpForward is called from two places:
 *	1. mipstate.c ... when a UDP registration is about to be acked.
 *		we call it to forward it to the peer.
 *		Note: in that case, the registration is the packet in the
 *		pinfo structure.  This MAY include the new MIP NAT extension.
 *	2. from the harp boottime dump code, to send an individual registration.
 *		
 * NAT note: because we don't send the ip/udp portion of the registration,
 * we have a special MIP HARP/NAT extension that provide the IP src of the
 * NAT box, as well as the udp src/dst ports.  Neither of which are included
 * in the Mobile-IP registration.  We need this supplement so that the HARP
 * peer can form the same UDP tunnel to the MN.
 */ 

void 
harpForward( int sock, struct packet_info *pinfo, int t)
{
	char buf[ 1024 ];
	char *buf_ptr = buf;
	u_short type, length;
	u_long tmp;
	/* debugging */

	if ( debug ){
		printf( "HARP: harpForward(): forwarding reg. packet to %s: %d\n",
			inet_ntoa( mip_config.coha.sin_addr ), HARP_PORT );
	}

	ha_stats.harpforward++;
	
	type = htons( HARP_FORWARD );/*	  (T) */  
	memcpy( buf_ptr, (char*)&type, sizeof(short));
	buf_ptr += sizeof(short);

	/*
	 * The size should include the size of the MIP Registration 
	 * PLUS the sizeof extensions, if any.
	 *
	 * if (pinfo->hasNat) {
	 *	size = sizeof(MIP Registration) + sizeof(MIP NAT Request Ext)
         *                                      + sizeof(HARP NAT Ext) 
	 * } else {
	 *      size = sizeof(MIP Registration)
	 * }    
	 *
	 */
	if (pinfo->hasNat) {
		length = sizeof(struct udp_register) + sizeof(struct mse_nat_ext_req)
					+ sizeof(struct harpNat_ext);
	} else {
		length = sizeof(struct udp_register);
	} 
	
	tmp = htons( length );
	memcpy( buf_ptr, (char*)&tmp, sizeof(short)); 
	buf_ptr += sizeof(short);

	memcpy( buf_ptr, pinfo->packet, pinfo->psize );
	buf_ptr += pinfo->psize;

	/* Append the NAT Request Ext and Harp-Nat Extension  */
	if ( pinfo->hasNat ){
	  
		u_short type,size;	  
		u_long  natSrc;
		u_short pad = 0x0000;
		u_short srcPort,dstPort;
		struct vlist *visitor;
		struct vlist *hhead;
		u_char ext_type; 
		
		/* load the NAT Request Extension, only if type == HARP_REG_DUMP  */
		if ( t == HARP_REG_DUMP ) {
			if (pinfo->nat_ext_req){		  
				memcpy( buf_ptr, pinfo->nat_ext_req, sizeof(struct mse_nat_ext_req));
				buf_ptr += sizeof(struct mse_nat_ext_req);
			}
		}

		ha_stats.harpforwardhasnat++;

		/* Load the HARP-NAT Extension */
		
		/* load the type */
		ext_type = HARP_NAT_EXT;
		memcpy(buf_ptr,(char*)&ext_type,sizeof(u_char)); 
		buf_ptr += sizeof(u_char);
		ha_stats.harpforwardhasharpnat++;
		
		/* load the length */
		size = sizeof(u_long) + (3 * sizeof(u_short)) ;
		memcpy(buf_ptr,(char*)&size,sizeof(u_char));
		buf_ptr += sizeof(u_char);
		
		/* load the value fields */
	
		/*  Pad bytes */
                pad = htons ( pad );
                memcpy(buf_ptr,(char*)&pad,sizeof(u_short));
                buf_ptr += sizeof(u_short);

	
		/* Nat IP Source */		  
		natSrc  = pinfo->udp_src.sin_addr.s_addr;
		natSrc  = htonl(natSrc);
		memcpy(buf_ptr,(char*)&natSrc,sizeof(u_long));
		buf_ptr += sizeof(u_long);
		
		/* Source Port the registration came in */ 
		srcPort = pinfo->udp_port;
		srcPort = htons(srcPort);
		memcpy(buf_ptr,(char*)&srcPort,sizeof(u_short));
		buf_ptr += sizeof(u_short);
		
		/* Destination Port of the Registration Packet */
		dstPort = UDP_PORT;
		dstPort = htons(UDP_PORT);
		memcpy(buf_ptr,(char*)&dstPort,sizeof(u_short));
		buf_ptr += sizeof(u_short);
	}
	
#ifdef MIPSEC
	storeMd5( buf, (buf_ptr-buf), buf_ptr, haha_key );
	if ( debug ){
		printf( "dumping hash:\n" );
		dumpme( buf_ptr, 16 );
	}
	buf_ptr += 16;
#endif

	if ( t == HARP_FORWARD ){
		sendOneUdp( harp_sendSock, buf, (buf_ptr - buf), 
			mip_config.coha.sin_addr, htons(HARP_PORT ));
	}
	else if ( t == HARP_REG_DUMP ){
		write( sock, buf, (buf_ptr - buf));
		if (debug ) {
			dumpme(buf,(buf_ptr - buf));
		}
	} 
}

/*
    0               1               2               3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     type=HARP_PING             |       size=4                 |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     HARP peer IP address                          |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*/

void harpPing()
{
	char buf[1024];
	char *buf_ptr = buf;
	u_short type, length, tmp;
	u_long my_ip;

	type = htons( HARP_PING );	/*  (T)  */
	memcpy( buf_ptr, (char*)&type, sizeof( short ));
	buf_ptr += sizeof( short );

	length = HARP_PING_SIZE;  /* (L) */
	tmp = htons( length );
	memcpy( buf_ptr, (char*)&tmp, sizeof( short ));
	buf_ptr += sizeof( short );

	/* HA ip */
	my_ip = getHaIp();
	if ( my_ip == 0 ){
		mip_log( MLOG_ERR, "HARP: harpPing, failed to get HA IP\n" );
		return;
	}
	my_ip = htonl( my_ip );
	memcpy( buf_ptr, (char*)&my_ip, sizeof(long)); 
	buf_ptr += sizeof( long );

#ifdef MIPSEC
	/* auth hash */
	storeMd5( buf, (buf_ptr - buf), buf_ptr, haha_key );
	buf_ptr += 16;
#endif

	if ( debug ) {
		printf( "HARP: harpPing(): sending harp ping\n" );
	}
	
	sendOneUdp( harp_sendSock, buf, (buf_ptr - buf), mip_config.coha.sin_addr, htons(HARP_PORT));
}

/*
    0               1               2               3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     type=HARP_ACK              |        size=4                |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     HARP peer IP address                          |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*/

void harpAck()
{
	char buf[1024];
	char *buf_ptr = buf;
	u_short type, length, tmp;
	u_long my_ip;

	type = htons( HARP_ACK );
	memcpy( buf_ptr, (char*)&type, sizeof( short ));
	buf_ptr += sizeof( short );

	length = 0;  /*  (L)  */
	tmp = htons( length );
	memcpy( buf_ptr , (char*)&tmp, sizeof( short ));
	buf_ptr += sizeof( short );

	/* 	HA ip */
	my_ip = getHaIp();
	if ( my_ip == 0 ){
		mip_log( MLOG_ERR, "HARP: harpAck, failed to get HA IP\n" );
		return;
        }
        my_ip = htonl( my_ip );
        memcpy( buf_ptr, (char*)&my_ip, sizeof(long));
	buf_ptr += sizeof( long );

#ifdef MIPSEC
	/* auth hash */
	storeMd5( buf, (buf_ptr - buf), buf_ptr, haha_key );
	buf_ptr += 16;
#endif

	if ( debug ) {
		printf( "HARP: harpAck(): sending harp ACK\n" );
	}

	sendOneUdp( harp_sendSock, buf, (buf_ptr - buf) , mip_config.coha.sin_addr, htons( HARP_PORT ));
}


/* turn off arp on interface for secodary HA */
#define ON 1
#define OFF 0
void harp2ArpOn(){ harp2Arp(ON); }
void harp2ArpOff(){ harp2Arp(OFF); }

void harp2Arp( int arg )
{

	/* if option is not on, ignore these calls */
	if ( !mip_config.harp2_arp_disable ){
		return;
	}
	if ( arg == OFF ){
		if ( debug ){
			printf( "HARP: harp2Arp(): turning arp off on interface %s\n",  mip_config.harp2_if );
		}
		mip_log( MLOG_INFO, "Turning arp off on interface %s\n", mip_config.harp2_if );
		set_arp( mip_config.harp2_if, "-arp", 128 );
	} else {
		if ( debug ){
			printf( "HARP: harp2Arp(): turning arp on on interface %s\n", mip_config.harp2_if );
		}
		mip_log( MLOG_INFO, "Turning arp on on interface %s\n", mip_config.harp2_if );
		set_arp( mip_config.harp2_if, "arp", -128 );
	}
}
	
#endif /* HARP */
