
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "icmpstats.h"

void
dumpBeacons(struct icmp_stats *s, int longflag) 
{
	int badauth;
	badauth = s->ha_netauthbad + s->fa_netauthbad + s->mn_netauthbad +
		s->ha_adhoc_authbad + s->fa_adhoc_authbad + s->mn_adhoc_authbad;

	printf("total beacons sent %d\n",
		s->ha_beacons_sent + s->fa_beacons_sent +
			s->combined_beacons_sent + s->mn_beacons_sent);
	printf("total beacons received %d\n", s->all_beacons_recv); 
	printf("total auth beacons received %d\n", s->all_netauth_recv); 
	printf("valid beacons received %d\n", s->valid_beacons_recv); 
	printf("valid non auth beacons received %d\n", s->valid_arp_beacons);
	printf("valid auth beacons received %d\n", s->valid_auth_beacons);
	printf("total ICMP protocol errors %d\n", 
		s->tooshort + s->malformed + s->msetooshort +
		s->invalidicmpmse + s->auth_nomac);
	printf("total authentication (present) failures %d\n", badauth);
	printf("total agent beacons ignored because lack authentication %d\n",
		s->ha_maybe_arp_lacks_auth + s->fa_maybe_arp_lacks_auth);
	if (!longflag)
		return;
	printf("---- even more info --------------\n");
	printf("counts:\n");
	printf("\tha beacons sent %d\n", s->ha_beacons_sent);
	printf("\tfa beacons sent %d\n", s->fa_beacons_sent);
	printf("\tcombined beacons sent %d\n", s->combined_beacons_sent);
	printf("\tmn beacons sent %d\n", s->mn_beacons_sent);
	printf("\tbeacons from fa that is our ha %d\n", s->faisha_recv);
	printf("\tcombined beacons recv %d\n", s->bothhafa);
	printf("ICMP protocol errors:\n");
	printf("\tICMP message too short %d\n",s->tooshort);
	printf("\tICMP message malformed %d\n",s->malformed);
	printf("\tICMP mse part too short %d\n",s->msetooshort);
	printf("\tICMP mse part invalid %d\n",s->invalidicmpmse);
	printf("\tICMP beacon authenticated but no MAC part %d\n",
			s->auth_nomac);
	printf("authentication errors:\n");
	printf("\tha network authentication failure %d\n",s->ha_netauthbad);
	printf("\tfa network authentication failure %d\n",s->fa_netauthbad);
	printf("\tmn network authentication failure %d\n",s->mn_netauthbad);
	printf("\tha adhoc authentication failure %d\n",s->ha_adhoc_authbad);
	printf("\tfa adhoc authentication failure %d\n",s->fa_adhoc_authbad);
	printf("\tmn adhoc authentication failure %d\n",s->mn_adhoc_authbad);
	printf("\tmn adhoc IPSEC route-key missing %d\n",s->mn_adhoc_noipsec);
	printf("agents lack authentication/ignored - may use arp?:\n");
	printf("\tha lacks authentication %d\n",s->ha_maybe_arp_lacks_auth);
	printf("\tfa lacks authentication %d\n",s->fa_maybe_arp_lacks_auth);
	printf("\tmn lacks authentication %d\n",s->mn_maybe_arp_lacks_auth);
	printf("\tpossible denial of service attacks due to non-increasing timestamp %d\n",
		s->ts_auth_failure);
	printf("misc:\n");
	printf("\tuse arp, ignored authenticated mn beacons %d\n", 
		s->arp_on_mn_beacons_ignored);
	printf("\tnetauth beacons rejected %d\n", s->netauth_rejected);
	printf("\tadhoc beacons rejected %d\n", s->adhoc_rejected);
}
