
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*
 * stats associated with beacons
 */


/* statistics for receiving icmp beacon 
 * used by both mnd/mipd
 */
struct icmp_stats {
/* basic counts */
	int ha_beacons_sent;	 /* ha beacons sent */
	int fa_beacons_sent;	 /* fa beacons sent */
	int combined_beacons_sent;	/* beacons with both agent bits set */
	int mn_beacons_sent;	 /* mn beacons sent */
	int all_beacons_recv;    /* all beacons that were recv */
	int all_netauth_recv;    /* all beacons of type auth recv */
	int valid_beacons_recv;  /* beacons that were not errors */
	int faisha_recv;	 /* fa beacons treated as ha beacons */
	int bothhafa;		/* beacon with both ha/fa bits set */
/* errors */
	int tooshort; 		/* message too short */
	int malformed;		/* message is malformed */
	int msetooshort;	/* mip part of message too short */
	int invalidicmpmse;     /* mse invalid */
	int auth_nomac;		/* authenticated but lacks mac part */
/* auth errors */
	int ha_netauthbad;	/* net auth from ha failed */
	int fa_netauthbad;	/* net auth from fa failed */
	int mn_netauthbad;	/* net auth from mn failed */
	int ha_adhoc_authbad;	/* adhoc auth from ha failed */
	int fa_adhoc_authbad;	/* adhoc auth from fa failed */
	int mn_adhoc_authbad;	/* adhoc auth from mn failed */
	int mn_adhoc_noipsec;	/* no ipsec key cached for LL route to MN */
	int ts_auth_failure;    /* ts not increasing on beacon */
/* auth vs arp (non-auth) general info */
	/* we want auth, beacons lack it */
	int ha_maybe_arp_lacks_auth;  /* auth in use, beacon doesn't have it */
	int fa_maybe_arp_lacks_auth;  /* auth in use, beacon doesn't have it */
	int mn_maybe_arp_lacks_auth;  /* auth in use, beacon doesn't have it */
	/* we use arp, ignore mn beacons */
	int arp_on_mn_beacons_ignored;     /* arp on, mn beacons discarded */
	/* authenticated beacons rejected */
	int netauth_rejected;	/* mnd configured to reject netauth */
	int adhoc_rejected;	/* mnd configured to reject adhoc */
	/* totals */
	int valid_arp_beacons;  /* total accepted valid non auth beacons */
	int valid_auth_beacons;	/* total count of authenticated beacons */
};

extern struct icmp_stats icmp_stats;
