/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  keylist.h
 */

#ifndef _KEYLIST_H
#define _KEYLIST_H

#define	SPI_OUTB	0	/* outbound spi */
#define SPI_INB		1	/* inbound spi */
#define SPI_MAX 	2

/* struct keylist is overloaded between ipsec and MobileIP --
 *     ipsec only fields:	ipaddr2, transform, inuse, spi_ah[1], spi_esp[]
 *     mobileIP only fields:	key, keysize
 */
struct keylist {
	unsigned long ipaddr;	/* index -- destination address */
	unsigned long ipaddr2;	/* NOT index -- spare ip addr */
	unsigned long 	spi_ah[SPI_MAX],
			spi_esp[SPI_MAX];
	unsigned long transform;
	char *key;		/* NOTE: key & keysize are for MobileIP only! */
	int  keysize;
	int  inuse;	/* TRUE if key configured as "on"; false if "off" */
	int  timeflag;  /* true if timestamp used for nonce */
	struct keylist *next;
};

#define KEYBUCKETS  131
struct keylist *keybuckets[KEYBUCKETS];
#define KEYHASH(a) ((a) % KEYBUCKETS)	/* input IP address */

/* prototypes */
struct keylist * search_keylist(struct keylist *, unsigned long ipaddr);
struct keylist *init_keylist();
int enter_keylist(struct keylist *, unsigned long, char *, int, 
	unsigned long *, int, unsigned long, int inuse, int timeflag);
char *get_key(struct keylist *, unsigned long ipaddr, int *, unsigned long *, int *,
	unsigned long *, int *);
int modify_key_transform(struct keylist *, unsigned long ipaddr, int);
void klist_initwalk(struct keylist *);
struct keylist * klist_nextwalk();
void dump_keylist(struct keylist *);

#endif /* undef _KEYLIST_H */
