/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef MADRP
#include "sr_mad.h"

extern struct sr_config sr_config; 
/*
 * defaults for MADRP.
 * Addition the following field which is set by user to /etc/mnd.conf:
 */
void sr_initconfig(struct sr_metric * sr_metric){
sr_config.SR_VERSION		=   1;
/* lifetime of a host route */
sr_config.SR_LIFETIME		=   180;  /* used to be 600 */
sr_config.SR_POWER_STRENGTH	=   80;
/* size of multicast discovery "ring", 10 may be too big */
sr_config.SR_UDPTTL		=   10;
/* reply is unicast OR multicast, if multicast you choose
 * to waste bandwidth and maximize the odds of a message arriving
 */
sr_config.SR_UNICAST_REPLY	=   true;
/* MADRP multicast ip address
*/
sr_config.SR_MULTICAST_IP	=   inet_addr("224.0.0.11");
/* MADRP UDP port
*/
sr_config.SR_UDP_PORT		=   1066;
/* use MIP-like network authentication.  
*/
sr_config.SR_NETAUTH		=   true;
sr_config.SR_MAC		=   false;
sr_config.SR_AGENT_REPLY_ACCEPTABLE  =  true;
sr_config.SR_POWER_STRENGTH_CONSTANT =	3;
sr_config.SR_SIGNAL_STRENGTH_CONSTANT= 	5;
sr_config.SR_HOP_COUNT_CONSTANT	  =	2;
sr_config.SR_IPSECAH		  =	false;

/* if it is set, then we will throw the packets of which the tuples contains
 * neighbours, except last tuple.
 * the opposite is metric-preference, which never throw these packets away,
 * and the host route is always purely decided by metrics.
 */
sr_config.SR_NEIGHBOUR_PREFERENCE = true;
/* when this is set, if we receive a request packet that contains 0 hops for
 * me, we would first push beacon then send a reply packet
 */
sr_config.SR_PUSHBEACON = true;
bzero(sr_metric, sizeof(sr_metric));
}

int sr_version(){
	return sr_config.SR_VERSION;
}
 
int sr_lifetime(){
	return sr_config.SR_LIFETIME;
}
int sr_udpttl(){
	return sr_config.SR_UDPTTL;
}
int sr_unicastreply(){
	return sr_config.SR_UNICAST_REPLY;
}
u_long sr_multicastip(){
	return sr_config.SR_MULTICAST_IP;
}
int sr_udpport(){
	return sr_config.SR_UDP_PORT;
}
int sr_netauth(){
	return sr_config.SR_NETAUTH;
}
int sr_mac(){
	return sr_config.SR_MAC;
}

int sr_agentreplyacceptable(){
	return sr_config.SR_AGENT_REPLY_ACCEPTABLE;
}
int sr_ipsecah(){
	return sr_config.SR_IPSECAH;
}
int sr_if_prefer_neighbour(){
	return sr_config.SR_NEIGHBOUR_PREFERENCE;
}
int sr_pushbeacon(){
	return sr_config.SR_PUSHBEACON;
}
#endif
