
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef MADRP
/*
 * srhlist.c
 *
 * srh list module.  Implement a linked list of srh
 * entries to be used by hosts.  The list head
 * is maintained by the caller.
 *
 * Module should hide how list is constructed from callers.
 * XXX. Current search mechanism is simple linear search.  Should
 * eventually be replaced with hash search on ip address.
 *
 * XXX. more important, sort entries according to delta lifetimes
 * ala callout list.
 */

#include "sr_mad.h"
struct SR_hlist  *srhlist;
int srhcount=0;
extern int debug; 
#ifdef MND
#include "alist.h"
extern int srhagentcount;
extern struct alist * sr_hagentlist;
#endif

/* create an entry and insert the node
 *
 * returns:
 *	struct *srhlist - pointer to srhlist entry if found
 *	NULL - out of buffer space
*/
struct SR_hlist *
srhlist_create()
{
	struct SR_hlist *n = NULL;
	struct SR_hlist *v = malloc(sizeof(struct SR_hlist));
	if (!v) {
		return(NULL);
	}
	bzero(v, sizeof(struct SR_hlist));

	/* for now just put at head of list
	*/
	if (srhlist)
		n = srhlist;
	srhlist = v;
	v->next = n;
	srhcount++;
		
	return(v);
}
/* 
 * search srhlist specified to see if mn is in list.
 *    
 * returns: 
 *      struct vlist * - pointer to vlist entry if found
 *      NULL - no such entry.
 */
       
struct SR_hlist *
srhlist_search( struct in_addr mn){
	struct SR_hlist * head=srhlist;
	while(head) {
		if (((unsigned long)head->destHost.s_addr)
		       == ((unsigned long)mn.s_addr))
		       return(head);
	  	head = head->next;
  	}
	return NULL;
}

void tick_clear_hlist(){

	struct SR_hlist *ss, * prev=NULL;
#ifdef MND
	struct alist * aa, * aprev=NULL;
#endif

	ss=srhlist;
	while (ss !=NULL){
	        if (--ss->lifetime <=0) {
#ifdef DEBUG
			if(debug)
			printf("===hlist lifetime ->0, delete host route %s\n",
				inet_ntoa(ss->destHost));
#endif
			route_delete_host(ss->destHost, ss->nextHop);
/*#ifdef MADRP
			vlist_delete(&vlist_gethahead(),ss->destHost);
			vlist_delete(&vlist_gethahead(),ss->destHost);
#endif*/
			if(prev==NULL) {
				srhlist=ss->next;
				free(ss);
				ss=srhlist;
				srhcount--;
			} else {
				prev->next=ss->next;
				free(ss);
				ss=prev->next;
				srhcount--;
			}
		} else {
			prev=ss;
			ss=ss->next;
		}
	}
#ifdef MND   /* tickclear sr_hagentlist */
	aprev=NULL;
	aa=sr_hagentlist;
	while(aa!=NULL) {
		if (--aa->lifetime<=0){
			if(aprev==NULL) {
				sr_hagentlist=aa->next;
				free(aa);
				aa=sr_hagentlist;
				srhagentcount--;
			} else {
				aprev->next=aa->next;
				free(aa);
				aa=aprev->next;
				srhagentcount--;
			}
		} else {
			aprev=aa;
			aa=aa->next;
		}
	}
#endif

}

void hlist_delete_all(){
	struct SR_hlist *ss;
#ifdef MND
	struct alist * aa;
#endif
	ss=srhlist;
	while (ss !=NULL){
		route_delete_host(ss->destHost, ss->nextHop);
		srhlist=ss->next;
		free(ss);
		ss=srhlist;
		srhcount--;
	}
#ifdef MND
	aa=sr_hagentlist;
	while (aa !=NULL){
		sr_hagentlist=aa->next;
		free(aa);
		aa=sr_hagentlist;
		srhagentcount--;
	}
#endif
}


void srhlist_delete(struct in_addr dest) {
	struct SR_hlist *ss, * prev=NULL;
	ss=srhlist;
	while (ss !=NULL){
		if(ss->destHost.s_addr==dest.s_addr){
			if(prev==NULL) {
				srhlist=ss->next;
				free(ss);
				ss=srhlist;
				srhcount--;
			} else {
				prev->next=ss->next;
				free(ss);
				ss=prev->next;
				srhcount--;
			}
			break;
		} else {
			prev=ss;
			ss=ss->next;
		}
	}
}

static struct SR_hlist *whead;

void
srhlist_initwalk()
{
	whead = srhlist;
}

struct SR_hlist *
srhlist_nextwalk()
{
	struct SR_hlist *cur = whead;
	if (whead == NULL)
		return(NULL);
	whead = whead->next;
	return(cur);
}

int srhlist_count(){
	return srhcount;
}

void srhlist_dump()
{
	struct SR_hlist *head;
	int i = 0;
	srhlist_header();
	srhlist_initwalk();
	while(1) {
		head = srhlist_nextwalk();
		if (!head)
			return;
		srhlist_dumpone(head);
	}
}

#ifdef TESTSRSENT
/*
 * test all primitives above
 */
main()
{
	struct SR_hlist *myhead = NULL;
	struct SR_hlist *myhead2 = NULL;
	struct in_addr mn;
	struct SR_hlist *v;

	/* test case 1.  put 5 entries in list and dump
	*/
	mn.s_addr = 0x01010101;
	srhlist_create(&myhead, mn);
	mn.s_addr = 0x02020202;
	srhlist_create(&myhead, mn);
	mn.s_addr = 0x03030303;
	srhlist_create(&myhead, mn);
	mn.s_addr = 0x04040404;
	srhlist_create(&myhead, mn);
	mn.s_addr = 0x05050505;
	srhlist_create(&myhead, mn);

	srhlist_dump(myhead); 

	/* now find all the list entries
	 * but first search for one that is not there.
	 */
	mn.s_addr = 0x01020304;
	v = srhlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, correct!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x01010101;
	v = srhlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x05050505;
	v = srhlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x03030303;
	v = srhlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	/* test series 2.  put 1 entry in list and delete it
	*/

	mn.s_addr = 0x01010101;
	srhlist_create(&myhead2, mn);
	srhlist_delete(&myhead2, mn);
	srhlist_dump(myhead2); 
	v = srhlist_search(myhead2, mn);
	if (v == NULL) 
		printf("correct - didn't find it\n");
		
	/* test series 3.  put 2 entries in list and delete first
	*/
	mn.s_addr = 0x02020202;
	srhlist_create(&myhead2, mn);
	mn.s_addr = 0x01010101;
	srhlist_create(&myhead2, mn);
	srhlist_delete(&myhead2, mn);
	printf("list should only have 02020202 in it\n");
	srhlist_dump(myhead2); 
	mn.s_addr = 0x02020202;
	v = srhlist_search(myhead2, mn);
	if (v == NULL) 
		printf("incorrect - didn't find it\n");
	else
		printf("found %x\n", v->mn.s_addr);

	/* test series 4. delete 3rd entry in original list
	*/
	mn.s_addr = 0x03030303;
	srhlist_delete(&myhead, mn);
	printf("following lisit should have 5/4/2/1 in that order\n");
	srhlist_dump(myhead); 

	/* test series 5. delete 1st and last entries in list
	*/
	mn.s_addr = 0x01010101;
	srhlist_delete(&myhead, mn);
	mn.s_addr = 0x05050505;
	srhlist_delete(&myhead, mn);
	printf("following list should only have 4/2 in it\n");
	srhlist_dump(myhead); 

	/* final test. put entry back in and print out.
	 * list should have 3/4/2 in that order
	 */
	mn.s_addr = 0x03030303;
	srhlist_create(&myhead, mn);
	printf("following list should have 3/4/2 in that order\n");
	srhlist_dump(myhead); 
}
#endif

#ifdef TESTDUMP
main()
{
	struct SR_hlist *head = NULL;
	struct SR_hlist *v;
	struct in_addr mn;
	struct in_addr coa;

	mn.s_addr = 0xc1d1e1f1;
	v = srhlist_create(&head, mn);
	v->vstate = 0;
	strncpy(v->location, "my pc", 5);
	v->lsize = 5;
	v->lifetime = 1;
	v->coa.s_addr = 0x01020304;
	v->auth_type = 0;

	mn.s_addr = 0x02020202;
	v = srhlist_create(&head, mn);
	v->vstate = 1;
	strncpy(v->location, "your pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccd02ce;
	v->auth_type = 0x1;
	v->lifetime = 2;

	mn.s_addr = 0x03030303;
	v = srhlist_create(&head, mn);
	v->vstate = 2;
	strncpy(v->location, "bobs pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xacbd02cd;
	v->auth_type = 0x2;
	v->lifetime = 3;

	mn.s_addr = 0x04040404;
	v = srhlist_create(&head, mn);
	v->vstate = 3;
	strncpy(v->location, "tena pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccccc12;
	v->auth_type = 0x3;
	v->lifetime = 4;

	mn.s_addr = 0x05050505;
	v = srhlist_create(&head, mn);
	v->vstate = 4;
	strncpy(v->location, "xyxz pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccccc11;
	v->auth_type = 0x0;
	v->lifetime = 5;

	srhlist_shortHeader();
	srhlist_initwalk(head);
	while(1) {
		head = srhlist_nextwalk();
		if (!head)
			return;
		srhlist_dumpone(head, 0);
	}
}
#endif
#endif
