/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef MADRP
#include "sr_mad.h"
#include "mip_log.h"
extern int debug; 

u_int 	init_sr_Udp();
void 	Send_SR_request(struct in_addr , u_int, struct in_addr );
void 	Recv_SR_request(char *, struct in_addr, struct sr_metric *);
void 	Send_SR_reply(int,int,char *, struct in_addr);
void 	Recv_SR_reply(char *,  struct in_addr, struct sr_metric *);
void 	handleMadInput( int, struct sr_metric *) ;
void    dumplists();
void 	tickBackoffTimer();

extern int madSocket;
extern struct sr_config  sr_config;
#ifdef MND
extern struct alist * sr_hagentlist;
#endif

void handleMadInput(int madSocket, struct sr_metric *sr_metric) 
{ 
    struct sr_request_packet * request;
    struct sr_reply_packet * reply;
    u_char packet[2048];
    struct sockaddr_in sin;
    int len;
    int fromlen;
    struct in_addr myad;
    int ifindex=0;


    fromlen=sizeof(struct sockaddr_in);
    len = recvMsg( madSocket,packet,2048, (struct sockaddr *)&sin, &fromlen,
		   & ifindex);
    if(len<0) {
	syslog( LOG_ERR, "handleMadInput(): recvfrom failed (%m).");
	exit(1);
    }
    myad.s_addr=indextoif(ifindex);
    request=(struct sr_request_packet *)packet;
    if(request->version==1){
	if(request->type==SRREQUEST){
		if(len<sizeof(struct sr_request_packet)){
			sr_metric->size++;	
			return;
		}

       		request->lifetime= ntohs(request->lifetime);
		request->ntp_sec=ntohl(request->ntp_sec);
		request->ntp_fra=ntohl(request->ntp_fra);
		request->udp_ttl=ntohs(request->udp_ttl);
		request->flag=ntohs(request->flag); 
		request->tuplesize=ntohs(request->tuplesize);
		request->tuplecount=ntohs(request->tuplecount);
#ifdef DEBUG
		if(debug){
		printf("recv request for %s ", inet_ntoa(request->desti_addr));
		printf("by %s ", inet_ntoa(request->sender_addr));
		printf("through %s\n", inet_ntoa(sin.sin_addr));
		}
#endif
		Recv_SR_request(packet,myad, sr_metric);
    	}
    	else if (request->type==SRREPLY){
		if(len<sizeof(struct sr_reply_packet)){
			sr_metric->size++;
			return;
		}

		reply=(struct sr_reply_packet *)packet;
       	 	reply->lifetime=ntohs(reply->lifetime); 
		reply->ntp_sec=ntohl(reply->ntp_sec);
		reply->ntp_fra=ntohl(reply->ntp_fra);
     		reply->udp_ttl=ntohs(reply->udp_ttl);
     		reply->flag=ntohs(reply->flag);
     		reply->tuplesize=ntohs(reply->tuplesize);
     		reply->tuplecount=ntohs(reply->tuplecount);
		reply->current_tuple_pointer=ntohs(reply->current_tuple_pointer);

#ifdef DEBUG
		if(debug) {
		printf("recv reply dst:%s ", inet_ntoa(reply->desti_addr));
		printf("initiator: %s ", inet_ntoa(reply->src_addr));
		printf("through %s\n", inet_ntoa(sin.sin_addr));
		}
#endif
		Recv_SR_reply(packet, myad, sr_metric);
	}
    }
}

#ifdef MND
void mn_tickBackoffTimer(){
	int i,j;
	struct in_addr tosend[10], myad;
	myad.s_addr=getmyaddr();
	j=tick_clear_sentlist(tosend);
	for(i=0;i<j;i++){
#ifdef DEBUG
		if(debug)
		printf("Send request again to %s\n", inet_ntoa(tosend[i]));
#endif
		Send_SR_request(tosend[i],FALSE, myad);  
	}
	tick_clear_recvlist();
	tick_clear_hlist();
	return;

}
#else
void mip_tickBackoffTimer(){
	tick_clear_recvlist();
	tick_clear_hlist();
	return;

}
#endif

/* mnstat TBD
 */
void dumplists(){
#ifdef MND
	srsentlist_dump();
	alist_dump(sr_hagentlist);
#endif
	srrecvlist_dump();
	srhlist_dump();
}

u_int init_sr_Udp()
{
	u_int s;
	struct sockaddr_in saddr;
	struct ip_mreq mreq;
	u_char ttl;
	u_char loop;
	int on=1;

	/* create INTERNET, udp datagram socket
	 */
	s = socket(AF_INET, SOCK_DGRAM, 0); 
	if ( s < 0 ) {
		perror("initUdpServer: socket");
		exit(1);
	}
	
	/* TBD : IP_RECV_INTERFACE */


	/* join the multicast group */
	mreq.imr_multiaddr.s_addr = sr_multicastip(); 
	mreq.imr_interface.s_addr = INADDR_ANY;
	if (setsockopt(s, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq,
			sizeof(mreq)) < 0) {       
		perror("joinGroup(): join IP_ADD_MEMBERSHIP .");
		exit(1);
	}

       	ttl= 1;
       	if (setsockopt(s, IPPROTO_IP, IP_MULTICAST_TTL,
	       &ttl, sizeof(ttl)) < 0) {
             perror("setsockopt(): set IP_MULTICAST_TTL.");
	     exit(1);
	}

	loop=0; /* 0: disable 1:enable */
       	if (setsockopt(s, IPPROTO_IP, IP_MULTICAST_LOOP,
	       &loop, sizeof(loop)) < 0) {
             perror("setsockopt(): set IP_MULTICAST_LOOP.");
	     exit(1);
	}

        /* configure socket to sendup i/f index so we can determine
	 * interface packet came in on.  This calls requires us
	 * to use recvMsg(2)
	 */
	if(setsockopt(s, IPPROTO_IP, IP_RECVIF, (char *)&on, sizeof(on))<0){
             perror("setsockopt(): set IP_RECVIF.");
	     exit(1);
	}

	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr = INADDR_ANY;   
	saddr.sin_port = htons(sr_udpport());     

	/* bind protocol to socket
	*/
	if (bind(s, (struct sockaddr *)&saddr, sizeof(saddr))) {
		perror("bind");
		exit(1);
	}
	return(s);
}
#endif

