/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _SR_MAD_H
#define _SR_MAD_H
 
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef TESTMAIN
#include "steal/mip_icmp.h"
#endif
#ifdef IPSEC
#include "route.h"
#endif

#ifndef MIN
#define MIN(x,y) ((x>y) ? y:x)
#endif

#define TUPLE_SIZE  8
#define MAX_INDEX 13
#define TRUE 1
#define FALSE 0
#define true 1
#define false 0

#define AGENT_REPLY_ACCEPTABLE_BIT 0x80
#define MULTICAST_BIT              0x40
#define UNICAST_BIT                0x20
#define AGENT_BIT                  0x10
#define DESTINATION_BIT            0x08

#define SRREQUEST    		1
#define SRREPLY 		2
#define UNICAST_REPLY 		0
#define MULTICAST_REPLY 	1
#define AM_MN 			1
#define AM_AGENT 		0
#define CURRENT_SR_VERSION 	1

#define MIPAGENTAUTH_OFFSET     20
#ifndef MD5CSUMSIZE
	#define MD5CSUMSIZE 	16
#endif

struct sr_reply_packet {
	u_char type ;
	u_char version ;
	u_short lifetime;
	u_long ntp_sec;  
	u_long ntp_fra; 
	struct in_addr replier_addr;
	struct in_addr src_addr;
	struct in_addr desti_addr;
	struct in_addr coa;   
	u_short udp_ttl;
	u_short pad ;
	u_short flag;	
	u_short tuplesize; 
	u_short tuplecount;
	u_short current_tuple_pointer;
};

struct sr_request_packet {
	u_char type;
	u_char version;
	u_short  lifetime;
	u_long ntp_sec;
	u_long ntp_fra;
	struct in_addr sender_addr;
 	struct in_addr desti_addr;
	u_short udp_ttl;
	u_short flag;
	u_short tuplesize;
	u_short tuplecount;
};

struct pathtuple{
	struct in_addr ipaddr;
	u_char signalstrength;
	u_char powerstrength;
	u_char tbd1;
	u_char tbd2;
};


struct mip_mse_udp_auth {
#define MIPMSEUDPAUTH 2
	u_char mse_type; 	/* 2 for mip_mse_udp_auth */
	u_char mse_length;                  /* length of data + spi */
/* no flag here, 'cause we will only use one key: SRAUTH */
	u_char pad[2];
	u_long  tssecs;         /* ntp secs timestamp */
	u_long  tsfract;        /* ts fractional part */
	u_long  agent;          /* ip addr for mn of current agent */
	/* how far checksum is calculated */
	unsigned long spi;
	u_char hash[MD5CSUMSIZE];        /* enough room for the hash */
};


/*1.2 Tables maintained at each host
 */
/* cache of SR request/reply sent
 */
struct SR_sent {
	u_long requestId_secs;  /* NTP seconds timestamp */
	u_long requestId_fract; /* NTP fractional seconds timestamp */
	struct in_addr ip_target; 
	u_char replied_flag;
	u_int backoff_count;
	u_int backoff_timer;
	struct SR_sent * next;
};

/* cache of request recv 
 */
struct SR_recv {
	u_char type; /* packet received type: SR request or reply */
	u_long requestId_secs;  /* NTP seconds timestamp */
	u_long requestId_fract; /* NTP fractional seconds timestamp */
	struct in_addr ip_initiator;
	struct in_addr ip_target;
	struct in_addr ip_replier;
	u_short orig_lifetime; /* the lifetime that won't decrement */
	u_short lifetime; /* will decrement this every click */
	u_short tuplecount; 
	struct pathtuple *tuple;
	struct SR_recv * next;
};

/* shadow routing table
 */
struct SR_hlist {
	u_short lifetime;  /* will decrement it every click */
	struct in_addr destHost;
	struct in_addr nextHop;
	int metric;
	int isAgent;       /* if the next hop is an agent , will always be
			    * overrided by a host nexthop */
#ifdef IPSEC
	struct routeSA  rosa;
#endif
	struct SR_hlist * next;
};

struct sr_config {
	int SR_VERSION;
	int SR_LIFETIME;
	int SR_POWER_STRENGTH;
	int SR_UDPTTL;
	int SR_UNICAST_REPLY;
	u_long SR_MULTICAST_IP;
	int SR_UDP_PORT;
	int SR_NETAUTH;
	int SR_MAC;
	int SR_AGENT_REPLY_ACCEPTABLE;
	int SR_POWER_STRENGTH_CONSTANT;
	int SR_SIGNAL_STRENGTH_CONSTANT;
	int SR_HOP_COUNT_CONSTANT;
	int SR_IPSECAH;

	/* if it is set, then we will throw the packets of which the tuples
	 * contains neighbours, except last tuple. 
	 * the opposite is metric-preference, which never throw these
	 * packets away, and the host route is always purely decided by metrics.
	 */
	int SR_NEIGHBOUR_PREFERENCE;

	/* when this is set, if we receive a request packet that contains 0
	 * hops for me, we would first push beacon then send a reply packet
	 */
	int SR_PUSHBEACON;
};

/* source routing error metrics
*/
struct sr_metric {
	int security;    /* not pass security check */
	int ttl;         /* ttl decrement to zero */
	int loop;        /* have loop in pathtuples */
	int samerequest; /* have received the same request ID 
			  * (not for me before      */
	int badmetric;   /* have received the same reuqest for me but this
			  * one's metric is not better than the old one 
			  */
 	int samedesti;   /* have received requests for the same destination, but
			  * this is an earlier one 
			  */
	int myself;      /* a reply/request sent by me */
	int size;        /* packet size too small */
        int linkloop;  	 /* have loop among neighbours (link) */
};

#ifndef DEBUG
	#define SENDTO sendto
	#define RECVFROM recvfrom
#endif
#endif /* undef _SR_MAD_H */
