/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef MADRP
#include "sr_mad.h"
extern struct sr_config sr_config;

/*
 * cumulative metrics calculation subroutine

 * input: 
	tuple, a pointer to a tuple, offset of the tuples in the 
	       receiving packet;
	tuple_count, tuple count starting from tuple pointer to the end;

 * output:
	the cumulative path metrics from the (candidate) destination
	host whose IP address is designated in the first tuple to the 
	current host;

 * Note:

 * 1) the host designated by the first tuple is not necessarily the
 * SR request initiator, it may be the intermediate system;

 *  the reason is for one incoming SR packet, we may compute the 
 * cumulative metrics starting from host, maybe from the source, 
 * maybe from the IS;

 * 2) likewise, the tuple_count value here may or may not be the 
 * value in SR request packet;

 * 3) int hop_count_conversion(int):

 * this function used below is a mapping from hop count value
 * to 0..100 range value; the trick is not how to convert (a mapping
 * function is the homework of cs161, so I do not bother to fill in
 * this function here), but how to define the mapping; see Metrics 
 * session for our first-cut mapping sketch;
 */


int cumulativePathMetrics (struct pathtuple *tuple, int tuple_count) {

	struct pathtuple * t;
	int ps_total=0, i, ss_total=0, PV, SV, HV, metric;
	
	/* calculate the average power strength */
	t= tuple;
	for (i=0; i<tuple_count; i++) {
	   ps_total += t[i].powerstrength;
	}
	ps_total += my_power_strength(); /* make sense ? */
	PV = ps_total / (tuple_count + 1);

	/* calculate the average signal strength */
	for (i=0; i<tuple_count; i++) {
	   ss_total += t[i].signalstrength;
	}
	ss_total += WIRESIGFUNCTION(t[tuple_count-1].ipaddr);
	SV = ss_total / (tuple_count + 1);

	/* calculate the 0..100 range value converted from hop count */
	HV = hop_count_conversion (tuple_count); /* see note above;*/

	/* calculate the weighted metrics 
	 *
	 * range:
	 * PV: 0..100, 100 for the best power strenth;
	 * PS(): 0..10, weight of PV among the three metrics; 
	 *      (not 0..1, to avoid FP calculation)
	 * SV: 0..100, 100 for the strongest signal strength;
	 * SS(): 0..10, weight of SV;
	 * HV: 0..100, 100 for shortest hop count (no hop?)
	 * HS(): 0..10, weight of HV;
	 */
	metric = PV * PS() + SV * SS() + HV * HS(); 
	return (metric);
}

/* just in case we need to compare the metric of SR with alists , not used rightnow
 */
int Alist_pathMetrics (struct in_addr neighbour) {

	int ps_total=0, i, ss_total=0, PV, SV, HV, metric;
	
	PV = my_power_strength(); /* make sense ? */
	SV = WIRESIGFUNCTION(neighbour);
	HV = 120;
	metric = PV * PS() + SV * SS() + HV * HS(); 
	return (metric);
}

int hop_count_conversion(int in){
	/* if no hop (just neighbour , I assume HV=120 */
	int conv_table[7]= { 100,90,80,60,40,20,20};
	if(in>7) return 0;
	else return (conv_table[in-1]);
}

int PS() {
  return (sr_config.SR_POWER_STRENGTH_CONSTANT);
}

int SS() {
  return (sr_config.SR_SIGNAL_STRENGTH_CONSTANT);
}

int HS() {
  return (sr_config.SR_HOP_COUNT_CONSTANT); 
}

int my_power_strength () {
  return (sr_config.SR_POWER_STRENGTH);
}

#endif
