/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef MADRP
#include "sr_mad.h"

extern struct sr_config sr_config;
extern int debug; 

void *  addnewentry(void * , int * , u_int);
int 	addudpauth(char * ,int);
char *	getNetauthKey();
int 	getNetauthSpi() ;

        
ssize_t SENDTO(int s, const char *packet, size_t len, int flags,
                const struct sockaddr * to, int tolen) {

	struct sr_request_packet *srq;
	struct sr_reply_packet *srp;
	struct pathtuple *tuple;
	int i,j;

	srp=(struct sr_reply_packet *)packet;
	if(srp->type==SRREPLY){
#ifdef  DEBUG
		if(debug) {
		printf("send a SR_reply packet to '%s'\n", 
			inet_ntoa(((struct sockaddr_in *) to)->sin_addr));
		printf("lifetime: %u\trequestId: %u+%u\tudp_ttl: %d\n",
			srp->lifetime,
			srp->ntp_sec, 
			srp->ntp_fra,srp->udp_ttl);
		printf("replier: %s\t", inet_ntoa(srp->replier_addr));
		printf("src: %s\t", inet_ntoa(srp->src_addr));
                printf("desti: %s\n",inet_ntoa(srp->desti_addr));
		printf("flag: %x\ttuplecount: %d\tcurrenttuple: %d\n",
		    	srp->flag, 
			srp->tuplecount,
			srp->current_tuple_pointer);
		if(srp->tuplecount>0) {
			tuple=(struct pathtuple *) 
				(packet+sizeof(struct sr_reply_packet));
			printf("tuples are: \n");
			for (j=0;j<srp->tuplecount;j++)
				printf("\t%s\n", inet_ntoa(tuple[j].ipaddr));
		}
		}
#endif
		srp->lifetime=htons(srp->lifetime);    
		srp->ntp_sec=htonl(srp->ntp_sec);
		srp->ntp_fra=htonl(srp->ntp_fra);
		srp->udp_ttl=htons(srp->udp_ttl);
		srp->flag=htons(srp->flag);
		srp->tuplesize=htons(srp->tuplesize);
		srp->tuplecount=htons(srp->tuplecount);
		srp->current_tuple_pointer=htons(srp->current_tuple_pointer);
		return(sendto(s,packet,len,flags,to,tolen)); 
	}
	else {
		srq=(struct sr_request_packet *)packet;
#ifdef DEBUG 
		if(debug){
		printf("send a SR_request packet to '%s'\n", 
			inet_ntoa(((struct sockaddr_in *) to)->sin_addr));
		printf("lifetime: %u\trequestId: %u+%u\tudp_ttl: %d\n",
			srq->lifetime,srq->ntp_sec, 
			srq->ntp_fra,srq->udp_ttl);
		printf("sender: %s\t", inet_ntoa(srq->sender_addr));
                printf("desti: %s\n",inet_ntoa(srq->desti_addr));
		printf("flag: %x\t\ttuplecount: %d\n",
		    	srq->flag, srq->tuplecount);
		if(srq->tuplecount>0) {
			tuple=(struct pathtuple *) 
				(packet+sizeof(struct sr_request_packet));
			printf("tuples are: \n");
			for (j=0;j<srq->tuplecount;j++)
				printf("\t%s\n", inet_ntoa(tuple[j].ipaddr));
		}
		}
#endif
		srq->lifetime= htons(srq->lifetime);
		srq->ntp_sec=htonl(srq->ntp_sec);
		srq->ntp_fra=htonl(srq->ntp_fra);
		srq->udp_ttl=htons(srq->udp_ttl);
		srq->flag=htons(srq->flag);
		srq->tuplesize=htons(srq->tuplesize);
		srq->tuplecount=htons(srq->tuplecount);
		return(sendto(s,packet,len,flags,to,tolen)); 
	}
	return len;
}

#ifdef SAVEENERGY

extern u_long changestate;
#define R(x,Y1,Y2,Y3) {x[0]=Y1;x[1]=Y2;x[2]=Y3;}

ssize_t RECVFROM(int s, void *buf, size_t len, int flags, 
		struct sockaddr *from, int *fromlen){
	ssize_t lenn;
	u_long A,B,C,D,N,N1,N2;
	u_long a[3],b[3],c[3],d[3];
	lenn=recvfrom(s,buf,len,flags,from,fromlen);
	A=inet_addr("204.203.67.205");
	B=inet_addr("204.203.67.206");
	C=inet_addr("204.203.67.210");
	D=inet_addr("10.0.0.88");
	N=inet_addr("1.1.1.1");
	N1=inet_addr("204.203.67.200");
	N2=inet_addr("204.203.67.115");
	switch(changestate) {
		default:
			printf("####### A-B-C-D #######");
			R(a,B,N,N); R(b,A,C,N); R(c,B,D,N); R(d,C,N1,N2);
			break;
		/*
		case 0: 
			printf("####### A/B-C-D #######");
			R(a,B,C,N); R(b,A,C,N); R(c,B,D,A); R(d,C,N,N);
			break;
		case 2:
			printf("####### A-B/C-D #######");
			R(a,B,C,N); R(b,A,C,D); R(c,B,D,A); R(d,C,B,N);
			break;
		*/
	}

#ifdef MND
	if (getmyaddr()==B){
		if (((struct sockaddr_in *)from)->sin_addr.s_addr == b[0] 
		 || ((struct sockaddr_in *)from)->sin_addr.s_addr == b[1]
		 || ((struct sockaddr_in *)from)->sin_addr.s_addr == B
		 || ((struct sockaddr_in *)from)->sin_addr.s_addr == b[2]){
			printf("recvmsg from %s \n",
			inet_ntoa(((struct sockaddr_in *)from)->sin_addr));
			return lenn;
		}
		else if (((struct sockaddr_in *)from)->sin_port==45569) {
			printf("==========recv registration reply \n");
			return lenn;
		}
	} else if (getmyaddr()== C){
		if (((struct sockaddr_in *)from)->sin_addr.s_addr == c[0] 
		 || ((struct sockaddr_in *)from)->sin_addr.s_addr == c[1]
		 || ((struct sockaddr_in *)from)->sin_addr.s_addr == C
		 || ((struct sockaddr_in *)from)->sin_addr.s_addr == c[2]){
			printf("recvmsg from %s \n", 
			inet_ntoa(((struct sockaddr_in *)from)->sin_addr));
			return lenn;
		}
		else if (((struct sockaddr_in *)from)->sin_port==45569) {
			printf("==========recv registration reply \n");
			return lenn;
		}
	} else if (getmyaddr()== A){
		if (((struct sockaddr_in *)from)->sin_addr.s_addr == a[0] 
		 || ((struct sockaddr_in *)from)->sin_addr.s_addr == a[1]
		 || ((struct sockaddr_in *)from)->sin_addr.s_addr == A
		 || ((struct sockaddr_in *)from)->sin_addr.s_addr == a[2]){
			printf("recvmsg from %s \n", 
			inet_ntoa(((struct sockaddr_in *)from)->sin_addr));
			return lenn;
		}
		else if (((struct sockaddr_in *)from)->sin_port==45569) {
			printf("==========recv registration reply \n");
			return lenn;
		}
	}
#else
	if (((struct sockaddr_in *)from)->sin_addr.s_addr == d[0] 
	 || ((struct sockaddr_in *)from)->sin_addr.s_addr == d[1]
	 || ((struct sockaddr_in *)from)->sin_addr.s_addr == D
	 || ((struct sockaddr_in *)from)->sin_addr.s_addr == d[2]){
		printf("recvmsg from %s \n", 
			inet_ntoa(((struct sockaddr_in *)from)->sin_addr));
		return lenn;
	}
	else if (((struct sockaddr_in *)from)->sin_port==45569) {
		printf("==========recv registration packet from 207\n");
		return lenn;
	}
#endif
	printf("!!!!!! throw the packet away from %s.%d!!!!!!!!!!\n",
	inet_ntoa(((struct sockaddr_in *)from)->sin_addr), 
			((struct sockaddr_in *)from)->sin_port);
	return 0;
}

ssize_t RECVMSG(int s, struct msghdr *msg, int flags){
	ssize_t lenn;
	struct sockaddr_in * from;
	u_long A,B,C,D,N; /* A-B-C-D */
	u_long a[3], b[3], c[3], d[3];
	lenn=recvmsg(s,msg,flags);
	from=(struct sockaddr_in *)msg->msg_name ;
	A=inet_addr("204.203.67.205");
	B=inet_addr("204.203.67.206");
	C=inet_addr("204.203.67.210");
	D=inet_addr("10.0.0.88");
	N=inet_addr("1.1.1.1");
	switch(changestate) {
		default:
			printf("####### A-B-C-D #######");
			R(a,B,N,N); R(b,A,C,N); R(c,B,D,N); R(d,C,N,N);
			break;
		/*
		case 0: 
			printf("####### A/B-C-D #######");
			R(a,B,C,N); R(b,A,C,N); R(c,B,D,A); R(d,C,N,N);
			break;
		case 2:
			printf("####### A-B/C-D #######");
			R(a,B,C,N); R(b,A,C,D); R(c,B,D,A); R(d,C,B,N);
			break;
		*/
	}


#ifdef MND
	if (getmyaddr()==B){
		if (from->sin_addr.s_addr == b[0] 
		 || from->sin_addr.s_addr == b[1]
		 || from->sin_addr.s_addr == B
		 || from->sin_addr.s_addr == b[2]){
			printf("recvmsg from %s \n", inet_ntoa(from->sin_addr));
			return lenn;
		}
		else if (from->sin_port==45569) {
			printf("==========recv registration reply \n");
			return lenn;
		}
	} else if (getmyaddr()== C){
		if (from->sin_addr.s_addr == c[0] 
		 || from->sin_addr.s_addr == c[1]
		 || from->sin_addr.s_addr == C
		 || from->sin_addr.s_addr == c[2]){
			printf("recvmsg from %s \n", inet_ntoa(from->sin_addr));
			return lenn;
		}
		else if (from->sin_port==45569) {
			printf("==========recv registration reply \n");
			return lenn;
		}
	} else if (getmyaddr()== A){
		if (from->sin_addr.s_addr == a[0] 
		 || from->sin_addr.s_addr == a[1]
		 || from->sin_addr.s_addr == A
		 || from->sin_addr.s_addr == a[2]){
			printf("recvmsg from %s \n", inet_ntoa(from->sin_addr));
			return lenn;
		}
		else if (from->sin_port==45569) {
			printf("==========recv registration reply \n");
			return lenn;
		}
	}
#else
	if (from->sin_addr.s_addr == d[0] 
	 || from->sin_addr.s_addr == d[1]
	 || from->sin_addr.s_addr == D
	 || from->sin_addr.s_addr == d[2]){
		printf("recvmsg from %s \n", inet_ntoa(from->sin_addr));
		return lenn;
	}
	else if (from->sin_port==45569) {
		printf("==========recv registration packet from 207\n");
		return lenn;
	}
#endif
	printf("!!!!!! throw the packet away from %s.%d!!!!!!!!!!\n",
	inet_ntoa(from->sin_addr), from->sin_port);
	return 0;
}
#else

ssize_t RECVMSG(int s, struct msghdr *msg, int flags){
	return recvmsg(s,msg,flags);
}

ssize_t RECVFROM(int s, void *buf, size_t len, int flags, 
		struct sockaddr *from, int *fromlen){
	return recvfrom(s,buf,len,flags,from,fromlen);
}

#endif
#endif

