/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef MADRP
#include "sr_mad.h"

struct in_addr defaultAgent;
int defaultagentMetrics=0;
extern int debug; 

extern u_int            madSocket;

void Unicast_Reply_Update ( struct pathtuple *tuple, int tupleCount, 
			    int replybyagent, int current_tuple, int lifetime, 
			    struct in_addr ip_replier, struct in_addr ip_dst, 
			    struct in_addr coa,struct in_addr ifa);
void Mcast_Reply_Update ( struct pathtuple *tuple, int tupleCount, int lifetime,
			int replybyagent, struct in_addr ip_replier,
			struct in_addr ip_dst,struct in_addr coa,
			struct in_addr ifa);
int reply_flooding_filter (char *, struct SR_recv **, struct in_addr, struct
sr_metric *)  ;
void reply_multicastForward(char *, struct in_addr ) ;
void    SR_Update_hlist (struct pathtuple * , int,int,int, struct in_addr,int,
                         struct in_addr);
void dumpsr_metric(struct sr_metric sr_metric);
struct SR_recv * srrecvlist_create();
struct SR_recv * srrecvlist_nextwalk();
#ifdef MND
#include "alist.h"
struct SR_sent * srsentlist_nextwalk();
struct SR_sent * srsentlist_search(struct in_addr);
struct SR_hlist * srhlist_search(struct in_addr);
struct alist * alist_getmnhead();
struct alist * sr_hagentlist=NULL;
int srhagentcount=0;
#else
#include "mipconfig.h"
#include "vlist.h"
struct vlist * vlist_gethahead();
struct vlist * vlist_getfahead();

struct vlist * vlist_search(struct vlist *, struct in_addr);
#endif
/*
 * 6.1 receive the SR_reply

 * NO sr_reply_recv table maintained in unicast mode.
 * sr_reply_recv table maintained in multicast mode.

 * input: the incoming SR reply packet;

 * ip_dst as the parameter to this function;
 */
void Recv_SR_reply(char *replyPacket, struct in_addr ifaddr, struct sr_metric *
sr_metric) {

  struct sr_reply_packet * reply;
  struct pathtuple * tuples;
  struct sockaddr_in to;
  struct SR_sent * ss;
  struct SR_recv * sr;
  int i,j,rc,offset,replybyagent;

  reply = (struct sr_reply_packet *) replyPacket;
  tuples=(struct pathtuple *) (replyPacket
			       + sizeof(struct sr_reply_packet));

  if (reply->flag & AGENT_BIT) {
      	replybyagent=TRUE;
  } else {
  	replybyagent=FALSE;
  }

  /* Unicast reply mode */
  if(reply->flag & UNICAST_BIT) {
      /* 
       * TBD: do we need filter here, probably not.
       * I try to justify that filter is not necessary here:
       * 
       * packet from self: impossible in unicast;
       * ttl: no ttl at all;
       * loop check: impossible;
       * same packet received via different: impossible;
       * replier reply twice or more: impossible;
       */

      /* update the hlist and routing table */
      Unicast_Reply_Update (tuples, reply->tuplecount,replybyagent ,
	  reply->current_tuple_pointer, reply->lifetime,reply->replier_addr,
	  reply->desti_addr, reply->coa,ifaddr);

      /* I am the desitnation of the srreply */
      if (reply->src_addr.s_addr== ifaddr.s_addr) {
#ifdef MND
         /* the replied_flag is in SR_Sent table, set it so that 
          * there won't be SR_Request again
          */
	 ss=srsentlist_search(reply->desti_addr);
	 if(ss)  ss->replied_flag=TRUE;
#endif
      }
      /* not yet the destination */ 
      else {
	 reply->lifetime=MIN(reply->lifetime, sr_lifetime());
	 reply->current_tuple_pointer++;
	 offset= sizeof(struct sr_reply_packet)+ 
		 reply->tuplecount * sizeof(struct pathtuple);
         if (sr_netauth()) {
		offset+= addudpauth(reply, offset);
	 }

	 to.sin_family = AF_INET; 
	 if ( reply->current_tuple_pointer== reply->tuplecount) 
		to.sin_addr.s_addr=reply->src_addr.s_addr;
	 else
	 	to.sin_addr.s_addr= tuples[reply->current_tuple_pointer].ipaddr.s_addr;
	 to.sin_port = htons(sr_udpport());

	 rc = SENDTO(madSocket,replyPacket, offset,0,
		 (struct sockaddr *) &to, sizeof(struct sockaddr_in));
					 
	 if(rc<0){
		 perror("sendto");
		 /*
		 exit(1);
		 */
	 }  
      }
  }

  /* Multicast reply mode */
  else {

      if (reply_flooding_filter(replyPacket, &sr, ifaddr, sr_metric) == 1){
#ifdef DEBUG
	if(debug)
	 printf("It is thrown away!\n");
	 dumpsr_metric(*sr_metric);
#endif
	 return;
      }


      /* update SR_recv table */
      if ( sr==NULL) {
	      /* create an SR_recv entry pointed by sr_request_recv;
      	       */
	       sr=srrecvlist_create();
	       if (!sr) {
			printf("no memory\n");
			exit(1);
		}
      		sr->type= SRREPLY;
      		sr->ip_initiator.s_addr= reply->src_addr.s_addr;
      		sr->ip_replier.s_addr= reply->replier_addr.s_addr;
      		sr->ip_target.s_addr= reply->desti_addr.s_addr;
      		sr->tuple= NULL; 
      }       
      sr->requestId_secs= reply->ntp_sec;
      sr->requestId_fract= reply->ntp_fra;
      sr->orig_lifetime = reply->lifetime; 
      sr->lifetime =reply->lifetime;
      sr->tuplecount = reply->tuplecount;
      if (reply->tuplecount!= 0){
      	if (sr->tuple==NULL){
      		sr->tuple= (struct pathtuple *) 
		   malloc(reply->tuplecount*sizeof(struct pathtuple));
      	} else {
      		sr->tuple= (struct pathtuple *)realloc(sr->tuple, 
		   reply->tuplecount*sizeof(struct pathtuple));
      	}
      	memcpy(sr->tuple,tuples,sizeof(struct pathtuple) * reply->tuplecount);

      	/* update SR_hlist and routing table
      	 */
      	Mcast_Reply_Update(tuples,reply->tuplecount,reply->lifetime,
		replybyagent, reply->replier_addr, reply->desti_addr, 
		reply->coa,ifaddr);
      }

      /* I am the destination of the SR reply 
       */
      if (reply->src_addr.s_addr==ifaddr.s_addr) {
#ifdef MND
         /* the replied_flag is in SR_Sent table, set it so that 
          * there won't be SR_Request again
          */
	 ss=srsentlist_search(reply->desti_addr);
	 if(ss)  ss->replied_flag=TRUE;
#endif
      }
      /* not yet the destination */ 
      else {
	 reply_multicastForward(replyPacket, ifaddr);
     }
  }
  dumplists();
  return;
}


/*
 * 6.2 update hlist and route table in unicast mode

 * input: 
 * 1. tuple, pointer to the first tuple in the SR reply packet;
 * 2. tupleCount, copied from the SR reply packet: tuple_count;
 * 3. current_tuple, current tuple, copied from SR reply packet;
 * 4. lifetime for any created/updated  SR_hlist entry;
 */
void Unicast_Reply_Update (tuple,tupleCount, replybyagent,
                           current_tuple, lifetime,ip_replier,ip_dst, coa,ifa)
struct pathtuple *tuple;
int tupleCount;
int replybyagent;
int current_tuple;
int lifetime;
struct in_addr ip_replier;
struct in_addr ip_dst;
struct in_addr coa;
struct in_addr ifa;
{
	struct in_addr dst_host_ip;
        /* loop through the tuples in the SR reply packet and for
         * each one, find out whether to add/update route for the 
         * corresponding host;
         */
        int tupleOffset=0;
        while (tupleOffset < current_tuple) {
           /* pick current dst host
            */
           if (tupleOffset == 0) {
              dst_host_ip.s_addr= ip_replier.s_addr;
           } else {
              dst_host_ip.s_addr= tuple[tupleOffset-1].ipaddr.s_addr;
           }
	   SR_Update_hlist(tuple,current_tuple,lifetime,tupleOffset,
			   dst_host_ip,FALSE, ifa);
           tupleOffset++;
        }
	if (replybyagent && current_tuple) {
	   	SR_Update_hlist(tuple,current_tuple,lifetime,0,ip_dst,TRUE,ifa);
#ifdef MND
	   	SR_Update_hagentlist(ip_replier, coa,ifa);
#endif
 	}	   
}

#ifdef MND
struct alist * alist_search(struct alist*, struct in_addr );
SR_Update_hagentlist( ip_replier, coa, ifa)
struct in_addr ip_replier;
struct in_addr coa;
struct in_addr ifa;
{

	struct SR_hlist * h;
	struct alist * ha;
	h= srhlist_search(ip_replier);
	if (h){
		ha=alist_search(sr_hagentlist, ip_replier);
		if( ha ) {
			if (   ha->sigaverage !=h->metric ){
				alist_delete(&sr_hagentlist, ip_replier);
				srhagentcount--;
				alist_search(sr_hagentlist, ip_replier);
			} else {
				return;
			}
		}
		alist_create(&sr_hagentlist, ip_replier);
		srhagentcount++;
		sr_hagentlist->identity=isHa(ip_replier)? OURHA:OURFA;
		sr_hagentlist->sawAuth=1;
		sr_hagentlist->auth_ok=1;
		sr_hagentlist->sawMac=0;
		sr_hagentlist->mnroute=1;
		sr_hagentlist->mac[0]=0;
		sprintf(sr_hagentlist->location, "several hops away");
		sr_hagentlist->lsize=20;
		/* TBD
		sr_hagentlist->auth_tssecs=;
		sr_hagentlist->auth_tsfract=;
		*/
		sr_hagentlist->lifetime=h->lifetime;
		sr_hagentlist->sigaverage=h->metric;
		sr_hagentlist->coa.s_addr = coa.s_addr;
		alist_sort(&sr_hagentlist,sr_hagentlist);
	} else {
#ifdef DEBUG
		if(debug)
		printf("strange, no entry in srhlist of %s\n", 
		       inet_ntoa(ip_replier));
#endif
	}
}
#endif

void Mcast_Reply_Update (tuple, tupleCount, lifetime, replybyagent,ip_replier,
			ip_dst, coa, ifa)
struct pathtuple *tuple;
int tupleCount;
int lifetime;
int replybyagent;
struct in_addr ip_replier;
struct in_addr ip_dst;
struct in_addr coa;
struct in_addr ifa;
{
	struct in_addr dst_host_ip;
        int tupleOffset=0;
        while (tupleOffset < tupleCount) {
           /* pick current dst host
            */
           if (tupleOffset == 0) {
              dst_host_ip.s_addr= ip_replier.s_addr;
           } else {
              dst_host_ip.s_addr= tuple[tupleOffset-1].ipaddr.s_addr;
           }
           SR_Update_hlist(tuple, tupleCount, lifetime, tupleOffset,
			   dst_host_ip, FALSE, ifa);
           tupleOffset++;
        }
	if (replybyagent && tupleCount) {
	   	SR_Update_hlist(tuple,tupleCount,lifetime, 0,ip_dst,TRUE, ifa);
#ifdef MND
	   	SR_Update_hagentlist(ip_replier, coa, ifa);
#endif
 	}	   
}





/* the flooding filter subroutine
 * This is a filter for a system that receives an SR reply applies
 * it to limit the multicast flooding and make other sanity checks.
 * The filter is also used to check authentication
 * (assuming IPSEC AH is not used).
 * Input:
 * replyPacket, pointer to the head of the incoming SR_reply;
 * Output:
 * where_recv, pointer to where the correspondant entry in SR_recv table is,
 * will be smaller than length of srrecvlist if an entry is found,and right the 
 * length of srrecvlist if not found (need to add a new entry)
 * Returns:
 * 1: discard packet
 * 0: packet ok
 * the algorithm here tends to make cheaper check earlier 
 */
int reply_flooding_filter (replyPacket,sr, ifaddr, sr_metric)  
char *replyPacket;
struct SR_recv ** sr;
struct in_addr ifaddr;
struct sr_metric * sr_metric;
{
	struct sr_reply_packet * rp;
	struct pathtuple * rep_paths;
	struct SR_recv * ss;
	int i, samerequestid, samerequesttype;
	struct mip_mse_udp_auth * mseauth;

	rp = (struct sr_reply_packet *) replyPacket;
        rep_paths=(struct pathtuple *) (replyPacket
			  + sizeof(struct sr_reply_packet));

	/* handle packet replied by myself */
	if (rp->replier_addr.s_addr== ifaddr.s_addr){
	   sr_metric->myself++;
	   return 1;
	}

	/* authenticate the packet */
	mseauth=(struct mip_mse_udp_auth *)(replyPacket+  
				sizeof(struct sr_reply_packet)+
				rp->tuplecount*sizeof(struct pathtuple));
	if (!sr_ipsecah() && (!udpSecurityCheck(replyPacket,mseauth))){
	   sr_metric->security++;
	   return 1; 
	}

	/* check ttl if at IS */
	if (rp->src_addr.s_addr != ifaddr.s_addr) {
	   rp->udp_ttl--;
	   if (rp->udp_ttl <= 0){
	      sr_metric->ttl++;
	      return 1;
	   }
	}

	/* loop check */
	for(i=0; i< rp->tuplecount; i++) 
		if ((rep_paths+i)->ipaddr.s_addr== ifaddr.s_addr){
			sr_metric->loop++;
	   		return 1;
		}

	if( sr_if_prefer_neighbour()){ 
	 /* check if there is loop in neighbour */
	 for(i=0; i<(rp->tuplecount-1);i++){
#ifdef MIPD 
		 struct vlist *v;
		 if (((v=vlist_search(vlist_gethahead(), (rep_paths+i)->ipaddr))
		   ||(v=vlist_search(vlist_getfahead(), (rep_paths+i)->ipaddr)))
		     && v->sawMac){
#else
		 if (alist_search(alist_getmnhead(), (rep_paths+i)->ipaddr)) { 
#endif
			sr_metric->linkloop++;
			 return 1; 
		 }
	 }
	 if((rp->src_addr.s_addr== ifaddr.s_addr) || (rp->tuplecount!=0)){
#ifdef MIPD
		 struct vlist *v;
		 if (((v=vlist_search(vlist_gethahead(), rp->replier_addr))
		    || (v=vlist_search(vlist_getfahead(), rp->replier_addr)))
		     && v->sawMac){
#else
	 	if (alist_search(alist_getmnhead(), rp->replier_addr) ){
#endif
		 	sr_metric->linkloop++;
		 	return 1;
		 }
	 }
	}

        /* handle the same reply received before
	 * (via different path and received at a later time)
	 */
	samerequestid=false;
	samerequesttype=false;
	srrecvlist_initwalk();
	while(1){
		ss=srrecvlist_nextwalk();
		if(!ss) break;
		if (ss->ip_initiator.s_addr== rp->src_addr.s_addr 
		    && ss->type==SRREPLY) 
		{
			if (ss->requestId_secs== rp->ntp_sec &&  
			    ss->requestId_fract== rp->ntp_fra){
				samerequestid= true;
				break;
			} else if (
		                ss->ip_target.s_addr==rp->desti_addr.s_addr &&
		                ss->ip_replier.s_addr==rp->replier_addr.s_addr
		  		/* what if ip_replier isnot the same ??*/
				   ) {
				samerequesttype=true;
				break;
			}
		}
	}
	*sr=ss;
	if (samerequestid) {
   		/* at IS: discard packet of the same reply */
   		if (rp->src_addr.s_addr != ifaddr.s_addr){
			sr_metric->samerequest++;
      			return 1;
		}
   		/* at destination: discard if the metric is not better */
   		else{
			if (ss->tuplecount==0){
				sr_metric->badmetric++;
				return 1;
			} else if ( rp->tuplecount!=0) {
			   if (cumulativePathMetrics(rep_paths, rp->tuplecount)
			       <= 
			       cumulativePathMetrics(ss->tuple, ss->tuplecount))
			       {
				sr_metric->badmetric++;
      				return 1;
			       }
			}
		}
	}
	/* if not the previous case, we should still consider the 
 	 * scenario of the same request type (different requestId but
 	 * same ip_src, ip_dst) when the source initiate SR reply 
 	 * more than once; 
 	 */
	else if (samerequesttype) {
   		/* smaller requestId*/
		if ((ss->requestId_secs > rp->ntp_sec) ||
		    ((ss->requestId_secs ==rp->ntp_sec) &&
		     (ss->requestId_fract > rp->ntp_fra))){
			sr_metric->samedesti++;
      			return 1;
		}
	}	
   	/* else, not found related SR_recv entry */ 
	
	/* packet passed filter */
	
	return (0);
}

/*
 *  subrouting to forward the incoming SR reply to multicast interface
 */
void reply_multicastForward(char *replyPacket, struct in_addr ifaddr) {
  	struct sockaddr_in to;
	u_char ttl;
	struct pathtuple * pt;
	int i,rc;
	int offset;
	struct sr_reply_packet * reply = 
			       (struct sr_reply_packet *) replyPacket;	
	 
   	/* append <MY_IP, my_power_strength(), incoming_signal_strength()> tuple
     	 * to the SR reply packet;
	 */
	pt= (struct pathtuple *) (replyPacket + 
			sizeof(struct sr_reply_packet) +
			sizeof(struct pathtuple) * reply->tuplecount);
	pt->ipaddr.s_addr= ifaddr.s_addr ;
	if(reply->tuplecount)
		pt->signalstrength=  WIRESIGFUNCTION((pt-1)->ipaddr);
 	else
		pt->signalstrength=  WIRESIGFUNCTION(reply->replier_addr);
	pt->powerstrength= my_power_strength();

   	reply-> lifetime = MIN( reply->lifetime, sr_lifetime());  
	reply->tuplecount++;
	reply->current_tuple_pointer++;

	offset=sizeof(struct sr_reply_packet)+ reply->tuplecount
	        *sizeof(struct pathtuple);
        if (sr_netauth()) {
		offset+= addudpauth(replyPacket, offset); 
	}    

	to.sin_family = AF_INET;
	to.sin_addr.s_addr = sr_multicastip();
	to.sin_port = htons(sr_udpport());

  	rc = SENDTO(madSocket, replyPacket, offset, 0,
		   (struct sockaddr *) &to, sizeof(struct sockaddr_in));
	if(rc<0){
		   perror("sendto");
		   exit(1);
	}

}

#endif
