/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef MADRP
#include "sr_mad.h"
#ifdef IPSEC
#include "ipsec.h"
#ifdef MND
#include "mnconfig.h"
#else
#include "mipconfig.h"
#endif  /* MND */
#endif  /* IPSEC */

extern int debug; 
void 	SR_Update_hlist (struct pathtuple * , int,int,int, struct in_addr,int,
				struct in_addr); 
int 	hop_count_conversion(int );
void 	request_multicastForward(char *, struct in_addr) ;
int	udpSecurityCheck(char *, struct mip_mse_udp_auth *);
struct SR_hlist * srhlist_create();
struct SR_hlist * srhlist_search();
struct SR_recv * srrecvlist_nextwalk();
struct SR_recv * srrecvlist_create();
#ifdef MND
#include "alist.h"
struct alist * alist_getmnhead();
struct alist * alist_search(struct alist*, struct in_addr );
#else
#include "mipconfig.h"
#include "mipstate.h"  
#include "mipstats.h"  
#include "vlist.h"
#include "mip_log.h"
struct vlist ** vlist_gethahead_pointer();
struct vlist ** vlist_getfahead_pointer();
struct vlist * vlist_gethahead();
struct vlist * vlist_getfahead();
struct vlist * vlist_search(struct vlist*, struct in_addr );
struct mip_ifs *getConfigIf(struct in_addr *);
extern struct ha_stats ha_stats;
extern struct fa_stats fa_stats;
#endif

extern u_int   		madSocket;

/* 
 * the flooding filter subroutine
 * This is a filter for a system that receives an SR request applies
 * it to limit the multicast flooding and make other sanity checks.
 * The filter is also used to check authentication
 * (assuming IPSEC AH is not used).
 * Input:
 * requestPacket, pointer to the head of the incoming SR_request;
 * Output:
 * where_recv, pointer to where the correspondant entry in SR_recv table is,
 * will be smaller than length of srrecvlist if an entry is found, and right the 
 * length of srrecvlist if not found (need to add a new entry
 * Returns:
 * 1: discard packet
 * 0: packet ok
 * the algorithm here tends to make cheaper check earlier 
 */
int request_flooding_filter (requestPacket, sr, ifaddr, sr_metric)  
char *requestPacket;
struct SR_recv ** sr;
struct in_addr ifaddr;
struct sr_metric * sr_metric;
{
	struct sr_request_packet * rp;
	struct pathtuple * req_paths;
	int i, samerequestid, samerequesttype;
	struct mip_mse_udp_auth * mseauth;
	struct SR_recv * ss;

	rp = (struct sr_request_packet *) requestPacket;
        req_paths=(struct pathtuple *) (requestPacket
			  + sizeof(struct sr_request_packet));

	/* handle packet sent by myself 
	*/
	if (rp->sender_addr.s_addr== ifaddr.s_addr){
	   sr_metric->myself++;
	   return 1;
	}

	/* authenticate the packet */
	mseauth=(struct mip_mse_udp_auth *)(requestPacket+  
				sizeof(struct sr_request_packet)+
				rp->tuplecount*sizeof(struct pathtuple));
	if (!sr_ipsecah() && (!udpSecurityCheck(requestPacket,mseauth))){
	   sr_metric->security++;
	   return 1; 
	}

	/* check ttl if at IS */
	if (rp->desti_addr.s_addr != ifaddr.s_addr) {
	   rp->udp_ttl--;
	   if (rp->udp_ttl <= 0){
	      sr_metric->ttl++;
	      return 1;
	   }
	}

	/* loop check */
	for(i=0; i< rp->tuplecount; i++) 
		if ((req_paths+i)->ipaddr.s_addr== ifaddr.s_addr){
			sr_metric->loop++;
	   		return 1;
		}

	if( sr_if_prefer_neighbour()){ 
	/* check if there is loop in neighbour */
	for(i=0; i<(rp->tuplecount-1);i++){
#ifdef MIPD
		struct vlist * v;
		if (((v=vlist_search(vlist_getfahead(), (req_paths+i)->ipaddr))
		  || (v=vlist_search(vlist_gethahead(), (req_paths+i)->ipaddr)))
		    && v->sawMac) {
#else
		if (alist_search(alist_getmnhead(), (req_paths+i)->ipaddr)) {
#endif
			sr_metric->linkloop++;
			return 1;
		}
	}
	if((rp->desti_addr.s_addr== ifaddr.s_addr) && (rp->tuplecount!=0)){ 
#ifdef MIPD
		struct vlist * v;
		if (((v=vlist_search(vlist_getfahead(), rp->sender_addr))
		  || (v=vlist_search(vlist_gethahead(), rp->sender_addr)))
		    && v->sawMac)
#else
	     	if (alist_search(alist_getmnhead(), rp->sender_addr) )
#endif
		{
			sr_metric->linkloop++;
			return 1;
		}
	}
	}

        /* handle the same request received before
	 * (via different path and received at a later time)
	 */
	samerequestid=false;
	samerequesttype=false;
	srrecvlist_initwalk();
	while(1){
		ss=srrecvlist_nextwalk();
		if(!ss) break;
		if (ss->ip_initiator.s_addr== rp->sender_addr.s_addr
		    && ss->type==SRREQUEST) 
		{
			if (ss->requestId_secs== rp->ntp_sec &&  
			    ss->requestId_fract== rp->ntp_fra){
				samerequestid= true;
				break;
			} else if (ss->ip_target.s_addr==
				   rp->desti_addr.s_addr) {
				samerequesttype=true;
				break;
			}
		}
	}
	*sr=ss;
	if (samerequestid) {
   		/* at IS: discard packet of the same request */
   		if (rp->desti_addr.s_addr != ifaddr.s_addr) {
			sr_metric->samerequest++;
      			return 1;
		}
   		/* at destination: discard if the metric is not better */
   		else{
			if (ss->tuplecount==0){
				sr_metric->badmetric++;
				return 1;
			} else if ( rp->tuplecount!=0) {
			   if (cumulativePathMetrics(req_paths, rp->tuplecount)
			       <= 
			       cumulativePathMetrics(ss->tuple, ss->tuplecount))
			       {
				sr_metric->badmetric++;
      				return 1;
			       }
			}
		}
	}
	/* if not the previous case, we should still consider the 
 	 * scenario of the same request type (different requestId but
 	 * same ip_src, ip_dst) when the source initiate SR request 
 	 * more than once; 
 	 */
	else if (samerequesttype) {
   		/* smaller requestId*/
		if ((ss->requestId_secs > rp->ntp_sec) ||
		    ((ss->requestId_secs ==rp->ntp_sec) &&
		     (ss->requestId_fract > rp->ntp_fra))){
			sr_metric->samedesti++;
      			return 1;
		}
	}	
   	/* else, not found related SR_recv entry */ 
	
	/* packet passed filter */
	
	return (0);
}

/*
 * update the corresponding table (hlist and routing table)

 * input: 
 * 1. tuple, pointer to the first tuple in the SR request packet;
 * 2. tupleCount, copied from the SR request packet: tuple_count;
 * 3. lifetime for any created/updated  SR_hlist entry;
 */
void Recv_Request_Update (tuple, tupleCount, lifetime, ip_initiator, ifa)
struct pathtuple *tuple;
int tupleCount;
int lifetime;
struct in_addr ip_initiator;
struct in_addr ifa;
{
	/* loop through the tuples in the SR request packet and for
	 * each one, find out whether to add/update route for the 
	 * corresponding host;
	 *
	 * tupleCount == 1, is the one ajacent to the current host,
	 * we aleady have link layer access to this host, so it
	 * is not considered;
	 *
	 * But we have missed the ip_src so far, so we check the
	 * initiator when tupleCount equals 1 and finish the
	 * loop right after that.
         *
	 */
	struct in_addr dst_host_ip;
	int tupleOffset;
	tupleOffset=0;
	while (tupleOffset < tupleCount) {
	   /* pick current dst host
	    */
           if (tupleOffset == 0) {
              dst_host_ip.s_addr= ip_initiator.s_addr;
           } else {
	      dst_host_ip.s_addr= tuple[tupleOffset-1].ipaddr.s_addr;
           }
	   SR_Update_hlist(tuple, tupleCount,
			lifetime, tupleOffset,dst_host_ip, FALSE, ifa);
	   tupleOffset++;
	}
}

/*
 *  update one hlist entry

 * input: 
 * 1. all the inputs to recv_Request_Update ( ... );
 * 2. the resulting cadidate host IP;
 */
void SR_Update_hlist (tuple, tupleCount, lifetime,offset,dstHostIP,
			replybyagent, ifa) 
struct pathtuple *tuple;
int tupleCount;
int lifetime;
int offset;  /* offset of the dstHostIP in tuples */
struct in_addr dstHostIP;
int replybyagent;
struct in_addr ifa;
{
	   int i,currentMetrics;
	   struct SR_hlist * thishlist;
	   int would_add_host;
#ifdef MND
	   struct alist * thisalist;
	   struct alist * alisthead;
#else 
	   struct vlist * thisvlist;
	   struct vlist * vhead;
	   struct vlist ** vheadp;
#endif
	   struct in_addr next_hop;
	   int isha=0;

	   /* calculate metrics to the current dst host 
	    */
           currentMetrics=cumulativePathMetrics(tuple+offset,tupleCount-offset);

#ifdef DEBUG	   
	if(debug){
	   printf("intend to add an entry to hlist:\n");
	   printf("dst: %s\t", inet_ntoa(dstHostIP));
	   printf("next hop: %s\t", inet_ntoa(tuple[tupleCount-1].ipaddr));
	   printf("metric: %d\tisAgent: %d\n", currentMetrics,replybyagent);
	}
#endif
	   thishlist= srhlist_search(dstHostIP);
	   /* create a new hlist entry */
	   if (thishlist==NULL) 
	   {
		thishlist= srhlist_create();
		if(!thishlist) {
			printf("no memory!\n");
			exit(1);
		}
		thishlist->destHost.s_addr=dstHostIP.s_addr;
	   }
	   /* stop if the old hlist is good enough
	    */
	   else {
#ifdef DEBUG	   
		if(debug){
		printf("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
	   	printf("a new hlist entry option:\n");
	   	printf("dst: %s\t", inet_ntoa(dstHostIP));
	   	printf("next hop: %s\t", inet_ntoa(tuple[tupleCount-1].ipaddr));
	   	printf("metric: %d\tisAgent: %d\n",currentMetrics,replybyagent);
		}
#endif
		if ( (replybyagent==thishlist->isAgent &&
		      currentMetrics<= thishlist->metric) ||
		     (replybyagent && !thishlist->isAgent )) {
			if ( thishlist->nextHop.s_addr
				!=tuple[tupleCount-1].ipaddr.s_addr){ 
#ifdef DEBUG
				if(debug){
				printf("old entry in hlist is good enough\n");
				srhlist_dump();
				}
#endif
	      			return;
			}
		}
	   }
	   /* for both the old entry or newly created entry */
	   thishlist->nextHop.s_addr = tuple[tupleCount-1].ipaddr.s_addr;
	   thishlist->metric = currentMetrics;
	   thishlist->lifetime= sr_lifetime();
	   thishlist->isAgent=replybyagent;
	  
	   would_add_host=0;
	   if (is_host_route(thishlist->destHost, &next_hop)) {
	      if (next_hop.s_addr != thishlist->nextHop.s_addr) {
#ifdef DEBUG
		if(debug){
		 printf("change routing table, the nexthop of %s",
			inet_ntoa(thishlist->destHost));
		 printf(" becomes %s", inet_ntoa(thishlist->nextHop));
		 printf(" the old one is %s\n", inet_ntoa(next_hop));
		}
#endif
		 route_delete_host(thishlist->destHost, next_hop);
		 would_add_host=1;
              }
	   }
	   else {
		 /* in case of old entry is link layer entry */
#ifdef MND
		 alisthead=NULL;
		 alisthead=alist_getmnhead();
		 thisalist=alist_search(alisthead,thishlist->destHost);
		 if(thisalist) {
			/* TBD , maybe we should wait no-beacon to delete it?
		 	alist_delete(&alisthead,thishlist->destHost);
		 	route_delete_link(thisalist->bip, thisalist->mac);
	         	route_add_host(thishlist->destHost, thishlist->nextHop, ifa);
			*/
		 } else 
#endif
			would_add_host=1;
	   }
	   if(would_add_host) {
#ifdef IPSEC
	       struct keylist * key;
               /* check for (and build) a route SA for destHost 
		*/
#ifdef MND
		for (i=0;i<MN_SECTYPE_MAX;i++) {
			if(mn_config.ipsec[i].inuse  
		  	  && (key = search_keylist( mn_config.ipsec[i].kl,
			  			thishlist->destHost.s_addr)) 
			  && key->inuse ) {
		      		set_rosa(TRUE, key, mn_config.mn, 
					&thishlist->rosa, i, 
					thishlist->destHost);
				break;
			}
	     	}
#else
#define MOBILENODE      3
		for (i=0;i<2;i++) {
			if(mip_config.ipsec[i].inuse  
		  	  && (key = search_keylist( mip_config.ipsec[i].kl,
			  			thishlist->destHost.s_addr)) 
			  && key->inuse ) {
		      		set_rosa(TRUE, key, ifa, 
					&thishlist->rosa, MOBILENODE, 
					thishlist->destHost);
				break;
			}
	     	}
#endif
#endif

	        route_add_host(thishlist->destHost, thishlist->nextHop, ifa
#ifdef IPSEC
				,&thishlist->rosa
#endif
				);
	   }
	   return;
}

#ifdef MIPD
int isHa(struct in_addr bip){
	 struct mip_ifs *mips;
	 int isha=0, i;
	/* find out if beacon matches any HA interface present
	 * here.  If so, we put MAC address in ha vlist.
	 * note that the interface pointer mifs points to
	 * the *input* i/f, which might be that of a FA.
	 */
	mips = &mip_config.mip_ifs[0];
 	for ( i = 0; i < mip_config.mip_ifcount; i++, mips++ ) {
        	if (mips->amfa) /* (1 if fa) */
                	continue;
		if ((bip.s_addr & mips->netmask.sin_addr.s_addr) ==
			(mips->if_sin.sin_addr.s_addr & mips->netmask.sin_addr.s_addr))  {
			isha = 1;
			break;
		}
	} 
#ifdef DEBUG
		if(debug)
		printf("isHa: %d\n", isha);
#endif
}
#else
#include "mnconfig.h"
extern struct mn_config mn_config;
int isHa(struct in_addr bip){
	return((bip.s_addr & mn_config.netmask.s_addr) ==
		(getmyaddr() & mn_config.netmask.s_addr)) ; 
}
#endif

/*
 *  receive an SR request
 *  input: the pointer to the head of the incoming SR_request packet;
 */
void Recv_SR_request(char *requestPacket, struct in_addr ifaddr, struct sr_metric
* sr_metric) {
	struct sr_request_packet * request ;
	struct pathtuple * tuples ;
	struct SR_recv   * ss ;
	int j=0,replybyagent;

	request = (struct sr_request_packet *) requestPacket;
	tuples=(struct pathtuple *) (requestPacket+ 
				     sizeof(struct sr_request_packet));

	/* first of all, do the anti-flooding job 
	 */
	if (request_flooding_filter(requestPacket, &ss, ifaddr, sr_metric) == 1){
#ifdef DEBUG
		if(debug){
		printf("it is thrown away\n");
		dumpsr_metric(*sr_metric);
		}
#endif
	   	return;
	}

	/* 
	 * update SR_recv table
	 *
	 * if we have an SR_recv entry of the same request type
	 * found during the filtering subroutine, we will use that
	 * entry and overwrite with newer data, else create a new entry
	 */ 
	if ( ss==NULL) {
		/* create an SR_recv entry pointed by sr_request_recv;
		 */
		ss=srrecvlist_create();
		ss->type= SRREQUEST;
		ss->ip_initiator.s_addr= request->sender_addr.s_addr;
		ss->ip_target.s_addr= request->desti_addr.s_addr;
		ss->tuple= NULL; 
	}
	ss->requestId_secs= request->ntp_sec;
	ss->requestId_fract= request->ntp_fra;
	ss->orig_lifetime =  request->lifetime;
	ss->lifetime =request->lifetime; 
	ss->tuplecount = request->tuplecount;

	if ( request->tuplecount !=0) {
		if (ss->tuple==NULL){
			ss->tuple= (struct pathtuple *)malloc
        		    (request->tuplecount * sizeof(struct pathtuple));
		} else {
			ss->tuple=(struct pathtuple *)realloc(ss->tuple, 
        			request->tuplecount * sizeof(struct pathtuple));
		}

		memcpy(ss->tuple,tuples,
	       		sizeof(struct pathtuple) * request->tuplecount);
		/* didn't set ip_replier */

		/* update SR_hlist and routing table 
	 	 * tuples, tuple_count and lifetime are the field in the 
		 * incoming packet;
		 */
		Recv_Request_Update(tuples, request->tuplecount, 
			    request->lifetime, request->sender_addr, ifaddr);
	}

	/* send the SR reply if I am the target 
	 */
	if (ifaddr.s_addr == request->desti_addr.s_addr) {
		if(request->tuplecount==0 && sr_pushbeacon()) {
#ifdef MIPD
			beaconLoop();
#else
			pushBeacon();
#endif
		}
	   	Send_SR_reply(IsAgent(), TRUE, requestPacket,ifaddr);
	   	return;
	}
	/* forward the SR request if I am not the target 
	 */
	else {
	   /* agent forward SR request packet */
	   if (IsAgent() && 
               (request->flag & AGENT_REPLY_ACCEPTABLE_BIT)
		 ==AGENT_REPLY_ACCEPTABLE_BIT) 
	      	Send_SR_reply( TRUE, FALSE, requestPacket,ifaddr);
	   request_multicastForward(requestPacket,ifaddr);
	}
#ifdef DEBUG
	if(debug){
	dumplists();
	}
#endif
}


/*
 *  subroutine to forward the incoming SR request to multicast interface
 */
void request_multicastForward(char *requestPacket, struct in_addr ifaddr) {
  	struct sockaddr_in to;
	u_char ttl;
	struct pathtuple * pt;
	int i,rc;
	int offset;
	struct sr_request_packet * request = 
			       (struct sr_request_packet *) requestPacket;	
	 
   	/* append <MY_IP, my_power_strength(), incoming_signal_strength()> tuple
     	 * to the SR request packet;
	 */
	pt= (struct pathtuple *) (requestPacket + 
			sizeof(struct sr_request_packet) +
			sizeof(struct pathtuple) * request->tuplecount);
	pt->ipaddr.s_addr=ifaddr.s_addr;

	if(request->tuplecount)
		pt->signalstrength=  WIRESIGFUNCTION((pt-1)->ipaddr);
 	else
		pt->signalstrength=  WIRESIGFUNCTION(request->sender_addr);

	pt->powerstrength= my_power_strength();

   	request-> lifetime= MIN(request->lifetime, sr_lifetime());  
	request->tuplecount++;
	    
	offset=sizeof(struct sr_request_packet)+ request->tuplecount
	        *sizeof(struct pathtuple);
        if (sr_netauth()) {
		offset+= addudpauth(requestPacket, offset); 
	}    

	to.sin_family = AF_INET;
	to.sin_addr.s_addr = sr_multicastip();
	to.sin_port = htons(sr_udpport());

  	rc = SENDTO(madSocket, requestPacket, offset, 0,
		   (struct sockaddr *) &to, sizeof(struct sockaddr_in));
	if(rc<0){
		   perror("sendto");
		   exit(1);
	}

}

#endif

