/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef MADRP
#include "sr_mad.h"

extern u_int 	madSocket;
extern int debug; 
#ifdef MIPD
#include "mipconfig.h"
extern struct mip_config mip_config;
#endif
/*
 *  send the SR reply:
 *
 * input: 
 * 1. isAgent, a flag showing whether it is from an agent;
 * 2. isDest, a flag showing whether it is from the SR target
 * 3. request->acket, SR request packet, we will build SR reply packet
 *    based on SR request packet;
 *
 * output: reply packet
 */
void Send_SR_reply (isAgent, isDest, requestpacket,ifaddr)
int isAgent;
int isDest;
char *requestpacket;
struct in_addr ifaddr;
{
        int i,j;
	int offset=0;
	u_char packet[1024];
	struct sr_reply_packet *srp;
	struct sr_request_packet *request;
	struct pathtuple * req_paths;
	struct pathtuple * reply_paths;
	struct sockaddr_in to;
	u_int rc;
	u_char ttl;

	request = (struct sr_request_packet *) requestpacket;
        req_paths=(struct pathtuple *) (requestpacket
		  + sizeof(struct sr_request_packet));
	srp=(struct sr_reply_packet *)packet;
	reply_paths=(struct pathtuple *)(packet
		    + sizeof(struct sr_reply_packet));

	/* build SR reply packet 
	 */
	srp->type     = SRREPLY;
	srp->version  = sr_version();
	srp->lifetime = MIN(request->lifetime,sr_lifetime());
	srp->ntp_sec  = request->ntp_sec;
	srp->ntp_fra  = request->ntp_fra;
	srp->replier_addr.s_addr = ifaddr.s_addr; 
	srp->src_addr.s_addr= request->sender_addr.s_addr; 
	srp->desti_addr.s_addr = request->desti_addr.s_addr;  
#ifdef MIPD
	srp->coa.s_addr=mip_config.fa_coa.sin_addr.s_addr;
#endif
	srp->tuplesize =  request->tuplesize;
	srp->pad =  0;
	srp->flag =0 ;
	srp->udp_ttl  = sr_udpttl();  
       
        if (sr_unicastreply() ) {
	      srp->flag = srp->flag | UNICAST_BIT;
	      srp->tuplecount = request->tuplecount;
	      srp->current_tuple_pointer =0;   
	
	      /*  reversely copy tuple(s) from requestpacket to reply 
	       *  packet;
	      */
	      j=0;
	      for (i = (int) request->tuplecount-1; i>=0; i--){
		memcpy(reply_paths+ j, req_paths+ i,request->tuplesize);
		j++;
	      }
	}
	else {
	      srp->tuplecount = 0;
              srp->flag = srp->flag | MULTICAST_BIT;
	}
	if (isAgent) 
	      srp->flag = srp->flag | AGENT_BIT;
	if (isDest) 
	      srp->flag = srp->flag | DESTINATION_BIT;

	offset = sizeof(struct sr_reply_packet) 
		 + srp->tuplecount * srp->tuplesize;

	if (sr_netauth() ) {
		offset+= addudpauth(packet, offset);
	}

	if (sr_unicastreply() ) {
	        /* unicast this packet to the last tuple in the tuple list;
		 */

 		to.sin_family = AF_INET;
		if(srp->tuplecount==0) {
			to.sin_addr.s_addr=srp->src_addr.s_addr;
		} else {
			to.sin_addr.s_addr=
			     reply_paths[0].ipaddr.s_addr;
		}
		to.sin_port = htons(sr_udpport());

		rc = SENDTO(madSocket,packet, offset,0,
	                   (struct sockaddr *) &to, sizeof(struct sockaddr_in));
		if(rc<0){
			perror("sendto");
			/*
			exit(1);
			*/
		}
	}
	else /* MULTICAST_REPLY */ {
 		to.sin_family = AF_INET;
		to.sin_addr.s_addr = sr_multicastip();
		to.sin_port = htons(sr_udpport());

		rc = SENDTO(madSocket,packet, offset,0,
	                   (struct sockaddr *) &to, sizeof(struct sockaddr_in));
		if(rc<0){
			perror("sendto");
			exit(1);
		}
	}
#ifdef DEBUG
	if(debug)
	dumplists();
#endif
	return;
}



#endif
