/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef MADRP
#include "sr_mad.h"


extern struct SR_sent * srsentlist;
extern u_int 	madSocket;
extern int debug; 

void *  addnewentry(void * , int * , u_int);
int 	addudpauth(char * ,int);
char *	getNetauthKey();
int 	getNetauthSpi() ;
struct SR_sent * srsentlist_create();
struct SR_sent * srsentlist_search(struct in_addr);

u_int backoffarray[MAX_INDEX] = {
		2,4,8,8,16,16,32,32,32,64,64,64,64
	};

/* send the SR request
 *
 * input: 1. target ip
 *        2. a boolean value indicating whether it is called first time
 *	  a) if it is called first time, then create SR_sent entry
 *	     and timeout is set to the initial value;
 *        b) if it is not called first time, then we only update
 *	     SR_sent entry, esp. the timeout value;
 *
 *  note, this algorithm only takes care of how to send the SR request->
 *  not when to send/resend SR request->
 */
void Send_SR_request(struct in_addr target_ip, u_int first_time,struct in_addr ifaddr) {

        int i,offset=0;
	u_char packet[1024];
	struct sr_request_packet *srq;
	struct sockaddr_in to;
	struct SR_sent * ss;
	u_int rc;
	u_char ttl;

	if(target_ip.s_addr==getmyaddr()){
#ifdef DEBUG
		if(debug)
		printf("why send request for yourself?");
#endif
		return;
	}
	ss=srsentlist_search(target_ip);
	if (ss && first_time) {
#ifdef DEBUG
		if(debug)
		printf("the same query have already been sent\n");
#endif
		return;
	}

	/*  build the SR request packet
	 */
	srq=(struct sr_request_packet *)packet;
	srq->type     = SRREQUEST;
	srq->version  = sr_version();
	srq->lifetime = sr_lifetime();
	srq->ntp_sec  = get_ntptime_secs();
	srq->ntp_fra  = get_ntptime_fract();
	srq->sender_addr.s_addr= ifaddr.s_addr; 
	srq->desti_addr.s_addr = target_ip.s_addr;  
	srq->udp_ttl  = sr_udpttl();
	if (sr_agentreplyacceptable() )
		srq->flag = AGENT_REPLY_ACCEPTABLE_BIT;
	else
		srq->flag = 0;
	srq->tuplesize =  TUPLE_SIZE;
	srq->tuplecount =  0;

	/*
	if (SR_MAC == true)
		set the mac TLV
	*/

	offset = sizeof(struct sr_request_packet);

	if (sr_netauth()) {
		/* set the netauth TLV
		*/
		offset+=addudpauth(packet, offset);
	}

	/*   multicast this packet
	 */
 	to.sin_family = AF_INET;
	to.sin_addr.s_addr = sr_multicastip();
	to.sin_port = htons(sr_udpport());
	
	rc = SENDTO(madSocket,packet, offset, 0,
	       (struct sockaddr *) &to, sizeof(struct sockaddr_in));
	if(rc<0){
		perror("sendto");
		exit(1);
	}

	/* create an SR_sent list entry
	 */
	if (first_time ){
		ss= srsentlist_create();
		if(!ss){
			printf("no more mmeory\n");
			exit(1);
		}
		ss->ip_target.s_addr = target_ip.s_addr;
		ss->backoff_count =  0;
	}
	else {
		ss=srsentlist_search(target_ip);
		if(ss){
			ss->backoff_count++;
		}
		else  {
			printf("didn't find old entries in SR_sent\n");
			exit(1);
		}
	}
	ss->requestId_secs  = ntohl(srq->ntp_sec);
	ss->requestId_fract = ntohl(srq->ntp_fra);
	ss->replied_flag = 0;
	ss->backoff_timer = backoffarray[ss->backoff_count];
#ifdef DEBUG
	if(debug)
	dumplists();
#endif
	return ;
}

#endif
