/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef MADRP
#include "sr_mad.h"
#ifdef MND
#include "alist.h"
#include "mnstate.h"
#endif

char *	getNetauthKey();
int 	getNetauthSpi() ;
        
/* to add auth packet after the send packet */
int addudpauth(char * outpack, int offset){

	u_char * key;
	int spi;

	struct mip_mse_udp_auth * mseauth = (struct mip_mse_udp_auth *)
		((char *) outpack + offset);
	mseauth->mse_type = MIPMSEUDPAUTH;
	mseauth->mse_length = sizeof(struct mip_mse_udp_auth) - 2; /*why?*/
	mseauth->pad[0] = 0;
	mseauth->pad[1] = 0;
	mseauth->tssecs = htonl(get_ntptime_secs());
	mseauth->tsfract = htonl(get_ntptime_fract());
#ifdef MND      
	mseauth->agent = (mns.agent) ? (mns.agent->bip.s_addr) : 0;
#else /* MIPD */ 
	mseauth->agent = 0;
#endif   

	spi = htonl(getNetauthSpi());  
	memcpy( &mseauth->spi, &spi, sizeof(int));

	/* do auth. over all payload up to here including
	 * first two fields of mseauth. timestamp + pad
	 * are included in hash.
	 */ 
	key= getNetauthKey();

	if (key){ 
		storeMd5((char *)outpack, offset +
			   MIPAGENTAUTH_OFFSET, mseauth->hash, key);
		return sizeof(struct  mip_mse_udp_auth);
	} else 
		return 0;
}


/*
 * check udp security if it shows up.  note packet should have
 * had ip header removed from it.
 *
 * we don't log anything here.  we do however keep
 * totals in stats 
 *
 * return:
 *	1: auth ok
 *	0: failed authentication.  That means we got an authentication
 *	but the hash recompute failed
 *     -1: no key (mode switch).
 */
int udpSecurityCheck(char *packet, struct mip_mse_udp_auth *mseauth)
{
	int spi = mseauth->spi;
	unsigned char hash[16];
	int size;
	char *key;

	/* sanity check */
	if (mseauth == NULL)
		return (0);

	/* do security hash check
	 */
	size = (((char *)mseauth) - packet) +
		MIPAGENTAUTH_OFFSET;
	/* check hash first */
        /* recompute MD5 hash, result in local hash[]
        */ 
	/* TBD. do something with spi */
        spi = htonl(getNetauthSpi());
	key = getNetauthKey();
	if (key == NULL)
		return(-1);
	computeMd5(packet, size, key, hash);

	if (bcmp(mseauth->hash, hash, 16) != 0) {
                return(0);
        }
	return(1);
}


#endif
