
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * md5authent.c
 *
 * compute keyed md5 checksum for mobile-ip registration and reply pkts.
 *
 * storeMd5 - used by sending side to compute and store hash where
 *	desired
 * computeMd5 - used by recv. side to check hash
 *
 * could probably be made into one function but have slightly
 * different purposes.
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/times.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <md5.h>
#include <unistd.h>

/*
 * compute md5 over packet of psize, and store in mse
 *
 * md5 hash is based on (key, data, key) method 
 */
char *
storeMd5(char *packet, int psize, unsigned char *hash, char *keys)
{
	MD5_CTX ctx;

	MD5Init(&ctx);
	MD5Update(&ctx,keys,16);
	MD5Update(&ctx,packet,psize);
	MD5Update(&ctx,keys,16);
	MD5Final(hash, &ctx);
}

static unsigned char hash[16];

/* compute the hash and store it in the hash[16]
 * array. This can be used for comparison by the recv. entity
 */
char *
computeMd5(char *packet, int psize, char *keys, char *hash)
{
	MD5_CTX ctx;

	MD5Init(&ctx);
	MD5Update(&ctx,keys,16);
	MD5Update(&ctx,packet,psize);
	MD5Update(&ctx,keys,16);
	MD5Final(hash, &ctx);
}

#ifdef TESTMAIN
/*
 * test main - benchmark time to do a particular packet size
 */

main(int argc, char **argv)
{
	char buf[2000];
	char keys[16];
	char hash[16];
	int i;
	int count = atoi(argv[1]);
	int bufsize = atoi(argv[2]);

	if (bufsize >= 2000) {
		printf("buf too big\n");
		exit(1);
	}
	/* compute md5 checksum
	*/
	for (i = 0; i < count; i++)
		storeMd5(buf, bufsize, hash, keys);
}
#endif
