
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  FILE:	mip_log.h
 *  SYNOPSIS:	
 *
 *  --	
 *
 */

#ifndef _MIP_LOG_H
#define _MIP_LOG_H

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <time.h>
#include <signal.h>

#ifdef MLOGROUTINES	/* mip_log() uses newer variable argument #defines */
#include <stdarg.h>
#define	TRUE	1
#define FALSE	0

#else			/* syslog() uses the older... */
#include <syslog.h>
#include <stdarg.h>

#endif /* MLOGROUTINES */


extern const int	sys_nerr;

#define MAXBUF	1024




/* ========  specific  ======== */

#define MIPD_LOGFILE    "/var/log/mipd.log"
#define MND_LOGFILE     "/var/log/mnd.log"


/* local logfile should only be owner (root) readable only as it may
 * contain sensitive information indicating the security state of
 * connections in the mobile IP infrastructure
 */
#define MLOG_MODE	0600	/* persimmon bits for open */
#define NOLFILE		-1	/* no local logfile */

/* local log priority types -- not necessarily in order */
#define MLOG_UNKNOWN	0	/* unknown type */
#define	MLOG_INFO	1	/* information message */
#define MLOG_SEC	2	/* security related */
#define MLOG_ERR	3	/* (packet parsing) error */

#define MLOG_PRI_MAX	3

/* strings to appear in the logfile at the head of a line logged under
 * a given priority type
 */
static char	*mtypestr[MLOG_PRI_MAX+1] = {
    /* MLOG_UNKNOWN */	"unknown",
    /* MLOG_INFO */	"info",
    /* MLOG_SEC */	"security",
    /* MLOG_ERR */	"error"
};




/* ========  debugging  ======== */

#define entryMark( level, function )				\
    if ( debug && level >= noticeFunction )			\
	fprintf( stderr, "\nIN FUNCTION %s().\n", function );	\



/* ========  prototypes  ======== */

void	mip_closelog( void ),
	mip_chlogname( char *logname );

int	mip_openlog( char *logfile ),
	mip_log( int pri, char *fmt, ... );

#endif /* _MIP_LOG_H */

