/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mipconfig.h
 *
 * constants needed for mipd configuration
 */

#ifndef CONFIG_H
#define CONFIG_H		1

#define  TRUE			1
#define	 FALSE			0

#ifdef HARP
#define MIPSEC
/* harp additions */
#define UDP_SOC               0       /* regular mobile IP UDP socket */
#define HARP_UDP_SOC          1       /* harp protocol forwarding socket */
#define HARP_DUMP_SOC         2       /* harp protocol, dump forwarding soc. */
#define HARP_PORT             1588    /* harp protocol forwarding port */
#endif
#define UDP_PORT	      434	/* mobile-ip agent port */
#define DIVERTPORTIN          3200      /* In-coming Packets forwarding port */
#define DIVERTPORTOUT         2300      /* Out-coming Packets forwarding port */
#define BEACONTIMER		1	/* default ICMP advert rate */
#define ICMPLIFETIME		60      /* lifetime for icmp router advert 
					 *  (secs) 
					 */
#define MOBILELIFETIME		180	/* default max reregistration time
					 * for mobile node
					 */

/* 
 * The time an at-home deregistration in ad-hoc mode is considered
 * to last.  This is also defined in mnconfig.h.
 */
#define HALIFETIME  180

#define MAX_INTERFACES		50	/* same as ifconfig, still XXX */
#define MAXLINE			1024
#define MAXLOCATION		255

#include "keylist.h"

/* sample config line
 * fa ifname address
 */ 
struct mip_ifs {
	char *ifname;			/* name of interface */
	int mcast;			/* true if dst is multicast */
	char *router;			/* address to use for router adverts */
	struct sockaddr_in r_sin;       /* translation of above */
	int  amfa;			/* 1 if fa, 0 if ha */
	struct sockaddr_in if_sin;	/* ip address assigned to if */
	struct sockaddr_in netmask;     /* interface netmask */
	struct sockaddr_in cloneip;     /* interface clone route (subnet) */
	int	link_index;		/* index of link in o.s. */
	u_char mac[6];			/* mac address for this interface */
	unsigned short seq;		/* sequence number for this if */
	unsigned char found;		/* if exists in kernel */
	unsigned char doArp;		/* do arp or not */
	unsigned char both;		/* true if this interface both ha/fa:
					 * 1 == ha, 2 == fa
					 */
	struct mip_ifs *nextAgent;  	/* circular list of agents on this 
					 * interface.  NULL == empty.
					 */
	int doProxy;			/* TRUE if a HA on this i/f should
					 * create proxy ARP entries for
					 * remote MNs
					 */
};

#define CONFIGDEFAULT	"/etc/mipd.conf"

/* five kinds of security associations; i.e., KEYS 
*/
#define MNHASECURITY	0   /* MIP between MN and HA */
#define FAMNSECURITY	1   /* MIP between FA and MN */
#define FAHASECURITY	2   /* MIP between FA and HA */
#define NETAUTHSECURITY 3   /* not MIP: optional net icmp auth. extension */
#define ADHOCSECURITY   4   /* not MIP: optional ad hoc icmp auth. extension */
#ifdef HARP
#ifdef MIPSEC
#define HAHASECURITY	5   /* required if HARP is in use */
#endif
#define MIPSAS 		6
#else
#define MIPSAS		5
#endif

/* <ip | network | hostname | domain> to be denied or allowed by access control 
*/
    
struct element {
        char address[80];
#define ACCESS_CONTROL_IP_TYPE          0
#define ACCESS_CONTROL_DOMAIN_TYPE      1       
        int type;
        struct element * next;
};  


struct mip_config {
	struct mip_ifs mip_ifs[MAX_INTERFACES];
	int mip_ifcount;
        /* MIP SA lists. indexed by ip address */
	struct sec_assoc {
		int spi;
		unsigned char key[16];	/* md5 key */
		struct keylist *kl;
	} sa[MIPSAS];
	int beaconRate;
	int mobileLifetime;
#ifdef HARP
        int harp;               /* doing HARP */
	int harp_tunnel;	/* home subnet is partitioned? */
	int harp2_arp_disable;	/* arp is dynamically disabled on the 
				* secondary HA	*/
	char harp2_if[24];	/* harp2 interface */
	int harp_ping;
	int dorip;
#endif
	int sawFa;    		/* at least do FA function */
	int sawHa;    		/* at least do HA function */
	int sawCoa;    	        /* coa passed in */
	int sawBroadcast;      /* true if at least 1 if wants 255.255.255.255 */
	int sawMcast;           /* true if at least 1 if wants 224.0.0.1 */
	struct sockaddr_in fa_coa;  /* global coa for all fa interfaces, but one fa machine:-> */
#ifdef HARP
	struct sockaddr_in coha;    /* address of co-HA (for HARP)*/
	int haha_security;	/* haha security association */
	struct sockaddr_in rip_if;	/* ripv2 i/f to use */
	struct sockaddr_in rip_ha;	/* ha network to advertise */
	struct sockaddr_in rip_mask;    /* ha mask to advertise */
#endif
	int faha_security;	/* faha security association required */
	int mnha_security;	/* mnha security association required */
	int famn_security;	/* famn security association required */
	int netauth_security;    /* netauth icmp security association set */
	int adhoc_security;      /* adhoc icmp security association set */
	unsigned long ntp_slop;   /* UDP timeslop stored in secs */
#ifdef DIVERT
        int doNat;                /*  do dhcp/nat, IP-UDP Tunnelling Enabler */
	int fr_number;		/* First Firewall Rule number */
#endif
	int doTimestamp;	 /* UDP timestamp replay else nonce */
	int sawLocation;		/* true if saw location */
	char location[MAXLOCATION];	/* location string for ICMP */
	/* following not used currently */
	int beacon_timeslop;	/* allowed in replay protection*/
	int beacon_timestamp;   /* true then use timestamp, else nonce */
	int beacon_replay;	/* replay protection, apllies to all */
	int logIPOnly;		/* TRUE if DNS lookups disallowed in logfiles */
	char *ha_ifname;	/* name of one allowed ha if, if any */
	struct sockaddr_in ha_sin;    /* one ha if - its ip address */

#ifdef KAMEIPSEC
  	int kameipsecEnable;          /* kame ipsec capable */ 
#endif

	/* access control configurations 
	*/
        struct element  * denylist;             /* hosts to be denied access */
        struct element  * allowlist;            /* hosts to be allowed access */
#define DENY_ALLOW      0
#define ALLOW_DENY      1
        int             access_control_order;   /* processing order --
                                                   denylist first or allowlist
                                                   first */
#define ALLOWED         0 
#define DENIED          1
        int             access_control_default_permission; /* default access --
                                                   deny or allow */
};

extern struct mip_config mip_config;

#endif /* CONFIG_H */
