/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mnconfig.h
 *
 * configuration constants for mnd - mobile node daemon
 */ 
#ifndef CONFIG_H
#define CONFIG_H		1

#define  TRUE			1
#define	 FALSE			0

#define UDP_PORT		434	/* mobile-ip agent port */
#define DIVERTPORTIN            3002   /* In-coming Packets are diverted to this port */
#define DIVERTPORTOUT           2003    /* Out going Packets are diverted to this port */ 
#define MAXLINE			1024
#define CONFIGDEFAULT		"/etc/mnd.conf"
#define CONFIGDEFAULTBAK	"/etc/mnd.conf.bak"
#define CONFIGDEFAULTSAFE	"/etc/mnd.conf.safe"

#ifndef MAXLOCATION
#define MAXLOCATION		255
#endif

/* number of beacons to recv. whilst examining request from
 * user to switch agents.  agent beacon must be recv. in this span
 */
#define HINTBEACONS	60

#ifndef MAXIFNAME 
#define MAXIFNAME 128
#endif
#define MNHASECURITY	0
#define FAMNSECURITY	1
#define NETAUTHSECURITY	2
#define ADHOCSECURITY	3

#define MD5TIMESPI	256
#define MD5NONCESPI	257
#define NETAUTHACCEPT 	1
#define NETAUTHREJECT 	0
#define ASKON  		1
#define ASKOFF 		0

#define ON		1
#define OFF		0

/* 802.11 defines
*/

#define L80211TIMER	20	/* seconds per layer 2 change */
#define L80211MAXESSID   256     /* max essid string size */
#define L80211ESSDEFAULT "any"

/* layer 3 flag/hints
*/
#define L802ARP		0
#define L802AAH		1

/* layer 2 flag/hints
*/
#define L802MANAGED	0
#define L802ADHOC	1

/* 
 * The time an at-home deregistration in ad-hoc mode is considered
 * to last.  This is also defined in mipconfig.h.
 */
#define HALIFETIME  180

#include "keylist.h"

struct mn_config {
	struct sec_assoc {
		int spi;
		unsigned char key[16];	/* md5 key */
		struct keylist *kl;     /* keylist pointer */
	} sa[4];
	int doArp;		/* use arp or use net. auth with ICMP beacons */
	int doNetauth;		/* use netauth key */
	int doAdhoc;		/* use supplemental adhoc key */
#ifdef  DHCP
        int dhcpEnable;         /* dhcp-capable */
#define DHCPRENEWINGPERIOD     600	/* default */
        /* time to renew lease.
         * if not set, then 0, means accept dhcp server lease time
	 * default is 10 minutes.
	 * Note: we tell the server this parameter. 
	 * we also make sure any lease told us is less < this value.
	 */
        int dhcp_renewing_period;
        /* name of interface for a reverse tunnel, home address will be
         *  bound to this interface in DHCP mode 
	*/
        char mvifname[MAXIFNAME]; 
#endif /* DHCP stuff */

#ifdef KAMEIPSEC
  	int kameipsecEnable;          /* kame ipsec capable */ 
#endif
	char ifname[MAXIFNAME]; /* name of interface used for mobility */
	u_char mac[6];		/* mac address for that interface */
	struct in_addr 	mn;	/* ip address of mn */
	struct in_addr 	netmask;/* netmask address of mn interface */
	struct in_addr 	cloneip;/* clone route ip address */
	int    link_index;      /* link #n for interface */
	struct in_addr  ha;     /* ip address of ha */
	int sawHa;		/* flag to indicate ha field is valid */ 
	int sawMn;		/* flag to indicate mn field is valid */
	int beaconRate;		/* second rate for icmp beacons */
	int mnha_security;	/* mnha security association required */
	int famn_security;	/* famn security association required */
	int netauth_security;	/* icmp from agents authenticated */
	int adhoc_security;     /* icmp beacons from mns authenticated */
	/* the mode is a boolean.  if 1, then we try to fix the
	 * time from the HA NAK, else we don't and hope somebody
	 * else is working on it (xntpd or ntpdate or hw clock)
	 */
	u_short beacon_seq;     /* sequence number in beacon */
	/* UDP replay protection */
	int tsmode;		/* time stamp mode for UDP msg */
	int doTimestamp;	/* type of replay, ts is default, for UDP,
				   only works when replay is set */
#ifdef DIVERT
	int doNat;		/* Can Do Nat */
	int fw_rule;            /* Used to store the firewall rule number */
        int divertPortIn;
        int divertPortOut;
#endif
        int dhcp_policy;
	int doDns;  
	/* NONE of these ICMP replay protection vars are used */
	int beacon_replay;	/* boolean: do we want replay protection */
	int beacon_timestamp;    /* ICMP type of replay */
	int beacon_timeslop;    /* timeslop for ICMP msg */

	char location[MAXLOCATION];  /* who/where we are */
	int sawLocation;

	unsigned short udp_port;	/* udp port to use */
	int logIPOnly;		/* TRUE if DNS lookups disallowed in logfiles */

	int useMulticast;	/* beacons: else broadcast 255.255.255.255 */
	struct in_addr maddr;   /* multicast address */
};

extern struct mn_config mn_config;

#endif /* CONFIG_H */

