
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mndumpstate.c
 */
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#include "mnconfig.h"
#include "mip_icmp.h"	/* icmp router advert + appropriate mse extension/s */
#include "mnstate.h"
#include "udp_reg.h"
#include "faha_pinfo.h"

/*
 * print out current mn state
 */
dumpMnState(int longflag )
{
	char *s = NULL;
	switch(mns.state) {
	case NOWHERE:
		printf("mns.state: NOWHERE\n");
		break;
	case ATHOME:
		if (mns.retry_timer_flag) {
			printf("mns.state: ATHOME-PENDING\n");
		}
		else {
			printf("mns.state: ATHOME\n");
		}
		break;
	case AWAY:
		/* retry timer flag indicates that retry timer
		 * is underway, THEREFORE we have sent a request out
		 * but haven't got a reply back yet.  If we don't
 		 * get a reply we may just be talking to an agent
		 * that is cutoff from home
		 */
		if (mns.retry_timer_flag)
			if (mns.retryTimeouts)
				printf("mns.state: AWAY-PENDING-DISCONNECTED?\n");
			else
				printf("mns.state: AWAY-PENDING\n");
		else
			printf("mns.state: AWAY-ACKED\n");
		break;
#ifdef DHCP
        case CO_LOCATED:
	  if (mns.retry_timer_flag)

		if (mns.retryTimeouts)	    
			printf("mns.state: CO_LOCATED-PENDING-DISCONNECTED?\n");		  

		else
			printf("mn.state: CO_LOCATED-PENDING \n");		  
	  else
		if (mns.fr_number > 0)
			printf("mns.state: CO_LOCATED-ACKED/NAT\n");
		else
			printf("mns.state: CO_LOCATED-ACKED\n");
	  
	  break;
#endif
	}
#ifdef MADRP
	switch(mns.sr_state) {
	case NOWHERE:
		printf("mns.sr_state: NOWHERE\n");
		break;
	case ATHOME:
		if (mns.retry_timer_flag) {
			printf("mns.sr_state: ATHOME-PENDING\n");
		}
		else {
			printf("mns.sr_state: ATHOME\n");
		}
		break;
	case AWAY:
		if (mns.retry_timer_flag)
			if (mns.retryTimeouts)
				printf("mns.sr_state: AWAY-PENDING-DISCONNECTED?\n");
			else
				printf("mns.sr_state: AWAY-PENDING\n");
		else
			printf("mns.sr_state: AWAY-ACKED\n");
		break;
	}
#endif

	if (mns.fa.s_addr != 0) {
		printf("mns.fa: %s\n", inet_ntoa(mns.fa));
		printf("mns.coa: %s\n", inet_ntoa(mns.coa));
	}

#ifdef DHCP
        if (mns.auto_mode >= 0) {
		char *policy;
		
		switch(mns.dhcp_present_policy) {
		case IPIP:
			policy = "DHCP/IPIP";
			break;
		case NAT:
			policy = "DHCP/NAT";
			break;
		case NOTUNNEL:
			policy = "DHCP/NOTUNNEL";
			break;
		case AGENT_MODE:
			policy = "MIP/AGENT_MODE";
			break;
		default:
			policy = "DHCP/NOSTATE";
			break;
		}
                printf("DHCP: tunnel/mode %s, automatic: %s\n",
			policy, mns.auto_mode ? "yes" : "manual");
        }
#endif
	if (!longflag)
		return;
	printf("mns.haroute %d\n", mns.haroute);
	printf("mns.ident %x %x\n", mns.ident[0], mns.ident[1]);
	printf("mns.seqno %ld\n", mns.seqno);
	printf("mns.orig_lifetime %d\n", mns.orig_lifetime);
	printf("mns.agent_timer: %d\n", mns.agent_timer);
	printf("lifetime flag/timer %d %d\n", mns.lifetime_flag,
		mns.lifetime);
	printf("reg_timer flag/timer %d %d\n", mns.reg_timer_flag,
		mns.reg_timer);
	printf("retry_timer flag/timer %d %d\n", mns.retry_timer_flag,
		mns.retry_timer);
	printf("retry index value %d\n", mns.retryIndex);
	printf("retry timeout count since ack %d\n", mns.retryTimeouts);
	printf("hint_set flag/last addr %d %s\n", mns.hint_set, 
		inet_ntoa(mns.hintaddr));
	printf("docoa flag %d\n", mns.docoa);
}
