
/*
 *
 * Old test code.
 *
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mnicmp.c
 *
 * handle icmp input for mn
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#include "mnconfig.h"
#include "mip_icmp.h"	/* icmp router advert + appropriate mse extension/s */
#include "mnstate.h"
#include "mnstats.h"
#include "unixsock.h"
#include "mip_log.h"
#include "mndversion.h"
#include "alist.h"

struct mn_stats mn_stats;

extern int debug;

/* prototypes
*/
static int
icmpSecurityCheck(char *, struct mip_mse_icmp_auth *);

void
mnStateMachine(int udpSocket, struct alist *a);


static int beaconSd;

/*
 * initIcmp
 *
 * open raw icmp socket so that we can send and recv router adverts
 * on it.  should be able to send solicitations too.
 *
 * returns:
 *	int beaconSd - socket descriptor
 */ 
int
initIcmp(sawBcast, sawMcast)
int sawBcast;
int sawMcast;
{
	struct protoent *proto;
	int on;
	int rc;

	/* TBD. do something for multicast router advert
	*/
	if (!(proto = getprotobyname("icmp"))) {
 		syslog(LOG_ERR, "initIcmp(): unknown protocol icmp (%m).");

		exit(1);
	}
	/* icmp raw socket for sending out ICMP adverts
	*/
	if ((beaconSd = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) {
 		syslog(LOG_ERR, "initIcmp(): socket (%m)");
		exit(1);
	}
	/* set ROUTETOIF for this socket.  packets sent to i/f ip
	 * address are not routed but delivered directly
	 */
  	on = 1;
        rc = setsockopt(beaconSd, SOL_SOCKET, SO_DONTROUTE, (char *)&on,
                    sizeof(on));
	if ( rc < 0 ) {
 		syslog( LOG_ERR, "initIcmp(): setsockopt(SO_DONTROUTE) (%m)");
		exit(1);
	}
	/* so we can broadcast on this socket
	*/
        rc = setsockopt(beaconSd, SOL_SOCKET, SO_BROADCAST, (char *)&on,
                    sizeof(on));
	if ( rc < 0 ) {
		syslog( LOG_ERR, "initIcmp(): setsockopt(SO_BROADCAST) (%m)");
		exit(1);
	}
	/* so our broadcasts will be limited broadcast 255.255.255.255
	 * as specified by mobile-ip (there is a kernel mod associated 
	 * with this, it only affects "route to interface" code
	 */
        rc = setsockopt(beaconSd, SOL_SOCKET, SO_LIMITBCAST, (char *)&on,
                    sizeof(on));
	if ( rc < 0 ) {
		syslog( LOG_ERR, "initIcmp(): setsockopt(SO_LIMITBCAST) (%m)");
		exit(1);
	}

	return(beaconSd);
}


/* read icmp advert message
 * sanity check it
 * store important info
 * check state
 * if we decide 
 *	we are at HOME (changed base)
 *		send UDP home request to HA's ip address
 *			set newrequest timer to HA, value 
 *				so we resend if we never hear
 *	we are at new FA (changed base)
 *		note: 1st cut algorithm, we don't change base unless
 *			we lose the FA timeout wise
 *		deinstall default route
 *		install default route
 *		set re-registration timer
 *		send UDP home request to FA's ip address
 *			set newrequest timer to FA, value
 *
 * note: ip header arrives as 1st part of packet.
 */

handleIcmpAdvert(int icmpSd )
{
 	register struct icmp_advert *icp;
        register struct mip_mseicmp *mse;
	struct mip_mse_icmp_auth *mseauth;
	char location[257];
	struct ip *ipp;
        int rc = -1;
	int auth_ok = 0;	/* we got auth ext., and auth passed */
	unsigned long auth_timestamp = 0;
	int sawAuth = 0;
        int cc;
	int total;
        extern int debug;
        u_short flags = 0;
	struct sockaddr_in from;
	int fromlen;
	u_char packet[MAXICMPPACKET];
	/* info extracted from MSE
	*/
	int sawha = 0;
	int sawfa = 0;
	int rbit = 0;
	int faisha = 0;	/* TRUE if HA only adverts as FA, but we know better */
	u_short mse_sequence;
	u_short mse_lifetime;
	int advert_length;
	int hlen;	/* ip header length */
	int i;
	struct in_addr mse_coa;
	int size, lsize;
	char *l = NULL;

	lsize = 0;
	mse_coa.s_addr = 0;
	fromlen = sizeof(from);
	/* read the icmp packet. note ip header is in there
	*/
	if ((cc = recvfrom(icmpSd, (char *)packet, MAXICMPPACKET, 0,
		    (struct sockaddr *)&from, &fromlen)) < 0) {
		if (errno == EINTR)
			return;
		mip_log( MLOG_ERR, "handleIcmpAdvert(): recvfrom failed");
		if (debug)
			perror("mnd: icmp recvfrom");
		return;
	}
	total = cc;

	/* sanity checks *******************************
	*/
	ipp = (struct ip*) packet;
	hlen = (int) (ipp->ip_hl << 2);
        if (cc < hlen + ICMP_MINLEN) {
		mn_stats.icmptooshort++;
		mip_log( MLOG_ERR, "handleIcmpAdvert(): icmp packet from %s less than min icmp length\n", inet_ntoa(from.sin_addr));
		if (debug) {
			printf("icmp packet less than min icmp length\n");
		}
                return;
        }
	cc -= hlen;	/* throw away icmp header */
	/* pointer to router advert body */
	icp = (struct icmp_advert *) ((int)ipp + hlen);

	/* we care about only router adverts. not worth logging otherwise
	*/
	if (icp->icmp_type != ICMP_ROUTERADVERT)  {
		return;
	}

	/* size of router advert, basic header + variable length
	 * ip address part at end
	 */
	advert_length = sizeof(struct icmp_advert) + (8 * (icp->icmp_naddrs-1));

	/* check size, invalid mobile ip packet if too small 
	 */
	if (cc <= advert_length ) {
		mn_stats.icmptooshort++;
		mip_log( MLOG_ERR, "handleIcmpAdvert(): icmp message from %s too short\n", inet_ntoa(from.sin_addr));
		if (debug) {
			printf("icmp message too short\n");
		}
		return;
	}

	/* now make mse checks, general size check first, include mse
	 * acc. to draft mobile ip mse must come before other extensions
	*/
	if (cc < (advert_length + (sizeof(struct mip_mseicmp) - 4))) { 
		mn_stats.msetooshort++;
		mip_log( MLOG_ERR, "handleIcmpAdvert(): icmp+mse message from %s too short\n", inet_ntoa(from.sin_addr));
		if (debug) {
			printf("icmp+mse message too short\n");
		}
		return;
	}

	/* ok we can't commit pointer suicide, now check type and length
	 */
	mse = (struct mip_mseicmp *) ((int)icp + advert_length);

	if (mse->mse_type != MIPAGENTAD) {
		mn_stats.invalidicmpmse++;
		mip_log( MLOG_ERR, "handleIcmpAdvert(): icmp packet from %s, invalid mse type in icmp advert\n", inet_ntoa(from.sin_addr));
		if (debug) {
			printf("invalid mse type in icmp advert\n");
		}
		return;
	}
	
	/* check flags */
	
	flags = ntohs(mse->mse_flags);
	/* came from FA */
	if (TST(flags, FABIT)) {
		struct in_addr router = *((struct in_addr *) &icp->icmp_router);
		if (router.s_addr == mn_config.ha.s_addr) {
			faisha = 1;
		}
		else {
			sawfa = 1;
			if (mse->mse_len < 10) {
				/* invalid mse++ */
				mn_stats.msetooshort++;
				mip_log( MLOG_ERR,
				    "handleIcmpAdvert(): mse from FA %s too short",
				    inet_ntoa(from.sin_addr));
				return;
			}
			if (TST(flags, RBIT)) {
				rbit = 1;
			}
			mse_coa.s_addr = mse->mse_coa;
			size = sizeof(struct mip_mseicmp);
		}
	}
	/* came from HA */
	if (TST(flags, HABIT) || faisha) {
		sawha = 1;
		if (mse->mse_len < 6) {
			/* invalid mse++ */
			mn_stats.msetooshort++;
			mip_log( MLOG_ERR, "handleIcmpAdvert(): mse from HA %s too short", inet_ntoa(from.sin_addr));
			return;
		}
		size = sizeof(struct mip_mseicmp) - 4;
	}
	if (sawfa && sawha) {
		/* icmp router advert, both ha, fa error */
		if (debug) {
			printf("advert claims BOTH ha/fa \n");
		}
		mn_stats.bothhafa++;
		mip_log( MLOG_ERR, "handleIcmpAdvert(): icmp routing advertisement claims BOTH ha/fa - invalid \n");
		return;
	}
	/* ignore packet if not ha or not fa
	*/
	if (sawfa == 0 && sawha == 0) {
		return;
	}
	mse_sequence = ntohs(mse->mse_sequence);
	mse_lifetime = ntohs(mse->mse_lifetime);

	/* now deal with optional parts if any
	*/
	cc = cc - (advert_length + size);
	mse = (struct mip_mseicmp *) (((char *)mse)  + size);

	/* unlike the draft says, we don't discard messages
	 * with types in range 0..127, we just ignore the extension
	 */
	while(cc >= 2) {
		int len;
		switch(mse->mse_type) {
		case 0:  /* one byte pad, should be end */
			cc = 0;
			break;
		case MIPAGENTAUTH:
			icp->icmp_cksum = 0;
			sawAuth = 1;
			mseauth = (struct mip_mse_icmp_auth *) mse;
			cc = 0;		/* last by definition */
			auth_ok = icmpSecurityCheck(packet + hlen, mseauth);
			auth_timestamp = mseauth->timestamp;
			break;
		case MIPAGENTINFO:
			lsize = (int) mse->mse_len & 0xff;
			/* put in location array and make sure it is
			 * null terminated
			 */
			strncpy(location, ((char *)mse) + 2, lsize);
			location[lsize] = 0;
			l = (char *)location;
			break;
		case 19: /* prefix message, ignore it */
		default:
			break;
		}
		/* advance pointer */
		if (cc) {
			cc -= (mse->mse_len + 2);
			mse = (struct mip_mseicmp *) (((char *)mse) + 
				(((int) mse->mse_len + 2) & 0xff));
		}
	}


    	/*  be sure that any beacon with the HA bit set is our HA.
	 *  if security checks are made, this should have failed above.
	 */
	if (sawha) {
		struct in_addr router = *((struct in_addr *) &icp->icmp_router);
		if (router.s_addr != mn_config.ha.s_addr) {
			if (debug) {
				char buf[1024];
				strcpy(buf, inet_ntoa(router));
				printf("ignoring HA beacon, my ha is %s, beacon is from %s\n", inet_ntoa(mn_config.ha), buf);
			}
			return;
		}
	}


	/* dump mse information
	*/
	if (debug) {
		printf("icmp packet:\n");
		dumpme(packet, total);	
		if (sawha) {
			printf("HA(%s): icmp_router_advert/mse, sequence %u, lifetime %u%s\n", 
				inet_ntoa(from.sin_addr),
				mse_sequence, mse_lifetime,
				faisha?" (HA advertising as FA)":""
				);
		}
		else {
			char buf[1024];
			strcpy(buf, inet_ntoa(mse_coa));
			printf("FA(%s): icmp_router_advert/mse, sequence %u, lifetime %u, coa %s\n", 
				inet_ntoa(from.sin_addr),
				mse_sequence, mse_lifetime, 
				buf
				);
		}
		if (sawAuth) {
			printf("authentication!: passed %d timestamp %x\n",
				auth_ok, auth_timestamp);
		}
		if (l) {
			printf("location <%s>\n", l);
		}
	}

	/* at this point:
	 * icp->icmp_router - address to use for default route
	 * mse_sequence - current sequence number
	 * sawha/sawfa - kind of agent
	 * mse_lifetime - lifetime before reregistration
	 * mse_coa - coa from fa
	 * sawAuth - true if authentication included
 	 * auth_ok - true if authentication verified
	 * auth_timestamp - value for auth timestamp in secs
	 * location - pointer to location, NULL if none.
	 * lsize - size of location 
	 */

        /* take this beacon and see if it is in the alist
         * if it is not, we will store it in there
         */
        checkAlist(sawha, *((struct in_addr *) &icp->icmp_router),
		mse_sequence, mse_lifetime, mse_coa, sawAuth,
		auth_ok, auth_timestamp, l, lsize);


	return(0);
}

/*
 * check icmp security if it shows up
 *
 * we don't log anything here.  we do however keep
 * totals in mn_stats
 *
 * return:
 *	1: auth ok
 *	0: failed authentication.  That means we got an authentication
 *	but the hash recompute failed
 */
static
int
icmpSecurityCheck(char *packet, struct mip_mse_icmp_auth *mseauth)
{
	int spi = mseauth->spi;
	unsigned int timestamp = mseauth->timestamp;
	unsigned char hash[16];
	extern int debug;
	int size;

	/* sanity check */
	if (mseauth == NULL)
		return (1);

	/* if we insist on optional agent security,
	 * then we check the timestamp.  we have to put
	 * off checking for freshness until later
	 * as we need a previous value in agent list
	 */
	if (mn_config.agent_security) {
		size = ((char *)mseauth) - packet;
		size += 8;
		/* check hash first */
		/* TBD. for now spi is ignored
                */
                spi = htonl(mn_config.sa[AGENTSECURITY].spi);
                /* recompute MD5 hash, result in local hash[]
                */ 
                computeMd5(packet, size,
                        mn_config.sa[AGENTSECURITY].key,
                        hash);

		 if (bcmp(mseauth->hash, hash, 16) != 0) {
                        if (debug) {
                                printf("mn agent authentication failed\n");
                        }
                        mn_stats.invalidicmpauth++;
                        /* authentication failed
                        */
                        return(0);
                }
		if (debug) {
			printf("mn agent authentication ok\n");
		}
	}
	return(1);
}
