
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mn_state.h
 */

#include <sys/time.h>
#ifdef DHCP
#include "keylist.h"
#include "route.h"
#endif
 
#ifdef DHCP
#define AGENT_MODE      0 
#define IPIP            1    
#define NAT             2
#define NOTUNNEL        3
#define NOSTATE         4   
#endif

/* fundamental states for mnd state machine
*/
#define NOWHERE 	     0	/* not home or away - lonely, in fact */
#define ATHOME		     1	/* at HA */
#define AWAY		     2	/* at FA */
#ifdef DHCP
#define CO_LOCATED           3       /* NOWHERE first, then DHCP worked */
#endif

#define TICK_EVENT 0
#define ADVERT_EVENT 1

/* retry array
*/
#define NORETRIES	10
#define SHORT_RETRY	(NORETRIES / 3)	    /* try another agent instead
					       of exhausting backoff cycle */
#define TYPE_ALIST 0
#define TYPE_MNLIST 1
#define TYPE_HLIST 2
#define PUBLIC   1
#define PRIVATE  2

/* time to go to NOWHERE state ==
 *	HA: time above + total times below in retryBackoff
 *	FA: 1/3 of lifetime + total times below
 */
/* total = 142 seconds or two minutes worth (roughly)
 * idea here is that we try a couple of times in case
 * packets or lost but we backoff rather steeply
 * so that we have time to get through if links nearby,
 * wireless or serial are congested.  
 */
static
int retryBackoff[NORETRIES] = {
        2,4,8,8,8,16,16,16,32,32
};

struct mn_state {
	int state;			/* current mn state */
#ifdef MADRP
	int sr_state;
#endif

	/* if following are non-zero,
	 * indicates that fa route is set
	 */
	struct in_addr fa;		/* ip addr to talk to fa */
	struct in_addr coa;		/* coa we tell ha */
	int haroute;			/* binary - if 1, ha route set */
	int cloneroute;			/* binary - if 1, clone route exists */
	unsigned long ident[2];		/* current identity */
	unsigned int seqno;		/* sequence number of agent */
	int dhcp_present_policy;  

        int fr_number;                 /* Store the firewall rules numbers Here */ 
	/* lifetime sent to us by ha or fa - life in visitor list
	*/
	int lifetime_flag;
	int lifetime;			
	int orig_lifetime;		/* last lifetime, not decremented */
 
	/* timer for resending registration at some fraction of time
	 * smaller than lifetime
	 */
	int reg_timer_flag;
	int reg_timer;			/* next registration timer */

	/* timer for resending registration packets
	*/
	int retry_timer_flag;
	int retry_timer;		
	int retryIndex;		/* position in retry cycle array */

	/* set to 0 if ACK is recv.
	*/
	int retryTimeouts;      /* # of times retry cycle timed out */

	struct timeval sendtime;	/* time we sent packet */ 
	struct timeval recvtime;	/* time we got packet back */
	int halifetime;			/* timer value for rereg at HOME */
	int docoa;		/* TBD. true if we should be our own FA */


	/* pointer to current agent (registered to or registering with...)
	 * and a timer until agent should be checked against the best
	 * possible choice
	 *
	 * state condition: if mns.state is NOWHERE then mns.agent == NULL, if
	 * mns.state is not NOWHERE, mns.agent is defined.
	 */
	struct alist	*agent;
	int		agent_timer;

	/* 
	 * A hint may be offered us by mnstat and
	 * consists of a user desired agent ip address.
	 * If hintaddr can be found as a bip or an coa in the agentlist
	 * we will switch from our current agent to that agent.
	 * This gives us a way to allow the user to determine the agent.
	 * The hinted agent, once found, is kept so long as the agent
	 * appears in the agent list, and then until a full retry cycle
	 * is attempted if it becomes unreachable, at which point the hint
	 * request is turned off.  The hint may be turned
	 * off at any time at which point the optimal agent is chosen
	 * from the alist as in the default case.
	 */
	int		hint_flag;	/* TRUE if hint is on */
	int		hint_set;	/* TRUE if hintaddr has been matched
					   with an agent in the alist */
	struct in_addr	hintaddr;

#ifdef DHCP
        int             mode;           /* either AGENT_MODE or DHCP_MODE */
        int             auto_mode;      /* automatic or manual mode switching */
#endif
};

extern struct mn_state mns;
