/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef IPSEC_H
#define IPSEC_H

/*
 * IPSEC policy types
 */

/* magic values and macros */
#define MINSPI	256
#define ROUNDUP(a) \
  ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n) \
    { x += ROUNDUP(n); }


/* mnd ipsec keylist type index */
#define MN_SECTYPE_ATHA		0
#define MN_SECTYPE_BYPASSFA     1
#define MN_SECTYPE_ATFA		2
#define MN_SECTYPE_MN		3
#define MN_SECTYPE_MAX		4

/* mipd ipsec keylist type index */
#define HA_SECTYPE_MN		0
#define FA_SECTYPE_MN		1
#define TUN_SECTYPE             2
#define MIP_SECTYPE_MAX		3

/*
 * security types
 */
#define	NOSEC		0x00
#define AHSEC		0x01	
#define ESPSEC		0x02

/*
 * prototypes
 */
void openKeySocket(void);
void closeKeySocket(void);

int verifyKey(struct in_addr src, struct in_addr dst,
					u_long spi, int isah, int outbound);

static void createKeyMsg(struct in_addr srcAddr, struct in_addr dstAddr,
		u_long spi, int transform);

static int keyMsg(int msgtype);
#endif    /* IPSEC_H */
