/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * route.h
 */


#ifndef ROUTE_H
#define ROUTE_H

/* prototypes */

/* agent (or mn) add mn as link-layer route
*/
int routeInit();
void routeRead();
void route_add_link(struct in_addr mn, char *mac, char *ifname
	);
void route_delete_link(struct in_addr mn, char *mac);

void route_add_proxy(struct in_addr mn, char *mac, char *ifname);
void route_delete_proxy(struct in_addr mn);

void route_add_host(struct in_addr, struct in_addr, struct in_addr
	);
void route_delete_host(struct in_addr, struct in_addr);
void route_delete_network(struct in_addr, struct in_addr, struct in_addr);

/* ha tunnel routes */
void ha_route_add_tunnel(struct in_addr mn, struct in_addr coa
	);
void ha_route_delete_tunnel(struct in_addr mn, struct in_addr coa);

/* mn to agent as default router */
void mn_route_add_default(struct in_addr agent, struct in_addr interface,
	int dolink,  char *mac, char *ifname );

#ifdef DHCP
void mn_route_add_tunnel(struct in_addr, struct in_addr, char *);
#endif

void mn_route_delete_default(struct in_addr router, int dolink);

/* new, only used with ad hoc routing
*/
void clone_route_add( struct in_addr cloneroute, struct in_addr interface, 
	struct in_addr subnet );

void reject_route_add( struct in_addr cloneroute, struct in_addr interface, 
	struct in_addr subnet);

void clone_route_delete(struct in_addr cloneaddr);

void flushArp(void);

/* imported by route.h from *config.c */
int noDefaultRoute();

#endif

