/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* setarp.c
 *
 * module turns arp flag on/off on a given interface
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <net/if_var.h>
#include <net/if_dl.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <arpa/inet.h>

#ifdef NOPE
#define	NSIP
#include <netns/ns.h>
#include <netns/ns_if.h>
#endif
#include <netdb.h>

#include <sys/protosw.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <nlist.h>
#include <kvm.h>
#include <fcntl.h>

struct	ifreq		ifr, ridreq;

static int	flags;
static int	metric;
static int	mtu;
static int	nsellength = 1;
static int	setaddr;
static int	setipdst;
static int	doalias;
static int	newaddr = 1;
static kvm_t	*kvmd;
extern	int errno;


#define	NEXTARG		0xffffff

void
set_arp(ifname, vname, value)
	char *ifname;
	char *vname;
	short value;
{
	int sd;

	sd = socket(AF_INET, SOCK_DGRAM, 0);
	if (sd<0) {
		perror("switching arp: socket");
		exit(1);
	}
	strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
	ifr.ifr_addr.sa_family = AF_INET;
 	if (ioctl(sd, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) {
 		perror("ioctl (SIOCGIFFLAGS)");
 		exit(1);
 	}
	strncpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
 	flags = ifr.ifr_flags;

	if (value < 0) {
		value = -value;
		flags &= ~value;
	} else
		flags |= value;
	ifr.ifr_flags = flags;
	if (ioctl(sd, SIOCSIFFLAGS, (caddr_t)&ifr) < 0)
		perror(vname);
	close(sd);
}

#ifdef TESTMAIN
main(int argc, char *argv[])
{
	if (argc!=3) {
		fprintf(stderr, "usage: turn i/f <arp>|<-arp>\n");
		exit(1);
	}
	if (strcmp(argv[2], "arp")==0)
		set_arp(argv[1], "arp", -128);
	else if (strcmp(argv[2], "-arp")==0)
		set_arp(argv[1], "-arp", 128);
	else {
		fprintf(stderr, "usage: turn i/f <arp>|<-arp>\n");
		exit(1);
	}
}
#endif
