
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * udprequest.c
 * 
 * test code - not part of normal setup 
 *
 * XXX: warning. still need to fix to use RECVIF. jrb
 *
 * static mobile node request (assumption is that timeout is
 *	infinite or this can just be used for testing)
 * sends one request packet, reads back reply.
 * if -d debug, will print that out verbosely.
 *
 * syntax:
 *
 *	udprequest -? - list commands
 *		   -d verbose debug
 * 	udprequest home - deregister
 *		   sapho - register at 113D FA
 *		   jimhome - register at jrb's residence
 *		   machine - register at machine room
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/times.h>
#include <errno.h>
#include <signal.h>

#include "mipconfig.h"
#include "udp_reg.h"

/* used to test IP_RECVIF IPPROTO_IP level sockopt */
#define LOCALHOST	"127.0.0.1"
#define SWIFTSURE	"10.0.0.25"
#define MYUDP		5001
#define THEIRUDP	5001

/*
 * mn is self 
 */
/* mn - name of self */
#define SELF "204.203.71.203"

/* ha */
#define GUINNESS "204.203.64.220"	/* wired address */
#define GUINNESSWL "204.203.71.254"     /* wireless address */

/* possible fa coas - wired ip addreses for fas */
#define DREADNOUGHT "204.203.64.228"
#define HOMEBREW    "204.203.64.211"
#define PRINCE      "204.203.64.28"
#define DEFENCE     "204.203.64.176"
/*
#define NEPTUNE     "not real yet"
*/

/* fa wavelan addresses (private) */
#define DREADNOUGHTFA "10.0.0.1"  /* jim's psu office */
#define PRINCEFA "10.0.0.2"       /* pcat machine room */
#define DEFENCEFA "10.0.0.3"      /* sapho 113D */
#define NEPTUNEFA "10.0.0.4"      /* nowhere */
#define HOMEBREWFA "10.0.0.10"    /* jim home */

static int udpSocket;
static int debug = 0;

static struct in_addr mn, ha, fa, coa;

/* write a MOCK registration packet to a given ip
 * address to port 434
 */
sendRegistration(mn, ha, fa, coa)
struct in_addr mn, ha, fa, coa;
{
	int rc;
	int size;
	struct sockaddr_in to;
	int s = initUdp();
	char packet[1024];
	int i;
	struct udp_register *reg;
	struct mse_authext *mse;
	unsigned long keys[4];
	int spi;

	udpSocket = s;

	reg = (struct udp_register *) packet;
	mse = (struct mse_authext *) (packet + sizeof (struct udp_register));

	/* set up dummy registration packet
	*/
	reg->type = UDP_REGISTER_TYPE;
	reg->flags = 0;
	reg->lifetime = htons(180);
	/* if deregistration
	*/
	if (coa.s_addr == 0)
		reg->lifetime = htons(0);
	/* my mn/ip address */
	reg->home_address.s_addr = mn.s_addr;
	/* my ha's ip address */
	reg->home_agent.s_addr = ha.s_addr;
	reg->coa.s_addr = coa.s_addr;
	reg->i.l[0] = htonl(1);
	reg->i.l[1] = htonl(2);

	mse->type = MN_HA_AUTHENTICATOR_TYPE;
	/* 16 is length of hash, 4 is spi */
	mse->length = 4 + 16;
	
	spi = htonl(256);
	bcopy(&spi, &mse->spi, 4);

	/* mock key */
	keys[0] = 0x01020304;
	keys[1] = 0x05060708;
	keys[2] = 0x09101112;
	keys[3] = 0x13141516;

	bcopy(keys, mse->hash, 16);

	/* send the packet to the  udp port THEIRUDP
	*/
        to.sin_family = AF_INET;
        to.sin_addr.s_addr = fa.s_addr;
	to.sin_port = htons(THEIRUDP);

	if (debug) {
		dumpUdpRequest(reg);
		dumpMse(mse);
	}
	rc = sendto(s, packet, 
		sizeof(struct udp_register) + sizeof(struct mse_authext),
		0,  (struct sockaddr *) &to, 
		sizeof (struct sockaddr_in));

	if (rc < 0) {
		perror("sendto");
		exit(1);
	}
	if (!debug) {
		printf("ack sent\n");
		return;
	}

	/* dump the packet */
	{ 
		int i = 0;
		char *p = packet;
		for ( ; i < rc; i++) {
			if ( (i != 0) && ((i % 16) == 0)) {
				printf("\n");
			}
			printf("%2.2x ", *p++ & 0xff);
		}
		printf("\n");
	}
}

/* resolve dns name or ip address as string to net order
 * socket "long"
 * 	returns: 0 - name could not be resolved
 *		 1 - ok
 */
int
resolveAddress(target, to)
char *target;		
struct sockaddr_in *to;
{
	struct hostent *hp;

	to->sin_family = AF_INET;
	to->sin_addr.s_addr = inet_addr(target);
	if (to->sin_addr.s_addr != (u_int)-1) {
		/*
		hostname = target;
		*/
		;
	}
	else {
		hp = gethostbyname(target);
		if (!hp)
			return(0);
		to->sin_family = hp->h_addrtype;
		bcopy(hp->h_addr, (caddr_t)&to->sin_addr, hp->h_length);
	}
	return(1);
}

int
initUdp()
{
	int s;
	struct sockaddr_in server;
	static	beenhere = 0;

	if (beenhere) 
	    return(udpSocket);


	/* create INTERNET, udp datagram socket
	*/
	s = socket(AF_INET, SOCK_DGRAM, 0); 
	if ( s < 0 ) {
		perror("initUdpServer: socket");
		exit(1);
	}
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = INADDR_ANY;   /* ok from any machine */
	/*
	server.sin_port = htons(UDP_PORT);     /* specific port * /
	*/
	server.sin_port = htons(MYUDP);     /* specific port */

	/* bind protocol to socket
	*/
	if (bind(s, (struct sockaddr *)&server, sizeof(server))) {
		perror("bind");
		exit(1);
	}
	beenhere = 1;
	return(s);
}


recvRegistration()
{
	int rc;
	int size;
	char packet[1024];
	struct udp_reply *rep;
	struct mse_authext *mse;
	struct sockaddr_in from;

	bzero( packet, 1024 );

	/* read the packet back from the udp port 434
	*/
	size = sizeof(struct sockaddr_in);
	rc = recvfrom(udpSocket, packet, 1024, 0, 
		(struct sockaddr *)&from, &size);
	if (rc < 0) {
		perror("recvfrom");
		exit(1);
	}
	printf("reply from addr=%s, port=%d, rc(size)=%d\n", 
		inet_ntoa(from.sin_addr.s_addr),
		ntohs(from.sin_port), rc );
	rep = (struct udp_reply *) packet;
	mse = (struct mse_authext *) (packet + sizeof (struct udp_reply));

	if (debug) {
		/* 
		dumpUdpReply(rep);
		dumpMse(mse);
		*/
		dumpme( packet, rc );
	}
}


void
parseArgs(argc, argv)
int             argc;
char          **argv;
{
	extern int      errno;
	extern char    *optarg;
	extern int      optind;
	int             ch, fd;
	int 		rc;
	int 		sawFlag = 0;
	char		*where;

	if (argc < 2) {
		usage();
	}
	while ((ch = getopt(argc, argv, "d")) != EOF) {
		switch (ch) {
		case 'd':
			debug = 1;
			break;
		case '?':
		default:
			usage();
		}
	}
	where = argv[optind];
	if (strcmp(where, "home") == 0) {
		coa.s_addr = htonl(0);
		/* fa is destination */
		fa.s_addr = inet_addr(GUINNESS);
	}
	else if (strcmp(where, "jimhome") == 0) {
		coa.s_addr = inet_addr(HOMEBREW);
		fa.s_addr = inet_addr(HOMEBREWFA);
	}
	else if (strcmp(where, "sapho") == 0) {
		coa.s_addr = inet_addr(DEFENCE);
		fa.s_addr = inet_addr(DEFENCEFA);
	}
	else if (strcmp(where, "machine") == 0) {
		coa.s_addr = inet_addr(PRINCE);
		fa.s_addr = inet_addr(PRINCEFA);
	}
	else if (strcmp(where, "local") == 0) {
		coa.s_addr = inet_addr(LOCALHOST)+100;
		fa.s_addr = inet_addr(LOCALHOST);
	}
	else {
		printf("sorry - don't know that location\n");
		usage();
	}
}

usage()
{
	fprintf(stderr,"udprequest: [-d] <location>\n");
	fprintf(stderr,"-d: verbose request/reply info\n");
	fprintf(stderr,"locations include: \n");
	fprintf(stderr,"\thome - deregister near guinness: \n");
	fprintf(stderr,"\tjimhome - fa at jim's home: \n");
	fprintf(stderr,"\tsapho - fa at 113D: \n");
	fprintf(stderr,"\tmachine - fa in machine room: \n");
	exit(1);
}


main(int argc, char **argv)
{
	int	count,
		on,
		rc,
		len;
	struct sockaddr_in whoto;

	/* name of self
	*/
	mn.s_addr = inet_addr(LOCALHOST);
	/* name of HA
	*/
	ha.s_addr = inet_addr(DEFENCEFA);

	parseArgs(argc, argv);

	/* before sockopt */
	sendRegistration(mn, ha, fa, coa);
	recvRegistration();

	on = 0;
	len = sizeof(int);
	rc = getsockopt(udpSocket, IPPROTO_IP, IP_RECVIF, (int *)&on,
		&len);
	if ( rc ) 
	    printf("getsockopt() failed: rc=%d\n", rc);
	else
	    printf("getsockopt(): on=%d (should be 0)\n", on);

	on = 1;
	rc = setsockopt(udpSocket, IPPROTO_IP, IP_RECVIF, (int *)&on,
		sizeof(int));
	if ( rc ) 
	    printf("setsockopt() failed: rc=%d\n", rc);

	on = 0;
	len = sizeof(int);
	rc = getsockopt(udpSocket, IPPROTO_IP, IP_RECVIF, (int *)&on,
		&len);
	if ( rc ) 
	    printf("getsockopt() failed: rc=%d\n", rc);
	else
	    printf("getsockopt(): on=%d (should be 1)\n", on);

	/* after sockopt turned on */
	sendRegistration(mn, ha, fa, coa);
	recvRegistration();

	on = 0;
	rc = setsockopt(udpSocket, IPPROTO_IP, IP_RECVIF, (int *)&on,
		sizeof(int));
	if ( rc ) 
	    printf("setsockopt() failed: rc=%d\n", rc);

	on = 0;
	len = sizeof(int);
	rc = getsockopt(udpSocket, IPPROTO_IP, IP_RECVIF, (int *)&on,
		&len);
	if ( rc ) 
	    printf("getsockopt() failed: rc=%d\n", rc);
	else
	    printf("getsockopt(): on=%d (should be 0)\n", on);

	/* after sockopt turned off */
	sendRegistration(mn, ha, fa, coa);
	recvRegistration();
}


