
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/stat.h>

/* mipd - mipstat socket names */
#define MOBYSOCK_SERVER "/tmp/mipd.sock.serv"
#define MOBYSOCK_CLIENT "/tmp/mipd.sock.client"

/* mipd commands */
#define U_STAT    1	/* send over stat structures */
#define U_FAVLIST 2	/* write fa vlist over */
#define U_HAVLIST 3	/* write ha vlist over */
#define U_BEACONS 4	/* send icmp_stats structure */

/* mnd-mnstat socket names and file mode */ 
#define MNDSOCK_SERVER	"/tmp/mnd.sock.serv"
#define MNDSOCK_CLIENT	"/tmp/mnd.sock.client"
#define XMNDSOCK_CLIENT	"/tmp/xmnd.sock.client"
#define US_MODE		(S_IRUSR | S_IWUSR)

#define U_MNSTAT  1	/* send of mn_stat structure */
#define U_MNSTATE 2	/* send over mn state structure */
#define U_MNHINT  3 	/* send over hint to mn about agent ip */
#define U_ALIST   4     /* mnd sends over agent list entry by entry */
#define X_MNSTAT  5	/* send mn_stat structure to xmnstat */
#define X_MNSTATE 6	/* send mn state structure to xmnstat */
#define X_ALIST   7	/* send alist over to xmnstat */
#define U_MNBEACONS 8	/* send icmp_stats structure */
#define X_MNBEACONS 9
#define U_MNLIST 10	/* send mnlist to mnstat */
#define X_MNLIST 11	/* send mnlist to xmnstat */
#define AUTHCHANGEKEY 12	/* change auth key value */
#define AUTHPOLICY 13	/* toggle accept/reject for netauth or adhoc */
#define LINKMODE 14	/* mode switch: arp <--> auth */
#define GETMNDPID 15	/* send mnd pid to xmnstat */
#define WRITECONF 16    /* write out the config file */ 
#ifdef MADRP	
#define U_HLIST   18	/* send hlist over to mnstat */
#endif
#ifdef DHCP
#define U_MODESWITCH    19      /* set manual or automatic modes */
#endif

/* key change types */
#define CHANGE_MNHA 0
#define CHANGE_FAMN 1
#define CHANGE_NETAUTH 2
#define CHANGE_ADHOC 3

/* structure for changing auth keys on the fly from xmnstat
*/
struct changekey {
	unsigned char change_type;
	unsigned char key[16];
};

/* auth policy change types */
#define ACCEPT_NETAUTH 0
#define REJECT_NETAUTH 1
#define ACCEPT_ADHOC 2
#define REJECT_ADHOC 3

/* link mode change types */
#define ARP_TO_AUTH 0
#define AUTH_TO_ARP 1


/* structure passed from mnstat to mnd for U_MNHINT message
*/
struct hintmsg {
	unsigned char code;
	unsigned char ip[4];
};

/* prototypes */
int initUnixSocket(char *whoami, int bepolite);
int sendUnixMessage(char *sockfile, char *buf, int size);
int recvUnixMessage(char *buf, int bufsize);

