
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * utils.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

int resolveAddress(char *, struct sockaddr_in *);
int pack_string(char *, unsigned char *, int);

/* resolve dns name or ip address as string to net order
 * socket "long"
 * 	returns: 0 - name could not be resolved
 *		 1
 */
int
resolveAddress(target, to)
char *target;		
struct sockaddr_in *to;
{
	struct hostent *hp;

	to->sin_family = AF_INET;
	to->sin_addr.s_addr = inet_addr(target);
	if (to->sin_addr.s_addr != (u_int)-1) {
		/*
		hostname = target;
		*/
		;
	}
	else {
		hp = gethostbyname(target);
		if (!hp)
			return(0);
		to->sin_family = hp->h_addrtype;
		bcopy(hp->h_addr, (caddr_t)&to->sin_addr, hp->h_length);
	}
	return(1);
}

char *
reverseAddress(struct in_addr in)
{
	struct hostent *hp;
	int	logIPOnly;

	logIPOnly = getConfiglogIPOnly();
	if (logIPOnly) {
		return(inet_ntoa(in));
	}

	/* turn off our timers because gethostbyaddr(3) is
	 * screwed up in 2.2.  This is unfortunate but we can't
	 * afford to hang here on a permanent basis.
	 */
	stopAlarm();
	hp = gethostbyaddr((caddr_t)&in, sizeof (struct in_addr), AF_INET);
	startAlarm();
	if (hp)
		return(hp->h_name);
	else {
		return("no name found");
	}
}

/* reverse the address unconditionally, to be used by access_control at HA */

char *
reverseAddress1(struct in_addr in)
{
	struct hostent *hp;

	/* turn off our timers because gethostbyaddr(3) is
	 * screwed up in 2.2.  This is unfortunate but we can't
	 * afford to hang here on a permanent basis.
	 */
	stopAlarm();
	hp = gethostbyaddr((caddr_t)&in, sizeof (struct in_addr), AF_INET);
	startAlarm();
	if (hp)
		return(hp->h_name);
	else
		return(inet_ntoa(in));
}

/*
 * pack_string
 *
 * input: pointer to hex string in nibbles; e.g.,
 *	0xdeadbeef01234567
 *
 * output: pack into bytes.
 *
 * len: length of input in bytes. output should be at least
 *	half the size of input.  May include 0x as prefix.
 *
 * returns:
 *	0: ok.
 *     -1: bad hex value on input. 
 */
int
pack_string(char *input, unsigned char *output, int len)
{
	register char *p = input;
	register unsigned char *o = output;
	int i;
	int cur_int;
	
	/* cycle input past hex prefix if any
	*/
	if (*p == '0' && (*(p + 1) == 'x' || *(p + 1) == 'X')) {
		p += 2;
		len -= 2;
	}
	len = len >> 1;

	/* loop thru input string, packing nibbles into bytes
	*/
	for (i=0; i < len; i++) {
		cur_int = 0;
		/* put p into upper nibble
		*/
		if ((*p >= '0') && (*p <= '9'))
			cur_int = (*p - '0') << 4;
		else if ((*p >= 'a') && (*p <= 'f'))
			cur_int = (*p - 'a' + 10) << 4;
		else if ((*p >= 'A') && (*p <= 'F'))
			cur_int = (*p - 'A' + 10) << 4;
		else {
			return(-1);
		}
		/* put p into lower nibble
		*/
		p++;
		if ((*p >= '0') && (*p <= '9'))
			cur_int |= (*p - '0');
		else if ((*p >= 'a') && (*p <= 'f'))
			cur_int |= (*p - 'a' + 10);
		else if ((*p >= 'A') && (*p <= 'F'))
			cur_int |= (*p - 'A' + 10);
		else {
			return(-1);
		}
		*o++ = cur_int & 0xff;
		p++;
	}
	return(0);
}



/*
 * extractString linebuf - a string of form "...." output - same string sans
 * quotes max - max size of output buffer
 * 
 * we want to extract the string and toss the quotes
 * 
 * returns: -1: failed 0: ok
 */
int
extractString(char *linebuf, char *output, int max)
{
	int             i = 0;
	int             l = strlen(linebuf);
	int             found = 0;
	char           *p = linebuf;

	if ((l - 2) >= max)
		return (-1);

	/*
	 * skip ahead to 1st double quote
	 */
	while (l--) {
		if (*p == '"') {
			found = 1;
			break;
		}
		p++;
	}
	if (!found) {
		return (-1);
	}
	found = 0;
	/*
	 * copy string into "value" buffer, leave out double quotes
	 */
	p++;
	l--;
	if (l == 0)
		return (-1);
	while (l--) {
		*output++ = *p++;
		/*
		 * terminate on double quote.
		 */
		if (*p == '"') {
			*output = 0;
			found = 1;
			break;
		}
	}
	if (!found) {
		return (-1);
	}
	return (0);
}

#ifdef TESTMAIN
main()
{
	struct in_addr target;

	target.s_addr = inet_addr("204.203.71.254");
	printf("%s\n", reverseAddress(target));
}
#endif
