
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define FALSE	0
#define TRUE	1
#include "route.h"

/* 
 * fa/ha visitor list
 */
struct vlist {
	int vstate;			/* per mn state */
	int fr_number;			/* divert firewall rule number */
	struct in_addr i_f;  		/* link mn is resident on */
	struct mip_ifs *vmifs;		/* pointer to that interface */
	struct sockaddr_in udp_src;     /* who sent us last msg */
	struct in_addr ha;  		/* home agent's address */
	struct in_addr mn;  		/* mobile node home address */
	struct in_addr coa;		/* for use at ha. mn's coa */
	u_short orig_lifetime;		/* original lifetime request */
	u_short lifetime;       	/* requested lifetime */
#ifdef DIVERT
	u_char  doIpUdp; 		/* nat/udp tunnel needed */
	u_char  hasNat;
#endif
#define ROUTELINK	1
#define ROUTETUNNEL	2
#define ROUTEHOST	3
#ifdef DIVERT
#define IPUDPTUNNEL     4
#endif
	int routeset;			/* if true, route has been set */
	struct identval
	{
		u_long l[2];
	} i;
	u_short udp_port;		/* mn udp port */
/* icmp info, if available */
	u_short link_sequence;
	u_char  sawBeacon;		/* true if beacon recv */
	u_char  sawMac;			/* true if mac passed in */
	u_char  auth_type;		/* auth flags */
	u_char  mac[6];			/* mac address */
	unsigned long tssecs;		/* ntp time in secs */
	unsigned long tsfract;		/* ntp time in fract */
	u_int   proxyArp;		/* true if proxy arp in use */
	char location[MAXLOCATION];	/* location string */
	u_char  lsize;			/* location size */
#ifdef HARP
#ifdef DIVERT
	/* harp-nat info */  
	u_char length;                 /* 6. Length in Bytes of this extension,  
					* not including the Type and Length 
					* Bytes 
					*/
	u_char reserved_1;             /* Reserved For Future Use */  
	u_char reserved_2;             /* 1st Bit(MostSign) ---> F (force) Flag   
					* ; 2nd Bit ----->  (Registration 
					* through FA Required) Flag 
					*/     
	u_char Encapsulation;          /* Indicates the type of Tunneled Data,  
					* using the same numbering as the 
					* IP header Protocol Field 
					*/    
	u_short reserved_3;            /* Reserved For Future Use */    
#endif
#endif
	struct vlist *next;	
};


/* vlist prototypes
*/
struct vlist * vlist_search(struct vlist *, struct in_addr );

struct vlist * vlist_create(struct vlist **, struct in_addr );

void vlist_delete(struct vlist **, struct in_addr );

void vlist_initwalk(struct vlist *);

struct vlist * vlist_nextwalk();

void vlist_dump(struct vlist *);

