/********************************************
 * Implementation of class cMnStat
 * a linux version of mnstat
 * Programmer: Ginny Mak
 * Location: Portland, Oregon
 * Date: Dec. 02, 2001
 *********************************************/

#include<stdio.h>
#include<stdlib.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <fcntl.h>
#ifdef LINUX
#include <error.h>
#endif
#include <errno.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>

#include "../mip_log.h"

#ifdef MADRP
#include "../sr_mad.h"
#endif

#include "cMnStat.h"


/* mnd-mnstat socket names and file mode */ 
#define MNDSOCK_SERVER	"/tmp/mnd.sock.serv"
#define MNDSOCK_CLIENT	"/tmp/mnd.sock.client"
#define XMNDSOCK_CLIENT	"/tmp/xmnd.sock.client"
#define US_MODE		(S_IRUSR | S_IWUSR)

#define U_MNSTAT  1	/* send of mn_stat structure */
#define U_MNSTATE 2	/* send over mn state structure */

#define U_ALIST   4     /* mnd sends over agent list entry by entry */

#define U_MNBEACONS 8	/* send icmp_stats structure */

#define U_MNLIST 10     /* send mnlist to mnstat */

#ifdef MADRP	
#define U_HLIST   18	/* send hlist over to mnstat */
#endif
#ifdef DHCP
#define U_MODESWITCH    19      /* set manual or automatic modes */
#endif


#define FAILURE -1
#define SUCCESS 0

// using server-client communication to get binary data.
// these routines are taken out of unixsock.c file
// of psu mnd modules
int initUnixSocket(char *whoami, int bepolite);
int sendUnixMessage(char *sockfile, char *buf, int size);
int recvUnixMessage(char *buf, int bufsize);           

/**************************************************
 * Implementation of class cMnStat                *
 **************************************************/
cMnStat::cMnStat() 
{
  int rc = initUnixSocket( MNDSOCK_CLIENT, TRUE );
  if( rc < 0 ) {
	perror( "socket setup failed" );
	exit( FAILURE );
  }
  mInitialized = true;
}

cMnStat::~cMnStat() 
{}


int cMnStat::getBeaconData( char beaconFields[][10] )
{
  int rc;
  if( !mInitialized )
	return FAILURE;

  buf[0] = U_MNBEACONS;
  rc = sendUnixMessage( MNDSOCK_SERVER, buf, 1 );
  if( rc < 0 ) {
	return FAILURE;
  }
  rc = recvUnixMessage( (char *)&icmpstats, sizeof(struct icmp_stats) );
  if( rc < 0 ) {
	return FAILURE;
  }
  dumpBeaconStats( beaconFields );
  
  return SUCCESS;
}


void cMnStat::dumpBeaconStats( char beaconFields[][10] )
{
  int badauth;
  badauth = icmpstats.ha_netauthbad +
	icmpstats.fa_netauthbad +
	icmpstats.mn_netauthbad +
	icmpstats.ha_adhoc_authbad +
	icmpstats.fa_adhoc_authbad +
	icmpstats.mn_adhoc_authbad;
	
  sprintf( beaconFields[0], "%d", 
		   icmpstats.ha_beacons_sent+
		   icmpstats.fa_beacons_sent+
		   icmpstats.combined_beacons_sent+
		   icmpstats.mn_beacons_sent );
  sprintf( beaconFields[1], "%d", icmpstats.all_beacons_recv );
  sprintf( beaconFields[2], "%d", icmpstats.all_netauth_recv );
  sprintf( beaconFields[3], "%d", icmpstats.valid_beacons_recv );
  sprintf( beaconFields[4], "%d", icmpstats.valid_arp_beacons );
  sprintf( beaconFields[5], "%d", icmpstats.valid_auth_beacons );
  sprintf( beaconFields[6], "%d",
		   icmpstats.malformed+
		   icmpstats.msetooshort+
		   icmpstats.invalidicmpmse+
		   icmpstats.auth_nomac );
  sprintf( beaconFields[7], "%d", badauth );
  sprintf( beaconFields[8], "%d", 
		   icmpstats.ha_maybe_arp_lacks_auth+
		   icmpstats.fa_maybe_arp_lacks_auth );
		   
}
  
  
int cMnStat::getUdpData( char udpFields[][10] )
{
  int rc;
  if( !mInitialized )
	return FAILURE;

  buf[0] = U_MNSTAT;
  rc = sendUnixMessage( MNDSOCK_SERVER, buf, 1 );
  if( rc < 0 ) {
	return FAILURE;
  }
  rc = recvUnixMessage( (char *)&mnstats, sizeof(struct mn_stats) );
  if( rc < 0 ) {
	return FAILURE;
  }
  dumpUdpStats( udpFields );
  return SUCCESS;
}

void cMnStat::dumpUdpStats( char udpFields[][10] )
{
  sprintf( udpFields[0], "%d", mnstats.reg );
  sprintf( udpFields[1], "%d", mnstats.replies );
  sprintf( udpFields[2], "%d", mnstats.identnomatch );
  sprintf( udpFields[3], "%d", mnstats.invalidhaauth );
  sprintf( udpFields[4], "%d", mnstats.invalidfaauth );
  sprintf( udpFields[5], "%d", mnstats.missinghaauth );
  sprintf( udpFields[6], "%d", mnstats.missingfaauth );
  sprintf( udpFields[7], "%d", mnstats.timeupdate );
}
  
int cMnStat::getState( mnstate *aMnState )
{
   int rc;
  if( !mInitialized )
	return FAILURE;

  buf[0] = U_MNSTATE;
  rc = sendUnixMessage( MNDSOCK_SERVER, buf, 1 );
  if( rc < 0 ) {
	return FAILURE;
  }
  rc = recvUnixMessage( (char *)&mstate, sizeof(struct mn_state) );
  if( rc < 0 ) {
	return FAILURE;
  }
  dumpStateData( aMnState );
  return SUCCESS;
}



// this function modifies dumpMnState function of mndumpstate.c
// in PSU mnd module
void cMnStat::dumpStateData( mnstate *aMnState )
{
  switch( mstate.state ) {
  case NOWHERE:
	sprintf( aMnState->state, "State: NOWHERE" );
	break;
  case ATHOME:
	if( mstate.retry_timer_flag ) 
	  sprintf( aMnState->state, "State: ATHOME-PENDING" );
	else
	  sprintf( aMnState->state, "State: ATHOME" );
	break;
  case AWAY:
	if( mstate.retry_timer_flag ) {
	  if( mstate.retryTimeouts )
		sprintf( aMnState->state, "State: AWAY-PENDING-DISCONNECTED?" );
	  else
		sprintf( aMnState->state, "State: AWAY-PENDING" );
	}
	else
	  sprintf( aMnState->state, "State: AWAY-ACKED" );
	break;
	#ifdef DHCP
  case CO_LOCATIED:
	if( mstate.retry_timer_flag ) {
	  if( mstate.retryTimeouts )
		sprintf( aMnState->state, "State: CO_LOCATED-PENDING-DISCONNECTED?" );
	  else
		sprintf( aMnState->state, "State: CO_LOCATED-PENDING" );
	}
	else
	  sprintf( aMnState->state, "State: CO_LOCATED-ACKED" );
	break;
	#endif
  } // end switch
  #ifdef MADRP
  switch( mstate.sr_state ) {
  case NOWHERE:
	sprintf( aMnState->state, "sr_state: NOWHERE" );
	break;
  case ATHOME:
	if( mstate.retry_timer_flag ) 
	  sprintf( aMnState->state, "sr_state: ATHOME-PENDING" );
	else
	  sprintf( aMnState->state, "sr_state: ATHOME" );
	break;
  case AWAY:
	if( mstate.retry_timer_flag ) {
	  if( mstate.retryTimeouts )
		sprintf( aMnState->state, "sr_state: AWAY-PENDING-DISCONNECTED?" );
	  else
		sprintf( aMnState->state, "sr_state: AWAY-PENDING" );
	}
	else
	  sprintf( aMnState->state, "sr_state: AWAY-ACKED" );
	break;
  } // end switch
  #endif

  if( mstate.fa.s_addr != 0 ) {
	sprintf( aMnState->fa, "fa: %s", inet_ntoa( mstate.fa ) );
	sprintf( aMnState->coa, "coa: %s", inet_ntoa( mstate.coa ) );
  }
}
	  

int cMnStat::getAgents( int *count, char agentBeacon[][100],
                        int agentSignal[],
                        char agentLocation[][100] ) 
{
  int rc;

  if( !mInitialized )
	return FAILURE;

  buf[0] = U_ALIST;
  rc = sendUnixMessage( MNDSOCK_SERVER, buf, 1 );
  if( rc < 0 ) {
	return FAILURE;
  }
  *count = 0;
  rc = recvUnixMessage( (char *)count, sizeof(int) );
  if( rc < 0 ) {
	return FAILURE;
  }
  for( int i = 0; i < *count; i++ ) {
	rc = recvUnixMessage( (char *)&agent, sizeof( struct alist ) );
	if( rc < 0 )
	  return FAILURE;
	dumpAgentData( i, agentBeacon[i], &agentSignal[i], agentLocation[i] );
  }
  return SUCCESS;
}

void cMnStat::dumpAgentData( int i, char *beacon, int *signal, char *location )
{
  char abuf[1024];

  sprintf(abuf, "%s", inet_ntoa(agent.bip));
  sprintf( beacon, "BeaconIP/Coa %s / %s", abuf, inet_ntoa( agent.coa ) );
  sprintf( location, "Location %s", agent.location );
  *signal = agent.sigstrength;
}


int cMnStat::getNodes( int *count, char nodeBeacon[][100],
                       int nodeSignal[],
                       char nodeLocation[][100] ) 
{
  int rc;

  if( !mInitialized )
	return FAILURE;

  buf[0] = U_MNLIST;
  rc = sendUnixMessage( MNDSOCK_SERVER, buf, 1 );
  if( rc < 0 ) {
	return FAILURE;
  }
  *count = 0;
  rc = recvUnixMessage( (char *)count, sizeof(int) );
  if( rc < 0 ) {
	return FAILURE;
  }
  for( int i = 0; i < *count; i++ ) {
	rc = recvUnixMessage( (char *)&node, sizeof( struct alist ) );
	if( rc < 0 )
	  return FAILURE;
	dumpNodeData( i, nodeBeacon[i], &nodeSignal[i], nodeLocation[i] );
  }
  return SUCCESS;
}

void cMnStat::dumpNodeData( int i, char *beacon, int *signal, char *location )
{
  char anode[1024];

  sprintf(anode, "%s", inet_ntoa(node.bip));
  sprintf( beacon, "BeaconIP/Coa %s/%s", anode,
		  inet_ntoa( node.coa ) );
  sprintf( location, "Location %s", node.location );
  *signal = node.sigstrength;
}


/*****************************************************************
 * The following functions are taken out of unixsock.c in
 * PSU's mnd modules.
 *****************************************************************/

static int linuxSocket;

/*
 * bind a unix datagram socket to sockfile
 *
 * returns:
 *	-1: failed
 *	0-n: socket descriptor value
 */
int initUnixSocket(char *whoami, int bepolite)
{
	int sd;
	int rc;
	struct sockaddr_un self;
	extern int errno;
	struct stat sbuf;
 
	sd = socket(AF_UNIX, SOCK_DGRAM, 0);
	if (sd < 0) {
		syslog( LOG_ERR, "initUnixSocket(): socket (%m)");
		return(-1);
	}

	/* Set close-on-exec flag to prevent fd leaks. */
	if (fcntl(sd, F_SETFD, 1) < 0) {
		syslog(LOG_INFO, "initUnixSocket(): F_SETFD (%m)");
		/* not catastrophic, don't exit */
	}

	self.sun_family = AF_UNIX;
	strcpy(self.sun_path, whoami);


	/* politeness is required by statistics gathering programs
	 * and is improper for daemons.
	 */
	if (bepolite) {
		/*
		 * politeness: exit if the socket exists.
		 */
		if (stat(whoami, &sbuf) == 0) {
		  //printf("socket busy.\n");
		  //exit(0);
		}
	} else {
		/* blast socket file, in case it is still there
		*/
		unlink(whoami);
	}

	rc = bind(sd, (struct sockaddr *)&self, sizeof(struct sockaddr_un));
	if (rc < 0) {
		syslog( LOG_ERR, "initUnixSocket(): bind (%m).");
		return(-1);
	}
	if ( chmod( whoami, US_MODE ) < 0 ) {
		syslog( LOG_ERR, "initUnixSocket(): chmod (%m).");
		return(-1);
	}

	linuxSocket = sd;
	return(sd);
}

/*
 * send a message to the server bound to sockfile
 *
 * returns: -1, failed
 *	     0, ok
 */
int
sendUnixMessage(char *sockfile, char *buf, int size) 
{
	struct sockaddr_un to;
	int rc;
	extern int errno;

	to.sun_family = AF_UNIX;
	strcpy(to.sun_path, sockfile);

again:
	rc = sendto(linuxSocket, buf, size, 0,  (struct sockaddr *)
		&to, sizeof (struct sockaddr_un));
	if ((rc < 0))
		goto again;

	if (rc < 0) {
		syslog( LOG_ERR, "sendUnixMessage(): sendto failed (%m).");
		return(-1);
	}
	return(0);
}

/*
 * recvUnixMessage
 *
 * 	recv message on UNIX datagram socket that bind
 *	was performed on.
 *
 * returns: -1, failed
 *	     rc > 0, amount actually received
 */
int
recvUnixMessage(char *buf, int bufsize)
{
	socklen_t size = sizeof(struct sockaddr_un);
	int rc;
	struct sockaddr_un from;

	rc = recvfrom(linuxSocket, buf, bufsize, 0, (struct sockaddr *)&from, &size);
	if (rc < 0) {
		syslog( LOG_ERR, "recvUnixMessage(): recvfrom failed (%m).");
		return(-1);
	}
	return(rc);
}


