/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "mnconfig.h"
#include "mip_log.h"

#define MN 0 
#define HA 1
#define MNHA_KEY 2
#define FAMN_KEY 3
#define NETAUTH_KEY 4
#define ADHOC_KEY 5
#define TSMODE 6
#define REPLAY 7
#define BEACON 8
#define BEACON_TIMESLOP 9
#define BEACON_REPLAY 10
#define ARP 11
#define NETAUTH 12
#define ADHOC 13
#define LOCATION 14
#define UDPPORT 15
#define LOGIPONLY 16

#define NUM_DIRECTIVES 17

/* write out config file
 *
 * overwrites: /etc/mnd.conf
 * 1st writes to mnd.conf.bak
 * saves copy of previous config file in mnd.conf.safe
 */

void op_switch(int op);

FILE *fp_w, *fp_r;


void write_mndconfig()
{
	int i;
	char linebuf[1024];
	char opbuf[1024];
	char *result;
	char *opword;

	char *directives[]={"mn","ha","mnha_key","famn_key","netauth_key",
		"adhoc_key","tsmode","replay","beacon","beacon_timeslop",
		"beacon_replay","arp","netauth","adhoc","location",
		"udp_port","logIPOnly"
		};
	int directive_done[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
		};

	fp_w = fopen(CONFIGDEFAULTBAK,"w"); 
	if (fp_w == NULL) {
		syslog(LOG_ERR, "write config: fopen of %s failed", CONFIGDEFAULTBAK);
		exit(1);
	} 

	fp_r = fopen(CONFIGDEFAULT,"r");  /* open config file for reading */
	if (fp_r == NULL) {
		syslog(LOG_ERR, "read config: fopen of %s failed", CONFIGDEFAULT);
		exit(1); 
	}

	while (feof(fp_r) == 0){
		result = fgets(linebuf, MAXLINE, fp_r);
		if (result == NULL)
			break;

		/* write comments to file */
		if (linebuf[0] == '#'){
			fprintf(fp_w, "%s", linebuf);
			continue;
		}

		/* skip empty lines */
		if (linebuf[0] == '\n' || linebuf[0] == '\r')
			continue;

		/* skip whitespace */
		while (*result == ' ' || *result == '\t')
			result++;
		strncpy(opbuf, result, 1024);
		opword = strtok(opbuf, " \t");
		if (opword == NULL || (strlen(opword) == 0))
			continue;

		for(i=0; i< NUM_DIRECTIVES; i++){
			if (strcmp(opword, directives[i]) == 0)
				break;
		}
		if ( i == NUM_DIRECTIVES ){
			syslog(LOG_ERR, "bad directive op\n");
		}
		else{
			directive_done[ i ] = TRUE; /* set flag to done */
			op_switch( i ); 	/* write directive to file */
		}
	}

	for (i=0; i< NUM_DIRECTIVES; i++){
		if (directive_done[i] == FALSE){
			op_switch( i );
		}
	}
			
	fclose(fp_r);
	fclose(fp_w); 

	/* copy original config file to conf.safe
	*/
	unlink(CONFIGDEFAULTSAFE); 
	link(CONFIGDEFAULT, CONFIGDEFAULTSAFE); 

	/* copy new config file to .conf
	 * We do not zap conf.bak just in case something went wrong.
	 */
	unlink(CONFIGDEFAULT);
	link(CONFIGDEFAULTBAK, CONFIGDEFAULT);
	unlink(CONFIGDEFAULTBAK); 
}

/* switch on config file operator
*  and write out line to config file
*/
void op_switch(int op)
{
	int i;

	char *policy_vals[]={"reject","accept"};
	char *replay_type[]={"nonce", "timestamp"};
	char *bool[]={"off","on"};

	switch(op){
		case MN: 		 /* mn	<ifname>   <ipaddr> */
			fprintf(fp_w, "mn\t%s\t%s\n", mn_config.ifname, inet_ntoa(mn_config.mn));
			break;
		case BEACON:		 /* beacon  <seconds>  */
			fprintf(fp_w, "beacon\t%d\n", mn_config.beaconRate);
			break;
		case HA:		 /* ha  <ipaddr>  */
			fprintf(fp_w, "ha\t%s\n", inet_ntoa(mn_config.ha));
			break;
		case MNHA_KEY:		 /* mnha_key  <transform>  <spi>  <hexstring> */
			if (mn_config.mnha_security){ 
				fprintf(fp_w, "mnha_key\t%s\tmd5\t%d\t0x", inet_ntoa(mn_config.ha),mn_config.sa[0].spi);
				for(i=0;i<16; i++)
					fprintf(fp_w,"%2.2x",mn_config.sa[0].key[i]);
				fprintf(fp_w,"\n");
				
			}
			break;
		case FAMN_KEY:		 /* famn_key  <transform>  <spi>  <hexstring>  */
			if (mn_config.famn_security){ 
				fprintf(fp_w, "famn_key\tmd5\t%d\t0x", mn_config.sa[1].spi);
				for(i=0; i<16; i++)
					fprintf(fp_w, "%2.2x", mn_config.sa[1].key[i]);
				fprintf(fp_w, "\n");
			}
			break;
		case NETAUTH_KEY:		 /* net_auth_key   <transform>  <spi>  <hexstring> */ 
			if (mn_config.netauth_security){ 
				fprintf(fp_w, "netauth_key\tmd5\t%d\t0x", mn_config.sa[2].spi);
				for(i=0; i<16; i++)
					fprintf(fp_w, "%2.2x", mn_config.sa[2].key[i]);
				fprintf(fp_w, "\n");
			}
			break;
		case ADHOC_KEY:		 /* adhoc_key  <transform>  <spi>  <hexstring> */
			if (mn_config.adhoc_security){ 
				fprintf(fp_w, "adhoc_key\tmd5\t%d\t0x", mn_config.sa[3].spi);
				for(i=0; i<16; i++)
					fprintf(fp_w, "%2.2x", mn_config.sa[3].key[i]);
				fprintf(fp_w, "\n");
			}
			break;
		case TSMODE:		 /*  tsmode  <0|1> */
			fprintf(fp_w, "tsmode\t%d\n", mn_config.tsmode);
			break;
		case BEACON_REPLAY:		 /*  beacon_replay  <timestamp|nonce>  <on|off>  */
			fprintf(fp_w, "beacon_replay\t%s\t%s\n", replay_type[mn_config.doTimestamp], bool[mn_config.beacon_replay]);
			break;
		case BEACON_TIMESLOP:		 /* beacon_timeslop  <seconds> */
			fprintf(fp_w, "beacon_timeslop\t%d\n", mn_config.beacon_timeslop);
			break;
		case ARP:		 /* arp   <on|off>  */
			fprintf(fp_w, "arp\t%s\n", bool[mn_config.doArp]);
			break;

		case NETAUTH:		/* netauth <accept | reject> */
			fprintf(fp_w, "netauth\t%s\n", policy_vals[mn_config.doNetauth]);
			break;
		case ADHOC:		/*  adhoc        <accept | reject> */
			fprintf(fp_w, "adhoc\t%s\n", policy_vals[mn_config.doAdhoc]);
			break;
		case REPLAY: 		/* replay    <timestamp | nonce> */
			fprintf(fp_w, "replay\t%s\n", mn_config.doTimestamp ? "timestamp" : "nonce");
			break;
		case LOCATION:		/* location 	<location string> */
			fprintf(fp_w, "location\t\"%s\"\n", mn_config.location);
			break;
		case UDPPORT:		/* udp_port	<number> */
			fprintf(fp_w, "udp_port\t%d\n", mn_config.udp_port);
			break;
		case LOGIPONLY:		/* logIPOnly	<off | on> */
			fprintf(fp_w, "logIPOnly\t%s\n", mn_config.logIPOnly ? "on" : "off" );
			break;
		default:
			syslog(LOG_ERR, "undefined config directive\n");
			return;
	}
}


