/*
 * mvifconfig.c
 *
 * configure mvif tunnel device with ip address for inner/outer tunnel
 * syntax:
 *	mvifconfig -w -i ipaddr  # use ipaddr for inner ip src
 *	mvifconfig -w -o ipaddr  # use ipaddr for outer ip src
 *	mvifconfig -r 		 # read out both addresses
 *			default as well
 *
 * note:
 *	1. passing in 0.0.0.0  turns off this facility
 *	2. normally the outer ip address does not need to be set
 *		as ip will set it correctly since the pkt is
 *		sent out a real interface.
 */
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <net/if_var.h>
#include <net/if_dl.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <arpa/inet.h>
#include <netinet/if_mvif.h>

#include <netdb.h>

#define EON
#include <sys/protosw.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <nlist.h>
#include <kvm.h>
#include <fcntl.h>

/* prototypes
*/
void parseArgs(int argc, char **argv);
void setOuterIP(int , struct in_addr);
void setInnerIP(int , struct in_addr);

struct	ifreq	ifr;

int readflag;
int writeflag;
int inipflag;
int outipflag;
struct in_addr inip;
struct in_addr outip;

main(argc, argv)
	int argc;
	char *argv[];
{
	int s;
	extern int errno;

	parseArgs(argc, argv);

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("socket");
		exit(1);
	}
	sprintf(ifr.ifr_name,"%s", "mvif0");
	ifr.ifr_addr.sa_family = AF_INET;

	if (writeflag && inipflag) {
		setInnerIP(s, inip);
	}

	if (writeflag && outipflag) {
		setOuterIP(s, outip);
	}

	if (readflag) {
		inip.s_addr = getInnerIP(s);
		printf("inner ip %s\n", inet_ntoa(inip));
		outip.s_addr = getOuterIP(s);
		printf("outip ip %s\n", inet_ntoa(outip));
	}
	close (s);
}

getInnerIP(int s)
{
	if (ioctl(s, SIOCGMVINIP, (char *) &ifr) < 0) {
		perror("ifconfig (SIOCGMVINIP)");
		exit (1);
	}
	return((int)ifr.ifr_data);
}

getOuterIP(int s)
{
	if (ioctl(s, SIOCGMVOUTIP, (char *) &ifr) < 0) {
		perror("ifconfig (SIOCGMVOUTIP)");
		exit (1);
	}
	return((int)ifr.ifr_data);
}

void 
setInnerIP(int s, struct in_addr inip)
{
	ifr.ifr_data = (caddr_t) inip.s_addr;
	if (ioctl(s, SIOCSMVINIP, (char *) &ifr) < 0) {
		perror("ifconfig (SIOCSMVINIP)");
		exit (1);
	}
}

void
setOuterIP(int s, struct in_addr outip)
{
	ifr.ifr_data = (caddr_t) outip.s_addr;
	if (ioctl(s, SIOCSMVOUTIP, (char *) &ifr) < 0) {
		perror("ifconfig (SIOCSMVOUTIP)");
		exit (1);
	}
}

void
parseArgs(argc, argv)
	int             argc;
	char          **argv;
{
	extern int      errno;
	extern char    *optarg;
	extern int      optind;
	int             ch, fd;
	int 		rc;
	char 		inipaddr[256];
	char 		outipaddr[256];

	readflag = 1;
	writeflag = 0;
	while ((ch = getopt(argc, argv, "rwi:o:")) != EOF) {
		switch (ch) {
		case 'r':
			readflag = 1;
			break;
		case 'w':
			writeflag = 1;
			break;
		case 'o':
			writeflag = outipflag = 1;	
			strncpy(outipaddr, optarg, 256);
			outip.s_addr = inet_addr(outipaddr);
			break;
		case 'i':
			writeflag = inipflag = 1;
			strncpy(inipaddr, optarg, 256);
			inip.s_addr = inet_addr(inipaddr);
			break;
		case '?':
		default:
			usage();
		}
	}

	if (outipflag == 0 && inipflag == 0 && writeflag) {
		usage();
	}

}

usage()
{
	fprintf(stderr,"mvifconfig: -r - dump mvif ip addresses\n");
	fprintf(stderr,"mvifconfig: -w -i <ipaddr> - set inner ip addr\n");
	fprintf(stderr,"mvifconfig: -w -o <ipaddr> - set outer ip addr\n");
	exit(1);
}

